/*
 * InferenceCore.java
 *
 * Created on 24 luglio 2004, 15.46
 */

package it.unive.dsi.pear.core;

import java.util.*;
import java.io.*;

import it.unive.dsi.pear.util.*;

import it.unive.dsi.pear.log.Log;
import it.unive.dsi.pear.core.parser.KeyType;
import it.unive.dsi.pear.core.parser.TaggedData;
import it.unive.dsi.pear.core.parser.TagType;
import it.unive.dsi.pear.core.parser.CipherInstruction;
import it.unive.dsi.pear.core.parser.Ciphertext;
import it.unive.dsi.pear.core.parser.Data;
import it.unive.dsi.pear.core.parser.Instruction;
import it.unive.dsi.pear.core.parser.InstructionNode;
import it.unive.dsi.pear.core.parser.Operation;
import it.unive.dsi.pear.core.parser.Protocol;
import it.unive.dsi.pear.core.parser.Rule;
import it.unive.dsi.pear.core.parser.SequentialProcess;
import it.unive.dsi.pear.core.parser.ValidationRules;
import it.unive.dsi.pear.core.evaluationtree.EvaluationTree;


/** This class implements the inference core.
 * @author Francesco Placella
 */
public class InferenceCore extends ValidationCore
    implements Inference
{
    private static final Map EMPTY_MAP = new HashMap(0);
    
    private static final Integer TAGGING_ITR = new Integer(-1);
    private static final int TAGGING_BRAND = 0;
    
    private static final int MAX_BT_LIMIT = Integer.parseInt(System.getProperty("btLimit", "2000"));
    private static final int EQUAL_LIMIT = 1;
    
    private static final int ENCRYPT_VALUE = 5;
    private static final int COMMIT_VALUE = 100;    
    private static final int TAGGABLE_VALUE = 2;
    private static final int DISTANCE_VALUE = 1;

    private static final String NEW = "new";
    private static final String RUN = "run";
    private static final String INPUT = "in";
    private static final String ENCRYPT = "encrypt";
    private static final String DECRYPT = "decrypt";
    private static final String DECRYPT_INIT = DECRYPT.substring(0, 1);
    static final String COMMIT = "commit";
    
    private static final Set TAGGABLE =
        new HashSet(Arrays.asList(new String[] { DECRYPT, ENCRYPT }));

    private static final Instruction[] INSTR_ARRAY = {};
    private static final Protocol[] PN_ARRAY = {};
    private static final SequentialProcess[] SP_ARRAY = {};

    private final Map substMaps = new HashMap();
    private final Map backupMap = new HashMap();
    
    private Map tagMap;
    private Map identifierTypeMap;
    private Protocol curTagging;
    private SequentialProcess curSeqProc;
    private SequentialProcess[] originalOrder;
    private InstructionNode lastInstruction;
    private Rule previousRule;
    private CommitHints commitHints;
    private boolean usingCommitHintsDefault;
    private boolean usingCommitHints;
    private boolean usingProtocolReordering = true;
    
    /** Holds value of property equalLimit. */
    private int equalLimit = EQUAL_LIMIT;    
    
    // ---------------------------------------------------------- inner classes

    private final class IdInfo
    {
        private String id;
        private InstructionNode instr;
        private int index;

        private IdInfo(String id,  InstructionNode instr, int index)
        {
            this.id = id;
            this.instr = instr;
            this.index = index;
        }

        public String toString() { return id + " (" + instr + ")"; }
    }

    
    // ----------------------------------------------------------- constructors
    
    /** Creates a new instance of InferenceCore
     * @throws InitializationException Any exception thrown during the
     * initialization process is caught and set as cause of a new instance of
     * <CODE>InitializationException</CODE>.
     */
    public InferenceCore()
        throws InitializationException
    {
        super(Log.VOID_LOG);
    }

    /** Creates a new instance of InferenceCore
     * @param log an implemetation of the logger
     * @throws InitializationException Any exception thrown during the
     * initialization process is caught and set as cause of a new instance of
     * <CODE>InitializationException</CODE>.
     */
    public InferenceCore(Log log)
        throws InitializationException
    {
        super(log);
    }
    
    /** Creates a new instance of InferenceCore
     * @param rules the validation rules to be set as the current ones
     * @param initLog an implementation of the logger
     * @throws InitializationException Any exception thrown during the
     * initialization process is caught and set as cause of a new instance of
     * <CODE>InitializationException</CODE>.
     */
    public InferenceCore(ValidationRules rules, Log initLog)
        throws InitializationException
    {
        super(rules, initLog);
    }
    
    
    // ------------------------------------------------------------------- main
    
    /** Runs the inference core in single-threaded mode.
     * @param args a file with the name equal to the value of the first argument
     * is looked for and, if found, opened.
     */
    public static final void main(String[] args)
        
    {
        Reader is;
        try
        {
            is = new BufferedReader(
                    (args.length > 0 ?
                        new FileReader(args[0]) :
                        new InputStreamReader(System.in)));

            InferenceCore ic = new InferenceCore();
            Protocol taggings[] =
                ic.inferTaggings(ic.parseProtocol(is));
            
            Log log = Log.SIMPLE_LOG;
            log.out("");
            int size = taggings.length;
            for (int i = 0; i < size; ++i)
                log.out(
                    "[tagging #" + (i + 1) + "] " + taggings[i].getTreeImage());
            
            log.info("finshed.");
        }
        catch (Exception e) { e.printStackTrace(System.err); }
    }

    
    // -------------------------------------------------------------- inference

    /** Infers all the possbile tagging for the specified protocol according to
     * the current validation rule set.
     * @param protocol the protocol whose tagging are to be inferred
     * @return an array containing the found taggings
     */    
    public synchronized 
        Protocol[] inferTaggings(final Protocol protocol)
    {
        curTagging = (Protocol)protocol.clone();
        identifierTypeMap = this.computeIdentifierTypes(curTagging);
        this.reorderProtocol(curTagging);
        
        int equalCounter = 0;
        int minBTNode = Integer.MAX_VALUE;
        boolean finished = false;
        Collection taggings = new HashSet();
        
        btHandler.setCurState(null);
        backupMap.clear();

        while (true)
        {
            substMaps.clear();
            
            try
            {
                if (this.validateProtocol(curTagging))
                {
                    if (!taggings.contains(curTagging))
                    {
                        taggings.add(curTagging.clone());
                        log.info(curTagging.getTreeImage());
                    }
                    else ++equalCounter;
                }
            }
            catch (StoppedException se) { break; }
            catch (BackTrackingHandler.LimitExcededException lee)
            {
                log.error(lee);
            }
            
            int available = btHandler.availableCount();
            if (finished || available == 0) break;
            
            try
            {
                // goes back to last "tagging" backtrack node
                finished = this.findTaggingBTNode(Integer.MAX_VALUE) == 0;
                
                // if equal counter is greater than limit, backtracks to the
                // first "tagging" node with "level" less than the minimum
                // reached until now
                if (equalCounter < equalLimit)
                {
                    available = btHandler.availableCount();
                    if (available < minBTNode) minBTNode = available;
                }
                else
                {
                    equalCounter = 0;
                    minBTNode = this.findTaggingBTNode(minBTNode);
                    finished = minBTNode == 0; 
                }
                
                // if the end of the stack is reached and still there is no
                // "tagging" node, quits, else it will after the next
                // validation session
                BackTrackingHandler.State state = btHandler.getCurState();
                if (finished && !this.isTaggingNode(state)) break;
            }
            catch (BackTrackingHandler.LimitExcededException lee)
            {
                log.error(lee);
            }
        }
        
        Iterator it = taggings.iterator();
        while (it.hasNext())
        {
            Protocol tagging = (Protocol)it.next();
            this.restoreProtocolOrder(tagging);
            tagging.setTagging(true);
        }

        Protocol[] res = 
            (Protocol[])taggings.toArray(PN_ARRAY);
        
        if (!this.isStopped() && !protocol.isTagging())
            protocol.setTaggings(res);
        
        return res;
    }
    
    private final boolean isTaggingNode(BackTrackingHandler.State state)
    {
        return state.getBrand() == TAGGING_BRAND;
    }
    
    /* findTaggingBTNode looks for tagging nodes in the backtracking handler
     * under the stack level specified as argument.
     */
    private int findTaggingBTNode(int min)
        throws BackTrackingHandler.LimitExcededException
    {
        BackTrackingHandler.State state = btHandler.getCurState();
        int available = btHandler.availableCount();
        while (available > 0)
        {
            if (this.isTaggingNode(state) && available < min)
                break;
                
            btHandler.backTrack();
            state = btHandler.getCurState();
            available = btHandler.availableCount();
        }

        return available;
    }
    
    /** Overridden method: keeps track of the current instruction. */
    public boolean validateInstruction(final InstructionNode instruction)
        throws BackTrackingHandler.LimitExcededException, InterruptedException
    {
        return super.validateInstruction(lastInstruction = instruction);
    }

    /** Overridden method: previous tagging is removed. */
    protected void noRuleApplied(InstructionNode instruction)
    {
        super.noRuleApplied(instruction);
        this.removePreviousTagging(instruction);
    }
    
    /** Overriden method: tries to find a valid tagging for the current
     * instruction using the specified rule.
     */    
    protected Map patternMatch(InstructionNode instruction, Rule rule)
    {
        // only "forward tagging" is permitted
        InstructionNode ci = instruction.getComplement();
        if (ci != null && instruction.isForwardComplement())
            return super.patternMatch(instruction, rule);
        
        BackTrackingHandler.State state = btHandler.getCurState();
        Iterator it = (Iterator)state.getFieldValue(TAGGING_ITR);
        boolean alreadyTagged = it != null;

        // if instruction is already tagged, removes prevous tagging, otherwise 
        // if rule and instruction aren't the same of the prevoius time,
        // updates tagging (after removing previous one, if present)
        if (alreadyTagged) this.removePreviousTagging(instruction);
        else
        {
            if (instruction != lastInstruction || rule != previousRule)
            {
                // if true, in btHandler there stack might be useless autosubst
                // nodes that must be removed in order not to generate loops
                if (!rule.equals(previousRule) &&
                    this.returningBackTrackingNode())
                {
                    if (alreadyTagged = this.removeUselessNodes())
                    {
                        state = btHandler.getCurState();
                        it = (Iterator)state.getFieldValue(TAGGING_ITR);
                    }
                    else return null;
                }

                previousRule = rule;
                this.removePreviousTagging(instruction);
            }
        }

        if (!alreadyTagged)
        {
            curSeqProc = instruction.getSequentialProcess();
            Map identifierTypes = this.getIdentifierTypes(curSeqProc);
            Instruction[] taggings =
                this.getTaggings(instruction, rule, identifierTypes);
            it = new ArrayIterator(taggings);
            state.setFieldValue(TAGGING_ITR, (CloneableIterator)it);
        }
        
        Map substMap = null;
        if (it.hasNext())
        {
            Instruction tagging = (Instruction)it.next();
            Instruction instr = instruction.getInstruction();
            
            if (tagging.containsTagged())
            {
                Instruction comp = null;
                if (ci != null)
                {
                    comp = ci.getInstruction();
                    this.tagComplement(instruction, tagging);
                }

                List pair = Arrays.asList(new Instruction[]{ instr, comp });
                backupMap.put(tagging, pair);
                instruction.setInstruction(tagging);
                substMap = (Map)substMaps.remove(tagging);
                log.info("current tagging: " + tagging, Log.FULL_OUTPUT);
            }
            else substMap = (Map)substMaps.remove(tagging);
            
            if (it.hasNext()) this.save(TAGGING_BRAND);
        }
        
        if (substMap == null)        
            substMap = super.patternMatch(instruction, rule);
        
        state.setFieldValue(TAGGING_ITR, null);
        return substMap;
    }

    /** Removes the previous tagging of the instruction, if any
     * @param instruction the current instruction
     */    
    protected void removePreviousTagging(InstructionNode instruction)
    {
        Instruction tagging = instruction.getInstruction();
        List pair = (List)backupMap.remove(tagging);
        if (pair != null)
        {
            Instruction instr = (Instruction)pair.get(0);
            instruction.setInstruction(instr);
            log.info("previous tagging removed: " + instr, Log.VERBOSE);
            InstructionNode ai = instruction.getComplement();
            instr = (Instruction)pair.get(1);
            if (ai != null && instr != null) ai.setInstruction(instr);
            log.info("complement tagging removed: " + instr, Log.VERBOSE);
        }
    }
    
    /** Overridden method: sets the initial state of the backtracking handler
     * to the appropriate value in order to restart validation from the needed 
     * point.
     */    
    protected BackTrackingHandler.State initBTHandler()
    {
        btHandler.setMaxBTNumber(MAX_BT_LIMIT);
        BackTrackingHandler.State state = btHandler.getCurState();
        if (state == null) state = super.initBTHandler();
        return state;
    }
    
    /** Overridden method: does nothing. In this case the validation process is
     * allowed to backtrack to previous sequential processes.
     */
    protected void resetBTHandler() {}

    /** Overridden method: does nothing. The instruction doesen't need decoration:
     * this way performances are improved.
     */
    protected void decorateInstruction(Rule rule) {}
    
    /** Overridden method: does nothing. Back steps are not allowed in this 
     * case.
     */    
    protected void handleBackStep(StepLevel stepLevel)
        throws ResetStepException {}

    /** Overridden method: before setting the validation rule set specified, all
     * the needed related data are collected.
     */    
    public void setValidationRules(ValidationRules vr)
        throws InitializationException
    {
        super.setValidationRules(vr);
        EvaluationTree et = (EvaluationTree)this.getValidationRules();
        tagMap = vr.getTagTypes();
        commitHints = CommitHints.collect(et);
        usingCommitHintsDefault = !this.isTaggedDecryptionDefined(et);
        usingCommitHints = usingCommitHintsDefault;
    }

    
    // ------------------------------------------------------ auxiliary methods
    
    /* computeIdentifierTypes computes the tag types that may be assigned to
     * each identifier appearing in the specified protocol.
     */
    private Map computeIdentifierTypes(Protocol pn)
    {
        int size = pn.getChildCount();
        Map identifierTypeMap = new HashMap(size);
        Map nonceSets = new HashMap(size);
        
        // scans sequential processes for known id types
        Iterator it = pn.seqProcIterator();
        while (it.hasNext())
        {
            SequentialProcess sp = (SequentialProcess)it.next();
            Set nonceSet = new HashSet();
            Map identifierTypes = new HashMap();
            this.scanSequentialProcess(sp, identifierTypes, nonceSet);
            nonceSets.put(sp, nonceSet);
            identifierTypeMap.put(sp, identifierTypes);
        }
        
        // tries to cross data through complement instructions
        it = pn.seqProcIterator();
        while (it.hasNext())
        {
            SequentialProcess sp = (SequentialProcess)it.next();
            Set nonceSet = (Set)nonceSets.get(sp);
            Map identifierTypes = (Map)identifierTypeMap.get(sp);
            
            Iterator jt = nonceSet.iterator();
            while (jt.hasNext())
            {
                IdInfo idInfo = (IdInfo)jt.next();
                if (!identifierTypes.containsKey(idInfo.id))
                {
                    InstructionNode compl = idInfo.instr.getComplement();
                    if (compl != null)
                    {
                        SequentialProcess s = compl.getSequentialProcess();
                        Map cIdentifierTypes = (Map)identifierTypeMap.get(s);
                        String id = this.getId(compl, idInfo.index);
                        TagType tt = (TagType)cIdentifierTypes.get(id);
                        if (tt != null)
                            identifierTypes.put(idInfo.id, tt);
                    }
                }
            }
        }
        
        // types all "free" identifiers as possible nonce
        it = pn.seqProcIterator();
        while (it.hasNext())
        {
            SequentialProcess sp = (SequentialProcess)it.next();
            Set nonceSet = (Set)nonceSets.get(sp);
            Map identifierTypes = (Map)identifierTypeMap.get(sp);
            this.typeAsNonce(identifierTypes, nonceSet);
            
            // to allow cloned sp to access identifier types, 
            // identifierTypeMap is indexed with the sp string image
            identifierTypeMap.remove(sp);
            identifierTypeMap.put(sp.toString(), identifierTypes);
        }
        
        return identifierTypeMap;
    }
    
    private final String getId(InstructionNode instr, int index)
    {
        Ciphertext cd = (Ciphertext)instr.getInstruction().getData()[0];
        return cd.getContent()[index].getInfo();
    }
    
    /* scanSequentialProcess finds the tag type assignable to identifiers
     * appearing in the specified sequential process.
     */
    private void scanSequentialProcess(
        SequentialProcess sp, Map identifierTypes, Set nonceSet)
    {
        // gets identities and keys from sp declaration
        identifierTypes.put(sp.getSelf().getInfo(), TagType.IDENTITY);
        Iterator it = new ArrayIterator(sp.getDeclaration().getData());
        while (it.hasNext())
        {
            Data d = (Data)it.next();
            String id = d.getInfo();
            TagType t = d instanceof TaggedData? TagType.IDENTITY : TagType.KEY;
            identifierTypes.put(id, t);                 
        }
        
        // scan sp to type identifiers
        it = sp.instrIterator();
        while (it.hasNext())
        {
            InstructionNode ai = (InstructionNode)it.next();
            Instruction instr = ai.getInstruction();
            Data[] args = instr.getData();
            String name = instr.getName();
            if (NEW.equals(name))
                nonceSet.add(new IdInfo(args[0].getInfo(), ai, 0));
            else if (RUN.equals(name) || COMMIT.equals(name))
            {
                String id = args[args.length - 1].getInfo();
                TagType tt = (TagType)identifierTypes.get(id);
                if (tt == null || tt.equals(TagType.CIPHERTEXT))
                    identifierTypes.put(id, TagType.MESSAGE);
            }
            else if (ENCRYPT.equals(name) || DECRYPT.equals(name))
            {
                Ciphertext cd = (Ciphertext)args[0];
                String id = cd.getInfo();
                if (!TagType.MESSAGE.equals(identifierTypes.get(id)))
                    identifierTypes.put(id, TagType.CIPHERTEXT);
                identifierTypes.put(cd.getKey().getInfo(), TagType.KEY);
                
                this.fillNonceSet(nonceSet, cd.getContent(), ai);
            }
            else if (INPUT.equals(name))
                this.fillNonceSet(nonceSet, args, ai);
        }
    }
    
    private void fillNonceSet(Set nonceSet, Data[] ids, InstructionNode instr)
    {
        int size = ids.length;
        for (int i = 0; i < size; ++i)
        {
            Data arg = ids[i];
            if (arg.isSimple()) 
                nonceSet.add(new IdInfo(arg.getInfo(), instr, i));
        }
    }
    
    /* typeAsNonce assigns the NonceTag type to each identifier belonging to the
     * given set.
     */
    private void typeAsNonce(Map identifierTypes, Set nonceSet)
    {
        Iterator it = nonceSet.iterator();
        while (it.hasNext())
        {
            IdInfo idInfo = (IdInfo)it.next();
            String id = idInfo.id;
            if (!identifierTypes.containsKey(id))
                identifierTypes.put(id, TagType.NONCE);
        }
    }
    
    /** Overridden method: used for the nonce typing. If there are more than one
     * rules defined for the new primitive, the backtracking node gets the
     * TAGGING_NODE brand.
     */    
    protected void moreRules(Instruction instr)
    {
        boolean isNew = instr.getName().equals(NEW);
        this.save(isNew ? TAGGING_BRAND : BackTrackingHandler.NO_BRAND);
    }
    
    /* getTaggings return all the taggings found for the given instruction and
     * rule.
     */
    private Instruction[] getTaggings(
        InstructionNode ai, Rule rule, Map identifierTypes)
    {
        Instruction instr = ai.getInstruction();
        Collection taggings = new LinkedList();
        String name = instr.getName();
        if (TAGGABLE.contains(name))
        {
            Instruction ruleInstr = rule.getInstruction();

            Ciphertext rd = (Ciphertext)ruleInstr.getData()[0];
            Ciphertext cd = (Ciphertext)instr.getData()[0];
            if (rd.keyMatches(cd.getKey()) && ai.getComplement() != null)
            {
                Collection usedTags = rd.getUsedTags();
                Collection c = Arrays.asList(new Collection[]{ usedTags });

                // if rule instruction doesn't include tagged data (and is a
                // decryption) commit hints are used: first tagging is the
                // empty one.
                if (usedTags.size() == 0)
                {
                    if (usingCommitHints)
                    {
                        if (DECRYPT.equals(name))
                        {
                            c = commitHints.getHints(cd.getKeyType());
                            if (c == null || c.isEmpty()) return INSTR_ARRAY;

                            Map sm = super.patternMatch(ai, rule);
                            if (sm != null)
                            {
                                taggings.add(instr);
                                substMaps.put(instr, sm);
                            }
                        }
                    }
                    else return INSTR_ARRAY;
                }

                Iterator jt = c.iterator();
                while (jt.hasNext())
                {
                    usedTags = (Collection)jt.next();
                    cd = (Ciphertext)instr.getData()[0];
                    Data[] content = cd.getContent();
                    if (content.length < usedTags.size()) continue;

                    // for each component of the ciphertext
                    // computes the possible taggings (if any)
                    List computed = new LinkedList();
                    this.computeTaggings(computed, instr, usedTags);

                    Iterator it = computed.iterator();
                    while (it.hasNext())
                    {
                        Instruction tagging = (Instruction)it.next();
                        ai.setInstruction(tagging);
                        Map sm = super.patternMatch(ai, rule);
                        ai.setInstruction(instr);

                        if (sm != null)
                        {
                            taggings.add(tagging);
                            substMaps.put(tagging, sm);
                        }
                    }
                }
            }
        }
        
        return (Instruction[])taggings.toArray(INSTR_ARRAY);
    }

    private Map getContentTypes(Ciphertext cd, Map identifierTypes)
    {
        Data[] content = cd.getContent();
        int size = content.length;
        Map contentTypes = new HashMap(size);
        for (int i = 0; i < size; ++i)
        {
            Data d = content[i];
            if (d.isSimple())
            {
                String id = d.getInfo();
                TagType tt = (TagType)identifierTypes.get(id);
                Map typeIds = (Map)contentTypes.get(tt);
                if (typeIds == null)
                    contentTypes.put(tt, typeIds = new HashMap());
                typeIds.put(id, new Integer(i));
            }
        }
        return contentTypes;
    }
    
    /* computeTaggings computes the possible taggings of the instruction passed
     * as argument according to the set of used tags received.
     */
    private void computeTaggings(
        Collection taggings, Instruction instr, Collection usedTags)
    {
        Map identifierTypes = this.getIdentifierTypes(curSeqProc);
        instr = (Instruction)instr.clone();
        taggings.add(instr);
        Ciphertext cd = (Ciphertext)instr.getData()[0];
        Data[] content = cd.getContent();
        Map contentTypes = this.getContentTypes(cd, identifierTypes);
        
        int size = content.length;
        for (int i = 0; i < size; ++i)
        {
            Data d = content[i];
            if (d.isSimple())
            {
                // gets the type of the indentifier d
                String id = d.getInfo();
                TagType tt = (TagType)identifierTypes.get(id);
                if (tt == null) continue;

                // gets the tags of type tt
                Collection c = ((Collection)tagMap.get(tt));
                if (c == null) continue;
                    
                Iterator it = c.iterator();
                while (it.hasNext())
                {
                    // only used tags can be applied
                    String tag = (String)it.next();
                    if (!usedTags.remove(tag)) continue;
                    
                    Map ids = (Map)contentTypes.get(tt);
                    if (ids.size() > 1)
                    {
                        Iterator kt = taggings.iterator();
                        Collection tmp = new LinkedList();
                        while (kt.hasNext())
                        {
                            Instruction cloned = (Instruction)kt.next();
                            Instruction node = (Instruction)cloned.clone();
                            
                            Iterator jt = ids.keySet().iterator();
                            boolean hasNext = jt.hasNext();
                            while (hasNext)
                            {
                                String jd = (String)jt.next();
                                int j = ((Integer)ids.get(jd)).intValue();
                                cd = (Ciphertext)cloned.getData()[0];
                                cd.getContent()[j] = d.tag(jd, tag);
                                
                                if (hasNext = jt.hasNext())
                                {
                                    cloned = (Instruction)node.clone();
                                    tmp.add(cloned);
                                }
                                else break;
                            }
                        }
                        taggings.addAll(tmp);
                    }
                    else
                    {
                        Iterator jt = taggings.iterator();
                        while (jt.hasNext())
                        {
                            Instruction cloned = (Instruction)jt.next();
                            cd = (Ciphertext)cloned.getData()[0];
                            cd.getContent()[i] = d.tag(id, tag);
                        }
                    }
                    
                    // tags of the same type are mutually exclusive
                    break;
                }
            }
        }
        
        if (!usedTags.isEmpty()) taggings.clear();
    }

    /* tagComplement applies the specified tagging to its complement. */
    private boolean tagComplement(InstructionNode instr, Instruction tagging)
    {
        InstructionNode compl = instr.getComplement();
        
        // only "forward tagging" is allowed
        if (compl.isForwardComplement())
        {
            Instruction ci = (Instruction)compl.getInstruction().clone();
            compl.setInstruction(ci);
            
            Ciphertext cd = (Ciphertext)tagging.getData()[0];
            Data[] taggedContent = cd.getContent();
            cd = (Ciphertext)ci.getData()[0];
            Data[] complContent = cd.getContent();
            
            int size = taggedContent.length;
            for (int i = 0; i < size; ++i)
            {
                Data d = taggedContent[i];
                if (d instanceof TaggedData)
                {
                    String info = complContent[i].getInfo();
                    String tag = ((TaggedData)d).getTag();
                    complContent[i] = d.tag(info, tag);
                }
            }
            
            return true;
        }

        return false;
    }
    
    /* getIdentifierTypes returns the identifier types corresponding to the
     * specified sequential process.
     */
    private final Map getIdentifierTypes(final SequentialProcess sp)
    {
        if (identifierTypeMap == null)
        {
            Protocol pn = sp.getProtocol();
            identifierTypeMap = this.computeIdentifierTypes(pn);
        }

        return (Map)identifierTypeMap.get(sp.toString());
    }
 
    /* removeUselessNodes removes form the backtracking stack all the useless
     * nodes that might generate loops.
     */
    private boolean removeUselessNodes()
    {
        try
        {
            Integer instrItr = ValidationState.INSTR_ITR;
            Integer ruleItr = ValidationState.RULE_ITR;
            
            BackTrackingHandler.State previousState = btHandler.getCurState();
            this.backTrack();
            BackTrackingHandler.State initialState = btHandler.getCurState();
            
            BackTrackingHandler.State state = initialState;
            Object instrNode = state.getFieldValue(instrItr);
            Object ruleNode = state.getFieldValue(ruleItr);

            Stack stateStack = btHandler.getStateStack();
            int size = stateStack.size();
            int index = size;

            while (state.getFieldValue(TAGGING_ITR) == null)
            {
                if (index == 0 ||
                    !instrNode.equals(state.getFieldValue(instrItr)) ||
                    !ruleNode.equals(state.getFieldValue(ruleItr)))
                {
                    this.restoreBTHandler(previousState, initialState);
                    return false;
                }

                state = (BackTrackingHandler.State)stateStack.get(--index);
            }

            if (!instrNode.equals(state.getFieldValue(instrItr)) ||
                !ruleNode.equals(state.getFieldValue(ruleItr)))
            {
                this.restoreBTHandler(previousState, initialState);
                return false;
            }

            btHandler.backTrack(size - index);

            return true;
           
        }
        catch (BackTrackingHandler.LimitExcededException lee)
        {
            log.error(lee);
        }
        
        return false;
    }
    
    /* restoreBTHandler push onto the backtracking stack the states passed as
     * arguments.
     */
    private void restoreBTHandler(
        BackTrackingHandler.State previousState,
        BackTrackingHandler.State initialState)
        throws BackTrackingHandler.LimitExcededException
    {
        initialState.resetBackup();
        btHandler.setCurState(initialState);
        btHandler.save();
        btHandler.setCurState(previousState);
        this.updateState();
    }
    
    /* isTaggedDecryptionDefined checks if a tagged decryption is defined in the
     * current validation rule set.
     */
    private boolean isTaggedDecryptionDefined(EvaluationTree vr)
    {
        Iterator it = new ArrayIterator(vr.getMatchingRules(DECRYPT));
        while (it.hasNext())
        {
            Rule rule = (Rule)it.next();
            if (rule.getInstruction().containsTagged()) return true;            
        }
        return false;
    }

    private final void restoreProtocolOrder(Protocol protocol)
    {
        if (usingProtocolReordering && originalOrder != null)
            protocol.reorderSeqProcesses(originalOrder);
    }

    /* reorderProtocol computes the best sequential process order according to
     * the used heuristic.
     */
    private void reorderProtocol(Protocol protocol)
    {
        if (usingProtocolReordering)
        {
            ValidationRules vr = this.getValidationRules();
            int minTagNumber = vr.getMinTagNumber();
            int maxTagNumber = vr.getMaxTagNumber();
            
            ArrayList seqProcesses = new ArrayList(protocol.getChildCount());
            Iterator it = protocol.seqProcIterator();
            while (it.hasNext())
            {
                SequentialProcess sp = (SequentialProcess)it.next();
                this.computeStatistics(sp, minTagNumber, maxTagNumber);
                seqProcesses.add(sp);
            }
            seqProcesses.trimToSize();
            
            originalOrder =
                (SequentialProcess[])seqProcesses.toArray(SP_ARRAY);
            
            Comparator statcomp = protocol.getStatisticsComparator();
            SequentialProcess[] reordered =
                (SequentialProcess[])seqProcesses.toArray(SP_ARRAY);
            Arrays.sort(reordered, statcomp);

            if (reordered.length > 2)
            {
                this.evaluateComplementOrder(reordered, minTagNumber);
                Arrays.sort(reordered, statcomp);
            }
            
            protocol.reorderSeqProcesses(reordered);
        }
    }
    
    /* evaluateComplementOrder finds the more convenient order of the
     * sequential processes following the first according to the used heuristic.
     */
    private void evaluateComplementOrder(
        SequentialProcess[] reordered, int minTagNumber)
    {
        int encrypt;
        int pos = 0;
        int max = 0;
        int size = reordered.length;
        Set ids = new HashSet();
        
        Iterator it = new ArrayIterator(reordered);
        while (it.hasNext())
        {
            SequentialProcess sp = (SequentialProcess)it.next();
            
            int statistics = sp.getStatistics();
            encrypt = 0;
            
            if (++pos == 1)
            {
                max = statistics;
                statistics += COMMIT_VALUE;
            }

            Iterator jt = sp.instrIterator();
            while (jt.hasNext())
            {
                InstructionNode ai = (InstructionNode)jt.next();
                String id = ai.getInstrLabel();
                Instruction instr = ai.getInstruction();
                
                if (pos == 1) ids.add(id);
                else if (instr instanceof CipherInstruction)
                {
                    CipherInstruction enc = (CipherInstruction)instr;
                    int tags = enc.getCiphertext().getContent().length;
                    if (ai.getComplement() != null && enc.isEncrypt() &&
                        tags >= minTagNumber && !ids.contains(id))
                    {
                        encrypt += ENCRYPT_VALUE;
                        ids.add(id);
                    }
                }
            }
            
            if (encrypt > 0)
                statistics = max + size - pos + encrypt;
            
            sp.setStatistics(statistics);
        }
    }
    
    /* computeStatistics computes the number of points to be assigned to the 
     * specified  sequential process according to the used heuristic.
     */
    private int computeStatistics(
        SequentialProcess sp, int minTagNumber, int maxTagNumber)
    {
        int statistics = 0;
        
        boolean taggableDecrypt = false;
        boolean hasComplements = false;
        boolean hasCommit = false;
        InstructionNode compl = null;
        int distance = 0;
        
        Iterator it = sp.instrIterator();
        while (it.hasNext())
        {
            InstructionNode ai = (InstructionNode)it.next();
            Instruction instr = ai.getInstruction();
            String name = instr.getName();

            if (COMMIT.equals(name))
            {
                statistics += (hasCommit = taggableDecrypt) ? COMMIT_VALUE : 1;
                if (taggableDecrypt)
                {
                    distance = -1;
                    statistics -= DISTANCE_VALUE;
                }
            }

            else if (ENCRYPT.equals(name) || DECRYPT.equals(name))
            {
                Ciphertext ciphertext = (Ciphertext)instr.getData()[0];
                int argNumber = ciphertext.getContent().length;
                
                if (argNumber >= minTagNumber)
                {
                    statistics += TAGGABLE_VALUE;

                    if (argNumber > maxTagNumber)
                       statistics -= argNumber - maxTagNumber;
                    
                    if (name.startsWith(DECRYPT_INIT))
                    {
                        taggableDecrypt = true;
                        distance = -1;
                    }
                    
                    InstructionNode ci = ai.getComplement();
                    if (ci != null)
                    {
                        compl = ci;
                        hasComplements = true;
                    }
                }
            }
            
            ++distance;
        }
        
        if (hasComplements && hasCommit)
        {
            SequentialProcess csp = compl.getSequentialProcess();
            int compStat = csp.getStatistics();
            if (compStat < statistics) csp.setStatistics(statistics++);
        }
        
        int compStat = sp.getStatistics();
        statistics = compStat > statistics ? compStat : statistics;
        sp.setStatistics(statistics);

        return statistics;
    }

    
    // -------------------------------------------------------------- properties
    
    /** Getter for property useCommitHints.
     * @return Value of property useCommitHints.
     *
     */
    public boolean isUsingCommitHints() {
        return usingCommitHints;
    }
    
    /** Checks if the commit hints are necessary or optional.
     * @return Value of property useCommitHints.
     */
    public boolean isUsingCommitHintsDefault() {
        return usingCommitHintsDefault;
    }

    
    /** Setter for property useCommitHints.
     * @param useCommitHints New value of property useCommitHints.
     *
     */
    public void setUsingCommitHints(boolean usingCommitHints) {
        this.usingCommitHints = usingCommitHints;
    }
    
    /** Getter for property usingProtocolReorder.
     * @return Value of property usingProtocolReorder.
     *
     */
    public boolean isUsingProtocolReordering() {
        return this.usingProtocolReordering;
    }
    
    /** Setter for property usingProtocolReorder.
     * @param usingProtocolReorder New value of property usingProtocolReorder.
     *
     */
    public void setUsingProtocolReordering(boolean usingProtocolReordering) {
        this.usingProtocolReordering = usingProtocolReordering;
    }
    
    /** Getter for property equalLimit.
     * @return Value of property equalLimit.
     *
     */
    public int getEqualLimit() {
        return this.equalLimit;
    }
    
    /** Setter for property equalLimit.
     * @param equalLimit New value of property equalLimit.
     *
     */
    public void setEqualLimit(int equalLimit) {
        this.equalLimit = equalLimit;
    }
    
    /**
     * Setter for property fullInfo.
     * @param fullInfo New value of property fullInfo.
     */
    public void setFullInfo(boolean fullInfo) {}
}

/** This class implements a set of commit hints. */
final class CommitHints
{
    private static final String DEC = "dec";
    private static final List EMPTY_HINTS = Arrays.asList(new Object[]{});
    
    private Map commitHintsMap = new HashMap();
    private int minTagNumber = Integer.MAX_VALUE;
    
    /** Creates a new instance of Commit Hints */    
    CommitHints() {}
    
    /** Collects the commit hints from the specified validation rule set.
     * @param vr the validation rule set to be searched for commit hints.
     * @return the commit hints found
     */    
    static CommitHints collect(EvaluationTree vr)
    {
        CommitHints commitHints = new CommitHints();
        
        Rule[] rules = vr.getMatchingRules(InferenceCore.COMMIT);
        Iterator it = new ArrayIterator(rules);
        while (it.hasNext())
        {
            Rule rule = (Rule)it.next();
            Iterator jt = rule.getConditions().operationIterator();
            while (jt.hasNext())
            {
                Operation op = (Operation)jt.next();
                Data args = op.getArguments()[0];
                Data d = null;
                
                if (args instanceof Ciphertext)
                {
                    Ciphertext cd = (Ciphertext)args;
                    d = cd.getContent()[0];
                }
                else if (args instanceof Instruction) d = (Instruction)args;
                else continue;
                
                commitHints.setHints(d);
            }
        }
        
        return commitHints;
    }
    
    /** Returns a collection of used tags for the specified key type.
     * @param kt the requested key type
     * @return a collection of used tags
     */    
    Collection getHints(KeyType kt)
    {
        List hints = (LinkedList)commitHintsMap.get(kt);
        List res = EMPTY_HINTS;
        
        if (hints != null)
        {
            res = new ArrayList(hints.size());
            Iterator it = hints.iterator();
            while (it.hasNext())
            {
                Collection usedTags = (Collection)it.next();
                res.add(new HashSet(usedTags));
            }
        }
        
        return res;
    }
    
    /** Returns the minimum number of tags used in the validation rule set.
     * @return the minimum number of tags used in the validation rule set
     */    
    int getMinTagNumber()
    {
        return minTagNumber != Integer.MAX_VALUE ? minTagNumber : 0;
    }

    /* setHints collects the hints from the given data. */
    private void setHints(Data d)
    {
        if (d instanceof Instruction && d.getName().startsWith(DEC))
        {
            Instruction instr = (Instruction)d;
            Ciphertext cd = (Ciphertext)instr.getData()[0];
            KeyType kt = cd.getKeyType().getComplement();
            this.putHints(kt, cd.getUsedTags());
        }
    }
    
    /* putHints stores the hints found. */
    private void putHints(KeyType kt, Collection usedTags)
    {
        Collection hints = (Collection)commitHintsMap.get(kt);
        if (hints == null)
            commitHintsMap.put(kt, hints = new LinkedList());
        
        int size = usedTags.size();
        if (size > 0)
        {
            hints.add(usedTags);
            if (size < minTagNumber) minTagNumber = size;
        }
    }
    
    /** Returns a string image of the object.
     * @return a string image of the object
     */    
    public String toString() { return commitHintsMap.toString(); }
}
