/*
 * ValidationCore.java
 *
 * Created on 4 marzo 2004, 17.16
 */

package it.unive.dsi.pear.core;

import java.util.*;
import java.lang.reflect.*;
import java.io.*;

import it.unive.dsi.pear.util.*;
import it.unive.dsi.pear.core.evaluationtree.*;

import it.unive.dsi.pear.log.Log;
import it.unive.dsi.pear.core.parser.AbstractData;
import it.unive.dsi.pear.core.parser.ParseException;
import it.unive.dsi.pear.core.parser.SemanticException;
import it.unive.dsi.pear.core.parser.Container;
import it.unive.dsi.pear.core.parser.Data;
import it.unive.dsi.pear.core.parser.Instruction;
import it.unive.dsi.pear.core.parser.InstructionNode;
import it.unive.dsi.pear.core.parser.Operation;
import it.unive.dsi.pear.core.parser.Protocol;
import it.unive.dsi.pear.core.parser.ProtocolNode;
import it.unive.dsi.pear.core.parser.Rule;
import it.unive.dsi.pear.core.parser.SequentialProcess;
import it.unive.dsi.pear.core.parser.ValidationRules;

/** This class implements the validation algorithm.
 * @author Francesco Placella
 */
public class ValidationCore implements Evaluator, Parser, Validation
{
    private static ValidationRules DEF_RULES = null;
    private static String defRulesName;
    
    private static it.unive.dsi.pear.core.parser.Parser dataFactory = ValidationCore.getRuleParser(null);
    
    private static final String DEF_RULES_SRC_PROP = "defaultRulesSource";
//    private static final String DEF_RULES_SRC = "default.vr";
    private static final String DEF_RULES_SRC = "synt-analysis.vr";
    private static final String DEF_RULES_RESOURCE =
        Configuration.getDefSourceName(DEF_RULES_SRC);
    
    private static final long FIXED_WAIT_DURATION = 1;
    private static final int MAX_BT_NUMBER = 200;
    
    /** The stopped exception. */    
    public final StoppedException STOPPED = new StoppedException();
    
    /** The exception raised to perform a back step. */    
    protected class ResetStepException extends InterruptedException {}
    /** This exception is thrown during a back step performing. */    
    protected final ResetStepException RESET_STEP = new ResetStepException();
    
    /** The current logger. */    
    protected final Log log;
    
    // --------------------------------------------------------- private fields

    private final List treePath = new LinkedList();
    /** The backtracking handler. */    
    protected final BackTrackingHandler btHandler =
        new BackTrackingHandler(null, MAX_BT_NUMBER);
    /** The back step handler. */    
    protected final BackTrackingHandler backStepHandler =
        new BackTrackingHandler(null);
    
    private Data curIdentity;
    private BackTrackingHandler.State state;
    private EvalInfo evalInfo;
    private Set uniqueIds; 
    private ProtocolNode lastInstruction;
    private SequentialProcess lastSeqProc;
    private Rule lastRule;
    private Rule lastBTRule;
    private Rule rule;
    private CloneableIterator lastBTInstr;
    private Iterator instrIterator;
    private boolean stepPerformed;
    private boolean stepReceived;
    private boolean moreStep;
    private boolean proceed;
    private boolean backStep;
    private boolean valid;
    private Map symbolTable;
    private StepLevel curStepLevel = StepLevel.STRAIGHT_ON_LEVEL;
    private CloneableMap containerMap = new CloneableMap();
    private Configuration conf = Configuration.getCurrent();
    
    
    private class Decoration implements InstructionDecoration
    {
        private Rule rule;
        private Collection containers;
        private String ruleName;

        public Collection getContainers() { return this.containers; }
        public void setContainers(Collection containers) { this.containers = containers; }
        public Rule getRule() { return this.rule; }
        public void setRule(Rule rule) { this.rule = rule; }
        public String getRuleName() { return ruleName; }
        public void setRuleName(String ruleName) { this.ruleName = ruleName; }
        
        public String toString() { return ruleName; }
    }
    
    // ------------------------------------------------------------- properties
    
    /** Holds value of property validationRules. */
    private EvaluationTree validationRules = null;
    
    /** Holds value of property protocolName. */
    private Protocol protocol = null;
    
    /** Holds value of property stopped. */
    private boolean stopped = false;    
    
    /** Holds value of property errMsg. */
    private String errMsg;
    
    /** Holds value of property errPath. */
    private Object[] errPath;
    /**
     * Holds value of property fullInfo.
     */
    private boolean fullInfo  = false;

    
    // ----------------------------------------------------------- constructors
    
    /** Creates a new instance of ValidationCore.
     * @throws InitializationException Any exception thrown during the
     * initialization process is caught and set as cause of a new instance of
     * <CODE>InitializationException</CODE>.
     */
    public ValidationCore()
        throws InitializationException
    {
        this(Log.SIMPLE_LOG);
    }
    
    /** Creates a new instance of ValidationCore.
     * @param log An implementation of the logger.
     * @throws InitializationException Any exception thrown during the
     * initialization process is caught and set as cause of a new instance of
     * <CODE>InitializationException</CODE>.
     */    
    public ValidationCore(Log log)
        throws InitializationException
    {
        this(ValidationCore.loadDefaultRules(log), log);
    }
    
    /** Creates a new instance of ValidationCore.
     * @param rules the current validation rule set
     * @param initLog An implementation of the logger.
     * @throws InitializationException Any exception thrown during the
     * initialization process is caught and set as cause of a new instance of
     */    
    public ValidationCore(ValidationRules rules, Log initLog)
        throws InitializationException
    {
        log = initLog != null ? initLog : Log.SIMPLE_LOG;

        if (rules != null)
        {
            int logLevel = log.getLogLevel();
            if (logLevel >= Log.FULL_OUTPUT && rules != DEF_RULES)
            {
                log.forcedInfo("current validation rules dump:");
                log.out("\n" + rules.getTreeImage());
            }
        }
        else 
        {
            log.info("null rules: using default ones");
            ValidationCore.loadDefaultRules(log);
            rules = ValidationCore.DEF_RULES;
        }
        
        this.setValidationRules(rules);
    }
    
    // ----------------------------------------------------------- initializers
    
    /* loadDefaultRules loads the default validation rules from the default
     * source reader.
     */
    private static ValidationRules loadDefaultRules(Log log)
    {
        if (DEF_RULES != null) return DEF_RULES;
        
        try
        {
            Reader is = ValidationCore.getDefRulesReader(log);
            DEF_RULES = ValidationCore.parseRules(is, log);
        }
        catch (Throwable t)
        {
            log.error("error during default rules parsing", t);
            throw new ExceptionInInitializerError(t);
        }
        
        if (log.getLogLevel() >= Log.DEBUG)
        {
            log.forcedInfo("default validation rules dump:");
            log.out("\n" + DEF_RULES.getTreeImage());
        }
        
        return DEF_RULES;
    }
    
    /** Initializes the state of the validation core. */    
    protected void resetEnvironment()
    {
        this.setStopped(false);

        uniqueIds = null;        
        lastInstruction = null;
        lastRule = null;
        errPath = null;
        stepReceived = true;
        moreStep = true;
        stepPerformed = false;
        proceed = false;
        backStep = false;
        valid = false;
        treePath.clear();
        
        state = this.initBTHandler();
        containerMap = 
            (CloneableMap)state.getFieldValue(ValidationState.CONTAINER_MAP);

        evalInfo = new EvalInfo();
        evalInfo.setFullInfo(fullInfo);
        evalInfo.setBtHandler(btHandler);
        evalInfo.setEvaluator(this);
        evalInfo.setSubstMapField(ValidationState.SUBST_MAP);
    }
   
    
    // ------------------------------------------------------------------- main
    
    /** Runs the validation core in single-threaded mode.
     * @param args a file with the name equal to the value of the first argument
     * is looked for and, if found, opened.
     */
    public static void main(java.lang.String[] args)
    {
        Reader is;
        try
        {
            is = args.length > 0 ? new FileReader(args[0]) : new InputStreamReader(System.in);
            is = new BufferedReader(is);
                    
            ValidationCore vc = new ValidationCore();
            vc.validateProtocol(vc.parseProtocol(is));
        }
        catch (Exception e) { e.printStackTrace(); }
    }
    
    
    // --------------------------------------------------------- static methods
    
    /** Returns a reader for the default validation rule set.
     * @throws IOException if any I/O problem is detected
     * @return a reader for the default validation rule set
     */    
    public final Reader getDefRulesReader()
        throws IOException
    {
        return this.getDefRulesReader(log);
    }
        
    private static Reader getDefRulesReader(Log log)
        throws IOException
    {
        Reader stream = null;
        Configuration conf = Configuration.getCurrent();
        boolean jnlp = conf.isUsingJNLP();
        
        // if fileName == null the related system property is not defined 
        String fileName = System.getProperty(DEF_RULES_SRC_PROP, DEF_RULES_SRC);
        File file = new File(fileName);
        boolean fileExists = false;
        
        if (!jnlp && (fileExists = file.exists()))
        {
            if (file.canRead())
                stream = new FileReader(file);
            else
            {
                log.error("error while getting default rules: can't read from file");
                InputStream is = conf.getResourceAsStream(DEF_RULES_RESOURCE);
                if (is != null) stream = new InputStreamReader(is);
                fileName = DEF_RULES_SRC;
            }
        }
        else
        {
            String defRules = Configuration.getDefSourceName(fileName);
            InputStream is = conf.getResourceAsStream(defRules);
            if (is != null) stream = new InputStreamReader(is);
        }
        
        if (stream == null)
        {
            log.error("can't get specified default rules: " + fileName);
            InputStream is = conf.getResourceAsStream(DEF_RULES_RESOURCE);
            if (is == null)
                throw new IOException("can't get default rules input stream");
            stream = new InputStreamReader(is);
            fileName = DEF_RULES_SRC;
        }
        
        if (DEF_RULES == null)
        {
            String msg = fileName + " ";
            msg += fileExists ? "(source file)" : "(built-in)";
            log.info("default rules source: " + msg);
        }
        
        defRulesName = fileName;
        
        return new BufferedReader(stream);
    }    
    
    /** Parses the source whose <CODE>Reader</CODE> is passed as argument.
     * @param input a <CODE>Reader</CODE> for a validation rule source
     * @throws ParseException if any syntax error is detected
     * @throws IOException if any I/O problem is detected
     * @return the parse tree of the specified validation rule source
     */    
    public final ValidationRules parseRules(Reader input)
        throws ParseException, IOException
    {
        return this.parseRules(input, log);
    }
        
    private static ValidationRules parseRules(Reader input, Log log)
        throws ParseException, IOException
    {
        it.unive.dsi.pear.core.parser.Parser ruleParser = ValidationCore.getRuleParser(input);

        ValidationRules vr = null; 
        ruleParser.setLog(log);

        try { vr = (ValidationRules)ruleParser.getParseTree(); }
        catch (SemanticException se) { ruleParser.warning(se); }
        finally { input.close(); }

        return vr;
    }
    
    private static final it.unive.dsi.pear.core.parser.Parser getRuleParser(Reader input)
    {
        return new it.unive.dsi.pear.core.parser.validationrules.ValidationRules(input);
    }
    
    
    // --------------------------------------------------------- public methods
    
    /** Parses the protocol source whose <CODE>Reader</CODE> is passed as
     * argument.
     * @param input a <CODE>Reader</CODE> for the current protocol source
     * @throws ParseException if any syntax error is detected
     * @throws IOException if any I/O problem is detected
     * @return the parse tree of the specified protocol source
     */    
    public Protocol parseProtocol(final Reader input)
        throws ParseException, IOException
    {
        it.unive.dsi.pear.core.parser.Parser protocolParser = 
            new it.unive.dsi.pear.core.parser.rospi.RoSpi(input);
        
        Protocol pn = null;
        protocolParser.setLog(log);
        protocolParser.setDefinedTags(validationRules.getDefinedTags());
        
        try { pn = (Protocol)protocolParser.getParseTree(); }
        catch (SemanticException se) { protocolParser.warning(se); }
        finally { input.close(); }
        
        return pn;
    }
    
    /** Runs a validation session on the specified protocol using the current 
     * validation rule set.
     * @param protocol the protocol to be validated
     * @throws StoppedException if the validation session is stopped before
     * terminating
     * @throws LimitExcededException if the limit of backtracking actions is 
     * exceeded
     * @return <CODE>true</CODE> if the protocol is valid,
     * <CODE>false</CODE> otherwise.
     */
    public synchronized 
        boolean validateProtocol(final Protocol protocol)
        throws StoppedException,
               BackTrackingHandler.LimitExcededException
    {
        this.protocol = protocol;
        this.resetEnvironment();
        
        Integer spItField = ValidationState.SEQ_PROC_ITR;
        try
        {
            moreStep = true;

            log.info("[" + protocol.getProtocolName() + "] starting validation");
            if (log.getLogLevel() >= Log.FULL_OUTPUT)
            {
                log.forcedInfo("protocol dump:" );
                log.out("\n" + protocol.getTreeImage());
            }

            boolean valid = true;
            Iterator it = (Iterator)
                state.getFieldValue(spItField, protocol.seqProcIterator());
            
            while (it.hasNext())
            {
                state.backup(spItField);
                SequentialProcess sp = (SequentialProcess)it.next();
                
                try
                {
                    valid = this.validateSeqProcess(sp);

                    if (valid)
                    {
                        this.waitForNextStep(StepLevel.SEQ_PROC_LEVEL);
                        state.setFieldValue(spItField, (CloneableIterator)it);
                    }
                    else if (btHandler.isAvailable())
                    {
                        String name = lastInstruction.getInstruction().getName();
                        log.info("[" + name + "] backtracking...", Log.VERBOSE);
                        this.backTrack();
                    } 
                    else break;
                }
                catch (InterruptedException ie)
                {
                    if (ie instanceof ResetStepException)
                    {
                        this.backTrack(backStepHandler);
                        btHandler.setCurState(state);
                        int iLevel = StepLevel.INSTRUCTION_LEVEL.getLevel();
                        if (curStepLevel.getLevel() <= iLevel)
                            curStepLevel = StepLevel.INSTRUCTION_LEVEL;
                    }
                    else 
                        throw STOPPED;
                }
                
                it = (Iterator)state.getFieldValue(spItField);
            }

            log.info(
                "[" + protocol.getProtocolName() +
                "] protocol local validation " +
                (valid ? "successful" : "failed"));
                
            this.valid = valid;
        }
        finally
        {
            stepPerformed = true;
            moreStep = false;
            state.setFieldValue(spItField, null);
            backStepHandler.clear();
            this.notifyAll();
        }

        return valid;
    }
    
    /** Validates the specified sequential process.
     * @param sp the sequential process to be validated
     * @throws InterruptedException if the validation session is interrupted
     * before terminating
     * @throws LimitExcededException if the limit of backtracking actions is
     * exceeded
     * @return <CODE>true</CODE> if the sequential process is valid,
     * <CODE>false</CODE> otherwise.
     */
    public boolean validateSeqProcess(final SequentialProcess sp)
        throws BackTrackingHandler.LimitExcededException, InterruptedException
    {
        evalInfo.setSymbolTable(symbolTable = sp.getSymbolTable());
        Instruction decl = sp.getDeclaration();
        String spName = decl.getName();
        curIdentity = sp.getSelf();
        uniqueIds = sp.getUniqueIds();

        if (!sp.equals(lastSeqProc))
        {
            lastSeqProc = sp;
            log.info(
                "[" + spName + "] validating sequential process: " + decl,
                Log.VERBOSE);            
        }

        boolean valid = true;
        Integer instrField = ValidationState.INSTR_ITR;
        CloneableIterator it =
            (CloneableIterator)state.getFieldValue(instrField);

        if (it == null)
        {
            this.clearContainersContent();
            valid = this.validateDeclBindings(sp);
            state.setFieldValue(instrField, it = sp.instrIterator());
        }

        while (valid && it.hasNext())
        {
            instrIterator = (Iterator)it.clone();
            state.backup(instrField);
            
            lastInstruction = (ProtocolNode)it.next();
            valid = this.validateInstruction((InstructionNode)lastInstruction);
            
            if (valid) state.setFieldValue(instrField, (CloneableIterator)it);
            else if (btHandler.isAvailable()) return false;
        }

        log.info(
            "[" + spName + "] sequential process validation " +
            (valid ? "successful" : "failed"), Log.VERBOSE);
        
        List ruleList = this.getRuleList();
        log.info("[" + spName + "] successfully applied rules: " + ruleList);
        if (ruleList != null) ruleList.clear();

        this.resetBTHandler();
        state.setFieldValue(instrField, null);
        
        return valid;
    }
    
    /** Validates the declarations found in the signature of the specified 
     * sequential process.
     * @param sp the sequential process whose signature has to be 
     * validated
     * @throws InterruptedException if the validation session is interrupted 
     * before terminating
     * @return <CODE>true</CODE> if the signature is valid,
     * <CODE>false</CODE> otherwise.
     */
    public boolean validateDeclBindings(final SequentialProcess sp)
        throws InterruptedException
    {
        boolean valid = true;
        lastInstruction = sp;
        
        Iterator di = sp.declIterator();
        while (di.hasNext())
        {
            Data arg = (Data)di.next();
            Instruction instr = sp.getDeclaration(arg);
            
            log.info("validating declaration: " + instr, Log.VERBOSE);
            
            Rule[] matchingRules =
                validationRules.getMatchingRules(instr.getName());
            Iterator it = new ArrayIterator(matchingRules);
            
            // if rules for the current instruction are defined, they are tried
            while (it.hasNext())
            {
                Rule rule = (Rule)it.next();
                Map substMap = rule.getInstruction().appliesTo(instr);
                
                // if the current rule is appliable...
                if (substMap != null)
                {
                    evalInfo.setSubstMap(substMap);
                    this.evalInfo.setSubstMap(substMap);
                    evalInfo.setBtHandler(new BackTrackingHandler(state));
                    valid = this.evaluateRule(rule, evalInfo);
                    evalInfo.setBtHandler(btHandler);
                    if (valid) break;
                }
                else
                {
                    this.setErrMsg(instr.toString());
                    valid = false;
                }
            }
            
            log.info(
                "declaration validation " + (valid ? "successful" : "failed"),
                Log.VERBOSE);
        }
        
        return valid;
    }
    
    /** Validates the specified instruction.
     * @param instruction the instruction to be validated
     * @throws InterruptedException if the validation session is interrupted 
     * before terminating
     * @throws LimitExcededException if the limit of backtracking actions is 
     * exceeded
     * @return <CODE>true</CODE> if the instruction is valid,
     * <CODE>false</CODE> otherwise.
     */
    public boolean validateInstruction(final InstructionNode instruction)
        throws BackTrackingHandler.LimitExcededException, InterruptedException
    {
        Instruction instr = instruction.getInstruction();
        log.info("-> validating instruction: " + instr, Log.VERBOSE);
        treePath.clear();
        this.waitForNextStep(StepLevel.INSTRUCTION_LEVEL);
        
        int triedRules = 0;
        boolean valid = false;
        
        Integer ruleField = ValidationState.RULE_ITR;
        String name = instr.getName();
        Rule[] matchingRules = validationRules.getMatchingRules(name);
        Iterator it = new ArrayIterator(matchingRules);
        it = (Iterator)state.getFieldValue(ruleField, (CloneableIterator)it);
        
        // if rules for the current instruction are defined, they are tried
        rule = null;
        while (it.hasNext())
        {
            state.backup(ruleField);
            state.backup(ValidationState.CONTAINER_MAP);
            
            rule = (Rule)it.next();
            String ruleId = rule.getIdentity();
            Map substMap = this.patternMatch(instruction, rule);
            
            // if the current rule is appliable...
            if (substMap != null)
            {
                ++triedRules;
                int stateCount = btHandler.availableCount();
                
                // nella mappa locale vengono riversati tutti i mapping globali
                log.debug("current identity: " + ruleId + "=" + curIdentity);

                evalInfo.setSubstMap(substMap);
                substMap.put(Data.UNIQUE_IDS, uniqueIds);
                
                if (valid = this.evaluateRule(rule, evalInfo))
                    this.decorateInstruction(rule);

                // checks if there are backtracking nodes available for
                // the current rule
                boolean b1 = stateCount == btHandler.availableCount();
                boolean b2 = this.returningBackTrackingNode();
                boolean unavailable = b1 && !b2;
                
                // it shouldn't happen
                if (!b1 && !b2) throw new RuntimeException("incoherent validation status!");
                
                if (unavailable)
                {
                    evalInfo.setBacktrackingEntity(null);
                    state.setBacktrackingEntity(null);
                }
                else if (!valid)
                {
                    log.info(
                        "[" + rule.getName() +"] backtracking...", Log.VERBOSE);

                    this.backTrack();
                    it = (Iterator)state.getFieldValue(ruleField);
                    continue;
                }
                
                if (it.hasNext())
                {
                    if (!valid) continue;
                    if (unavailable)
                    {
                        state.backup(ruleField);
                        this.moreRules(instr);
                    }
                }
                break;
            }
            else
            {
                this.setErrMsg(instr.toString());
                log.debug("failed matching of: " + rule.getName());
            }
            
            if (substMap != null) substMap.clear();
            evalInfo.setBacktrackingEntity(null);
        }
        
        if (valid && rule != null)
        {
            List ruleList = this.getRuleList();
            if (ruleList != null) ruleList.add(rule.getName());
        }

        if (triedRules == 0) this.noRuleApplied(instruction);

        log.info(
            "instruction validation " + (valid ? "successful" : "failed"),
            Log.VERBOSE);

        evalInfo.setBacktrackingEntity(null);
        state.setFieldValue(ruleField, null);
        state.retrieveFieldValue(ValidationState.SUBST_MAP);
        return valid;
    }
    
    protected void decorateInstruction(Rule rule)
    {
        Decoration d = new Decoration();
        d.setRuleName(rule.getName());
                    
        if (fullInfo)
        {
            ETRule r = (ETRule)rule;
            Map m = (Map)containerMap.clone();
            d.setContainers(m.values());
            Rule rr = (Rule)r.getLastInstantiation();
            d.setRule(rr);
        }

        lastInstruction.setDecoration(d);
    }

    
    /** It's invoked if the instruction is valid, but there are other rules 
     * defined for it.
     * @param instr the instruction being validated
     */    
    protected void moreRules(Instruction instr) { this.save(); }

    
    // ---------------------------------------------- Metodi d'interazione
    
    /** Collects all the informations needed to display an image of the current
     * validation state.
     * @param vo a data structure used to store the collected data
     * @throws StoppedException if the validation session is stopped during the
     * collecting process
     * @return <CODE>true</CODE> if there are more steps,
     * <CODE>false</CODE> otherwise.
     */    
    public synchronized boolean getStep(ValidationOutput vo)
        throws StoppedException
    {
        try
        {
            while (!stepPerformed) this.waitOn();

            stepReceived = true;
            this.notifyAll();

            HashMap sm = (HashMap)evalInfo.getSubstMap();
            if (sm != null)
            {
                sm = (HashMap)sm.clone();
                sm.remove(Data.UNIQUE_IDS);
            }
            vo.setSubstMap(sm);
            vo.setContainers(this.containersContentImage(false));
            vo.setTreePath(treePath.toArray());
            vo.setInstruction(lastInstruction);
            vo.setRule(lastRule);
            vo.setProtocol(protocol);
            if (!moreStep) vo.setValid(valid);

            stepPerformed = false;
        }
        finally { vo.setMessage(log.getLastMsg()); }
        
        return moreStep;
    }
    
    /** Tells the validation core to proceed with the validation session. */    
    public synchronized void proceed()
    {
        proceed = true;
        this.notifyAll();
    }
    
    /** Checks if there are back steps available.
     * @return <CODE>true</CODE> if there are back steps available,
     * <CODE>false</CODE> otherwise.
     */    
    public final boolean isBackStepAvailable()
    {
        return backStepHandler.isAvailable();
    }
    
    /** Performs a back step. */    
    public synchronized void stepBack()
    {
        if (this.isBackStepAvailable())
        {
            proceed = backStep = true;
            this.notifyAll();
        }
    }

    /** Waits for the permission to proceed to the next step.
     * @param stepLevel the step level associated with the current point of the 
     * process
     * @throws InterruptedException if the validation session is interrupted
     */    
    public synchronized void waitForNextStep(StepLevel stepLevel)
        throws InterruptedException
    {
        boolean waited = false;
        this.waitOn(FIXED_WAIT_DURATION);
        
        if (curStepLevel != StepLevel.STRAIGHT_ON_LEVEL)
            while (!stepReceived) this.waitOn();
        else
            if (this.isStopped()) throw STOPPED;

        stepReceived = false;
        stepPerformed = true;
        this.notifyAll();
        
        int level = stepLevel.getLevel();
        if (curStepLevel != StepLevel.STRAIGHT_ON_LEVEL &&
            level <= curStepLevel.getLevel())
            while (!proceed) waited = this.waitOn();
        else
            if (this.isStopped()) throw STOPPED;
        
        if (curStepLevel != StepLevel.STRAIGHT_ON_LEVEL)
            this.handleBackStep(stepLevel);
        
        proceed = false;
    }
    
    /** If the validation core is proceeding saves the data needed to perform a
     * back step, otherwise starts the performing of a back step by throwing a
     * <CODE>ResetStepException</CODE>.
     * @param stepLevel the step level associated with the current point of the 
     * process
     * @throws ResetStepException if a back step has been requested
     */    
    protected void handleBackStep(StepLevel stepLevel)
        throws ResetStepException
    {
        if (backStep) 
        {
            backStep = proceed = false;
            throw RESET_STEP;
        }
        else if (stepLevel == StepLevel.INSTRUCTION_LEVEL)
        {
            BackTrackingHandler.State state = btHandler.getCurState();
            state = state.cloneState();
            Integer contField = ValidationState.CONTAINER_MAP;
            state.setFieldValue(contField, this.state.getFieldValue(contField));
            state.backup(contField);
            backStepHandler.setCurState(state);
            backStepHandler.save();
        }
    }

    private final boolean waitOn()
        throws StoppedException
    {
        return this.waitOn(0);
    }
    
    /* waitOn waits for the specified time if it is not interrupted before. */
    private synchronized boolean waitOn(long timeout)
        throws StoppedException
    {
        try { this.wait(timeout); } catch (InterruptedException ie) {}        
        if (this.isStopped()) throw STOPPED;
        return true;
    }
    
    
    // -------------------------------------------------------- Metodi ausliari
    
    /* containersContentImage returns the list of the containers' current
     * string representation; if requested it clears the containers.
     */
    private List containersContentImage(boolean clear)
    {
        List containerOutput = new LinkedList();

        Map cm = (Map)state.getFieldValue(ValidationState.CONTAINER_MAP);
        Iterator it = cm.values().iterator();
        while (it.hasNext())
        {
            Container c = (Container)it.next();
            containerOutput.add(c.getName() + ": " + c.getContentImage());
            if (clear) c.clear();
        }
        
        return containerOutput;
    }
   
    private void clearContainersContent()
    {
        Iterator it = containerMap.values().iterator();
        while (it.hasNext()) ((Container)it.next()).clear();
    }
    
    /* initContainerMap instatiates the containers declared in the apposite
     * section of the validation rule set definition.
     */
    private void initContainerMap(ValidationRules validationRules)
        throws InitializationException
    {
        containerMap.clear();
        Map containerMap = validationRules.getContainerMap();
        
        try
        {
            Iterator entries = containerMap.entrySet().iterator();
            while (entries.hasNext())
            {
                Map.Entry entry = (Map.Entry)entries.next();
                String type = (String)entry.getValue();
                
                Class containerClass = conf.getPluginClass(type);
                Container container = (Container)containerClass.newInstance();

                String name = (String)entry.getKey();
                container.setName(name);
                this.containerMap.put(name, container);
            }
        }
        catch (Exception e)
        {
            log.error("error during container map init", e);
            throw new InitializationException(e);
        }
    }
    
    /* evaluateRule evaluates the rule passed as argument using the data stored
     * in the data structure received as argument.
     */
    private boolean evaluateRule(Rule rule, EvalInfo evalInfo)
        throws InterruptedException
    {
        lastRule = null;
        treePath.clear();
        treePath.add(validationRules);
        treePath.add(rule);
        evalInfo.setTreePath(treePath);
        
        log.info("applying rule: " + rule.getName(), Log.VERBOSE);
        Set substset = evalInfo.getSubstMap().entrySet();
        log.debug("pattern matching result: ", substset);
        this.waitForNextStep(StepLevel.RULE_LEVEL);

        // if there are conditions available, they are evaluated
        boolean valid = false;
        Evaluable conditions = (Evaluable)rule.getConditions();
        if (conditions != null)
        {
            evalInfo.setAutoSubstEnabled(true);
            try { valid = conditions.evaluate(evalInfo); }
            catch (Container.OperationException oe)
            {
                Throwable cause = oe.getCause();
                if (cause instanceof InterruptedException)
                    throw (InterruptedException)cause;

                log.error("error during condition evaluation: " + cause);
            }
            
            log.info("rule conditions: evaluation " +
                (valid ? "successful" : "failed"), Log.VERBOSE);
        }
        else valid = true;
        
        // if the evaluation of defined conditions is successful and
        // there are actions available, they are performed
        Evaluable actions = (Evaluable)rule.getActions();
        if (valid && actions != null)
        {
            treePath.clear();
            treePath.add(validationRules);
            treePath.add(rule);
            
            boolean actRes = false;
            evalInfo.setAutoSubstEnabled(false);
            try { actRes = actions.evaluate(evalInfo); }
            catch (Container.OperationException oe)
            {
                Throwable cause = oe.getCause();
                if (cause instanceof InterruptedException)
                    throw (InterruptedException)cause;
                
                log.error("error during actions performing: " + cause);
                valid = false;
            }

            log.info("rule actions performing result: " + actRes,
            Log.VERBOSE);
        }

        if (valid)
        {
            lastRule = rule;
            this.setErrPath(null);
        }
        
        return valid;
    }
    
    /** Performs a backtrack step.
     * @throws LimitExcededException if the limit of backtracking actions is
     * exceeded
     */    
    public final void backTrack()
        throws BackTrackingHandler.LimitExcededException
        { this.backTrack(btHandler); }

    /** Performs a backtrack step on the backtracking handler specified as
     * argument.
     * @param btHandler the backtracking handler which has to perform the 
     * backtrack step
     * @throws LimitExcededException if the limit of backtracking actions is
     * exceeded
     */    
    protected void backTrack(BackTrackingHandler btHandler)
        throws BackTrackingHandler.LimitExcededException
    {
        state = btHandler.backTrack();
        Object entity = state.getBacktrackingEntity();
        evalInfo.setBacktrackingEntity((Operation)entity);

        lastBTRule = (Rule)
            state.getFieldValue(ValidationState.BT_RULE);
        lastBTInstr = (CloneableIterator)
            state.getFieldValue(ValidationState.BT_INSTRUCTION);
        containerMap = (CloneableMap)
            state.getFieldValue(ValidationState.CONTAINER_MAP);
    }
    
    /** Saves the current state. */    
    protected final void save()
    {
        this.save(BackTrackingHandler.NO_BRAND, btHandler);
    }
    /** Saves the current state through the specified backtracking handler.
     * @param btHandler the backtracking handler used to save the state
     */    
    protected final void save(BackTrackingHandler btHandler)
    {
        this.save(BackTrackingHandler.NO_BRAND, btHandler);
    }
    /** Saves the current state with the specified brand.
     * @param brand the brand to be associated to the state
     */    
    protected final void save(int brand)
    {
        this.save(brand, btHandler);
    }
    /** Saves the current state with the specifed brand through the backtracking
     * handler passed as argument.
     * @param brand the brand to be associated to the state
     * @param btHandler the backtracking handler used to save the state
     */    
    protected void save(int brand, BackTrackingHandler btHandler)
    {
        btHandler.save(brand);
        if (btHandler == this.btHandler)
            log.info("backtrack node here", Log.VERBOSE);        

        if (instrIterator != null)
        {
            CloneableIterator ci = (CloneableIterator)instrIterator;
            lastBTInstr = (CloneableIterator)ci.clone();
            state.setFieldValue(ValidationState.BT_INSTRUCTION, lastBTInstr);
        }

        if (rule != null)
        {
            lastBTRule = rule;
            state.setFieldValue(ValidationState.BT_RULE, lastBTRule);
        }
    }
    
    /** Performs the pattern-matching between the instruction and the rule 
     * passed as arguments.
     * @param instr the instruction to be matched
     * @param rule the rule to be checked for matching
     * @return the substitution map if pattern-matching is successful,
     * <CODE>null</CODE> otherwise
     */    
    protected Map patternMatch(InstructionNode instr, Rule rule)
    {
        String ruleId = rule.getIdentity();
        Instruction ruleInstr = rule.getInstruction();
        return ruleInstr.appliesTo(instr.getInstruction(), ruleId, curIdentity);
    }
    
    /** Initializes the backtracking handler.
     * @return the initial state
     */    
    protected BackTrackingHandler.State initBTHandler()
    {
        btHandler.setStartingState(state = new ValidationState());
        backStepHandler.setStartingState(state);

        this.clearContainersContent();
        state.setFieldValue(ValidationState.CONTAINER_MAP, containerMap);
        state.setFieldValue(ValidationState.RULE_LIST, new CloneableList());

        return state;
    }
    
    /** Clears the state of the backtracking handler. */    
    protected void resetBTHandler() { btHandler.clear(); }
    
    /** It's invoked whenever no rule can be applied to the specified 
     * instruction.
     * @param instruction the instruction unmatchable
     */
    protected void noRuleApplied(InstructionNode instruction)
    {
        this.setErrPath(null);
        log.info("no rule could be applied", Log.VERBOSE);
    }
    
    /** The current state is updated with the value obtained from the 
     * backtracking handler.
     */    
    protected final void updateState() { state = btHandler.getCurState(); }
    
    /** Checks if the current instruction and rule match the last ones creating
     * a backtracking node.
     * @return <CODE>true</CODE> if the check is succeessful,
     * <CODE>false</CODE> otherwise.
     */    
    protected final boolean returningBackTrackingNode()
    {
        return rule.equals(lastBTRule) && instrIterator.equals(lastBTInstr);
    }
    
    
    // -------------------------------------------------- getter/setter methods
    
    /** Getter for property validationRules.
     * @return Value of property validationRules.
     *
     */
    public ValidationRules getValidationRules() {
        return this.validationRules;
    }
    
    /** Getter for property protocolName.
     * @return Value of property protocolName.
     *
     */
    public final String getProtocolName() {
        return protocol != null ? protocol.getProtocolName() : null;
    }
    
    /** Getter for property ruleList.
     * @return Value of property ruleList.
     *
     */
    public List getRuleList() {
        return (List)state.getFieldValue(ValidationState.RULE_LIST);
    }
    
    /** Getter for property curStepLevel.
     * @return Value of property curStepLevel.
     *
     */
    public StepLevel getCurStepLevel() {
        return curStepLevel;
    }
    
    /** Setter for property curStepLevel.
     * @param curStepLevel New value of property curStepLevel.
     *
     */
    public void setCurStepLevel(StepLevel curStepLevel) {
        this.curStepLevel = curStepLevel;
    }
    
    /** Getter for property stopped.
     * @return Value of property stopped.
     *
     */
    public boolean isStopped() {
        return this.stopped;
    }
    
    /** Setter for property stopped.
     * @param stopped New value of property stopped.
     *
     */
    public synchronized void setStopped(boolean stopped)
    {
        this.stopped = stopped;
        if (stopped)
        {
            this.notifyAll();
            log.info("[" + protocol.getProtocolName() + "] validation stopped");
        }
    }
    
    /** Getter for property errorMessage.
     * @return Value of property errorMessage.
     *
     */
    public String getErrMsg() {
        return this.errMsg;
    }
    
    /** Setter for property errorMessage.
     * @param errMsg
     */
    private void setErrMsg(String errMsg) {
        this.errMsg = errMsg;
    }
    
    /** Getter for property errPath.
     * @return Value of property errPath.
     *
     */
    public Object[] getErrPath() {
        return this.errPath;
    }
    
    /** Setter for property errPath.
     * @param errPath New value of property errPath.
     *
     */
    private void setErrPath(Object[] errPath) {
        this.errPath = errPath;
    }
    
    /** Setter for property validationRules.
     * @param vr the validation rules to be set
     * @throws InitializationException Any exception thrown during the
     * setting process is caught and set as cause of a new instance of
     * <CODE>InitializationException</CODE>.
     */
    public void setValidationRules(ValidationRules vr)
        throws InitializationException
    {
        if (!(vr instanceof EvaluationTree))
            vr = new ETValidationRules(vr);
        
        validationRules = (EvaluationTree)vr;
        
        if (validationRules != null)
            this.initContainerMap(validationRules);
    }
    
    /** Getter for property protocol.
     * @return Value of property protocol.
     *
     */
    public Protocol getProtocol() {
        return this.protocol;
    }
    
    /** Setter for property protocol.
     * @param protocol New value of property protocol.
     *
     */
    protected void setProtocol(Protocol protocol) {
        this.protocol = protocol;
    }
    
    /** Getter for property defRulesName.
     * @return Value of property defRulesName.
     *
     */
    public String getDefRulesName() { return defRulesName; }

    /**
     * Setter for property fullInfo.
     * @param fullInfo New value of property fullInfo.
     */
    public void setFullInfo(boolean fullInfo) {
        this.fullInfo = fullInfo;
    }

    
    
    // -------------------------------------------------------------------------

    /** Performs the operation passed as argument.
     * @param evalInfo a data structure containing all the data needed to 
     * perform the operation requested
     * @throws OperationException if any problem is detected during the 
     * operation execution
     * @return the boolean value returned by the operation if any,
     * <CODE>true</CODE> otherwise.
     */    
    public boolean evaluate(EvalInfo evalInfo)
        throws Container.OperationException
    {
        evalInfo.setBacktrackingEntity(null);
        Operation op = evalInfo.getOperation();

        try
        {
            Map substMap = evalInfo.getSubstMap();
            String name = op.getContainerName();
            Container container = (Container)containerMap.get(name);

            if (container == null)
            {
                container = (Container)substMap.get(name);
                name = container.getName();
                if (container == null)
                    throw new Container.OperationException("can't recognize native container");
            }

            Container[] args = this.getSubstArgs(substMap, container, op);
            if (evalInfo.isFullInfo()) evalInfo.setInstantiatedArguments(this.toData(args));
            
            Object res = this.performOperation(op, container, args);

            String opMsg = op.toString(name, args);
            log.info("performed: " + opMsg, Log.FULL_OUTPUT);
            this.setErrMsg(opMsg);
            this.setErrPath(evalInfo.getTreePath().toArray());

            if (log.getLogLevel() == Log.DEBUG)
                log.debug("operation result: " + res + log.newline() +
                          "container content: " + container.getContentImage());

            this.waitForNextStep(StepLevel.OPERATION_LEVEL);

            if (res != null && res instanceof Boolean) 
                return ((Boolean)res).booleanValue();
            else
                return true;
        }
        catch (RuntimeException rte) { throw rte; }
        catch (InvocationTargetException ite)
        {
            throw new Container.OperationException(ite.getCause());
        }
        catch (Exception e)
        {
            throw new Container.OperationException(e);
        }
    }

    private Class[] getArgClasses(int size)
    {
        Class[] argClasses = new Class[size];
        Arrays.fill(argClasses, Data.class);
        return argClasses;
    }

    /* getSubstArgs return the array of arguments of the operation passed as
     * argument substituted with the values present in the substitution map and
     * the ones computed by autoSubst.
     */
    private Container[] getSubstArgs(Map substMap, Container container, Operation op)
    {
        Data[] arguments = op.getArguments();
        int size = arguments.length;
        List args = new LinkedList();

        for (int i = 0; i < size; ++i)
        {
            Data ruleArg = arguments[i];
            if (ruleArg instanceof AbstractData)
            {
                ((AbstractData)ruleArg).subst(substMap, args);
                continue;
            }

            Container c;
            String argInfo = ruleArg.getInfo();            
            if ((c = (Container)containerMap.get(argInfo)) != null) args.add(c);
            else
            {
                Data arg = (Data)ruleArg.clone();
                if (evalInfo.isAutoSubstEnabled())
                {
                    String absInfo = AbstractData.ABSTRACT_INFO;
                    List abstractVal = null;
                    if (evalInfo.isResetAbstract())
                        abstractVal = (List)substMap.remove(absInfo);
                    arg.subst(substMap, false);
                    arg = this.autoSubst(arg, substMap, container, op);
                    if (abstractVal != null) substMap.put(absInfo, abstractVal);
                }
                else arg.subst(substMap);

                substMap.remove(AbstractData.ABSTRACT_DATA);
                args.add(arg);
            }
        }

        state.setFieldValue(ValidationState.PMATCH_ITR_MAP, null);
        return (Container[])args.toArray(new Container[args.size()]);
    }

    /* autoSubst tries to perform pattern-matching between the argument and all
     * the elements belonging to the specified container; if a successful
     * matching is found substitution, is performed on the fly and a 
     * backtracking node is created.
     */
    private Data autoSubst(Data arg, Map sm, Container container, Operation op)
    {
        Data argClone = (Data)arg.clone();

        Integer field = ValidationState.PMATCH_ITR_MAP;
        Map autoSubstMap =
            (Map)state.getFieldValue(field, new CloneableMap(true));

        CloneableIterator it = (CloneableIterator)autoSubstMap.get(arg);
        if (it == null) it = container.infoIterator();

        Object mapField = evalInfo.getSubstMapField();
        Map map = (Map)state.retrieveFieldValue(mapField);
        if (map != null && map != sm) sm.putAll(map);

        while (it.hasNext())
        {
            Data data = (Data)it.next();
            CloneableMap cm = (CloneableMap)sm;
            Map substMap = (Map)cm.clone();
            boolean autosubst = container.autoPatternMatch(arg, data, substMap);
            if (autosubst && substMap.size() != sm.size())
            {
                log.debug("auto pattern matching (backtrack node here): ",
                          substMap.entrySet());

                state.setFieldValue(mapField, (CloneableMap)sm);
                state.setBacktrackingEntity(op);
                autoSubstMap.put(argClone, it.clone());
                this.save(evalInfo.getNodeBrand());
                state.setBacktrackingEntity(null);

                sm.putAll(substMap);
                return arg;
            }
        }

        return argClone;
    }

    /* performOperation performs the specified operation checking if it is
     * native or external.
     */
    private Object performOperation(
        Operation operation, Container container, Container[] arguments)
        throws NoSuchMethodException, ClassNotFoundException,
               IllegalAccessException, InstantiationException,
               InvocationTargetException
    {
        Method op = null;
        Object res = null;
        String operationName = operation.getOperationName();
            
        if (operation.isNative())
        {
            Class[] argClasses = this.getArgClasses(arguments.length);
            op = container.getClass().getMethod(operationName, argClasses);
            res = op.invoke(container, arguments);
        }
        else
        {
            Class c = conf.getPluginClass(operationName);
            Class[] argClasses = ExternalOperation.ARG_CLASSES;
            op = c.getMethod(ExternalOperation.METHOD_NAME, argClasses);
            ExternalOperation opInstance = (ExternalOperation)c.newInstance();
            opInstance.setSymbolTable(evalInfo.getSymbolTable());
            res = op.invoke(opInstance, new Object[]{ container, arguments });
        }
        
        log.debug("invoked method: ", op);
        return res;
    }
    
    private Data[] toData(Container[] c)
    {
        Data[] d = new Data[c.length];
        
        for (int i = 0; i < c.length; ++i)
        {
            Container ci = c[i];
            if (!(ci instanceof Data)) 
                ci = dataFactory.toData(ci.toString());
            d[i] = (Data)ci;
        }

        return d;
    }
}