/*
 * DefaultLog.java
 *
 * Created on 5 marzo 2004, 16.46
 */

package it.unive.dsi.pear.log;

import java.util.*;
import java.text.*;
import java.io.*;

/** This class is complete implementatioin of the interface <CODE>Log</CODE>.
 * @author Francesco Placella
 */
public class DefaultLog implements Log
{
    private static final boolean DEFAULT_CLASS_HEADER =
        Boolean.valueOf(System.getProperty("classHeader", "false")).booleanValue();
    
    private static final boolean DEFAULT_TIME_HEADER =
        Boolean.valueOf(System.getProperty("timeHeader", "true")).booleanValue();
    
    private static final boolean DEFAULT_WARNINGS_ENABLED =
        Boolean.valueOf(System.getProperty("enableWarnings", "true")).booleanValue();

    private static final String NL[];
    private static final DateFormat df;
    private static final Message[] WARN_ARRAY = {};
    
    static
    {
        df = DateFormat.getDateTimeInstance(DateFormat.SHORT, DateFormat.MEDIUM);

        String prefix = "[" + DefaultLog.DEBUG_HEADER + "] ";
        String date = df.format(new Date(System.currentTimeMillis())) + ": ";
        
        int size = prefix.length();
        int[] sizes = new int[]{ size, size + date.length() };
        
        int sSize = sizes.length;
        NL = new String[sSize];
        for (int j = 0; j < sSize; ++j)
        {
            size = sizes[j];
            StringBuffer buf = new StringBuffer(size + 1);
            buf.append("\n");
            
            for (int i = 0; i < size; ++i) buf.append(" ");
            
            NL[j] = buf.toString();
        }
    }
    
    
    private static final String ERROR_HEADER = "ERR";
    private static final String WARNING_HEADER = "WRN";
    private static final String INFO_HEADER = "INF";
    private static final String DEBUG_HEADER = "DBG";
    
    private static final String STARTING_MSG = "created new log with level: ";
    private static final String CHANGING_LEVEL_MSG = "changing log level to: ";

    private static final int DEFAULT_LINE_LENGTH = 256;
    
    private final PrintStream err;
    private final PrintStream out;
    private final Class loggedClass;
    private final LinkedList warnings = new LinkedList();
    
    private int logLevel = DEFAULT_LOG_LEVEL;
    private boolean timeHeader = DEFAULT_TIME_HEADER;
    private boolean classHeader = DEFAULT_CLASS_HEADER;
    private boolean warningsEnabled = DEFAULT_WARNINGS_ENABLED;
    
    private String lastMsg;
    
    /** Creates a new instance of Log. */
    public DefaultLog()
    {
        this(null, null, null, DEFAULT_TIME_HEADER);
    }
    
    /** Creates a new instance of Log.
     * @param loggedClass the class creating the logger
     */    
    public DefaultLog(Class loggedClass)
    {
        this(loggedClass, null, null, DEFAULT_TIME_HEADER);
    }
    
    /** Creates a new instance of Log.
     * @param loggedClass the class creating the logger
     * @param err the error stream
     * @param out the output stream
     * @param timeHeader if <CODE>true</CODE> the time will be appended to each 
     * print
     */    
    public DefaultLog(
        Class loggedClass, PrintStream err, PrintStream out, boolean timeHeader)
    {
        this.err = err;
        this.out = out;
        this.loggedClass = loggedClass;
        this.timeHeader = timeHeader;
        this.classHeader = loggedClass == null ? false : DEFAULT_CLASS_HEADER;

        if (logLevel > Log.WARNINGS)
        {
            String header = loggedClass != null ? loggedClass + " " : "";
            this.getOut().println(this.getMsg(INFO_HEADER, header + STARTING_MSG + logLevel));
        }
    }
    
    protected final PrintStream getErr() { return err != null ? err : System.err; }
    protected final PrintStream getOut() { return out != null ? out : System.out; }

    /** Prints the data if the current log level is greater or equal to NORMAL.
     * @param msg the data to be printed
     */    
    public final void info(Object msg) { this.info(msg, Log.NORMAL); }    
    /** Prints the data if the current log level is greater or equal to specifed
     * one.
     * @param msg the data to be printed
     * @param logLevel the log level assigned to the data
     */    
    public void info(Object msg, int logLevel)
    {
        // mute has no output!
        if (logLevel <= Log.MUTE) logLevel = Log.MUTE + 1;        

        if (this.logLevel >= logLevel)
        {
            lastMsg = msg.toString();
            this.getOut().println(this.getMsg(INFO_HEADER, lastMsg));
        }
    }

    
    /** Treats the data as informations ignoring the current log level.
     * @param msg the data to be printed
     */    
    public final void forcedInfo(Object msg)
    {
        lastMsg = msg.toString();
        this.getOut().println(this.getMsg(INFO_HEADER, lastMsg));
    }
    
    
    /** Prints the given <CODE>Throwable</CODE> object with its stack trace.
     * @param ex the throwable object to be printed
     */    
    public final void error(Throwable ex) { this.error(null, ex); }
    /** Prints the given object treating it as an error.
     * @param msg the object to be printed
     */    
    public final void error(Object msg) { this.error(msg, null); }
    /** Prints the given message and then appends the specified
     * <CODE>Throwable</CODE> object and its stack trace.
     * @param msg the error message
     * @param ex the throwable object
     */    
    public void error(Object msg, Throwable ex)
    {
        String exMsg = ex != null ? ex.getMessage() : "";
        if (exMsg == null) exMsg = ex.getClass().getName();
        
        String errMsg = msg != null ?
            msg + (ex != null ? " (" + exMsg + ")" : "") : exMsg;
        
        if (logLevel > Log.MUTE)
        {
            this.getOut().println(this.getMsg(ERROR_HEADER, lastMsg = errMsg));
            this.getErr().println(this.getMsg(ERROR_HEADER, lastMsg = errMsg));
        }
        
        // stack trace is printed anyway if present
        if (ex != null)
        {
            this.getErr().print(this.getMsg(ERROR_HEADER, ""));
            ex.printStackTrace(this.getErr());
        }
    }
    

    /** Treats the given message as a warning.
     * @param warning the warning to be printed
     */    
    public void warning(Message warning)
    {
        if (warningsEnabled && logLevel >= Log.WARNINGS)
        {
            warnings.add(warning);
            lastMsg = warning.toString();
            this.getOut().println(this.getMsg(WARNING_HEADER, lastMsg));
        }
    }
    
    /** Checks if there are warning recorded in the related history.
     * @return <CODE>true</CODE> if there are warnings,
     * <CODE>false</CODE> otherwise.
     */    
    public boolean hasWarnings() { return warnings.size() > 0; }
    /** Returns an array containing the set of warnings recorded in the related
     * history.
     * @param clear if <CODE>true</CODE> history is cleared after creating the
     * resulting array.
     * @return an array containing the set of warnings recorded in the related
     * history
     */    
    public Message[] getWarnings(boolean clear)
    {
        Message[] res = (Message[])warnings.toArray(WARN_ARRAY);
        if (clear) warnings.clear();
        return res;
    }

    /** Treats the given message as debug data, it's printed only if the current
     * log level is DEBUG.
     * @param msg the message to be printed
     */    
    public final void debug(Object msg) { this.debug(null, msg); }
    
    /** Treats the given message as debug data, it's printed only if the current
     * log level is DEBUG ad it's appended the string image of the given object.
     * @param header the header of the message
     * @param msg the message to be printed
     */    
    public void debug(String header, Object msg)
    {
        if (this.logLevel >= Log.DEBUG)
        {
            StringBuffer sb = new StringBuffer();
            if (header != null) sb.append(header);
            sb.append(msg);
            this.getOut().println(this.getMsg(DEBUG_HEADER, sb));
        }
    }
    
    /** Prints the given messag with no header, it's a way to access the output
     * stream provided to the logger at creation time.
     * @param msg the message to be printed
     */    
    public final void out(Object msg) { this.getOut().println(msg); }

    /** Returns a new line string with appended the apropriate indentation according
     * to che current header length.
     * @return a newline string
     */    
    public final String newline() { return this.timeHeader ? NL[1] : NL[0]; }
    
    /** Constructs the output line of the logger.
     * @param prefix the header string
     * @param msg the message to be printed
     * @return the string to be printed
     */    
    protected String getMsg(final String prefix, final Object msg)
    {
        StringBuffer message = new StringBuffer(DEFAULT_LINE_LENGTH);
        
        if (timeHeader)
        {
            message.
                append(df.format(new Date(System.currentTimeMillis()))).
                append(": ");
        }

        message.
            append("[").
            append(prefix).
            append("] ").
            append(msg);
        
        if (classHeader)
        {
            message.
                append(" (").
                append(loggedClass.getName()).
                append(")");
        }
                
        return message.toString();
    }
    
    
    /** Getter for property logLevel.
     * @return Value of property logLevel.
     *
     */
    public int getLogLevel() { return logLevel; }
    
    /** Setter for property logLevel.
     * @param logLevel New value of property logLevel.
     *
     */
    public void setLogLevel(int logLevel)
    {
        if (logLevel != this.logLevel &&
            logLevel >= Log.MUTE && logLevel <= Log.DEBUG)
        {
            this.logLevel = logLevel;
            this.getOut().println(this.getMsg(INFO_HEADER, CHANGING_LEVEL_MSG + logLevel));            
        }
    }
    
    /** Getter for property longHeader.
     * @return Value of property longHeader.
     *
     */
    public boolean isTimeHeader() {
        return timeHeader;
    }
    
    /** Setter for property longHeader.
     * @param longHeader New value of property longHeader.
     *
     */
    public void setTimeHeader(boolean longHeader) {
        this.timeHeader = timeHeader;
    }
    
    /** Getter for property warningsEnabled.
     * @return Value of property warningsEnabled.
     *
     */
    public boolean isWarningsEnabled() {
        return warningsEnabled;
    }
    
    /** Setter for property warningsEnabled.
     * @param warningsEnabled New value of property warningsEnabled.
     *
     */
    public void setWarningsEnabled(boolean warningsEnabled) {
        this.warningsEnabled = warningsEnabled;
    }
    
    /** Getter for property lastMsg.
     * @return Value of property lastMsg.
     *
     */
    public java.lang.String getLastMsg() {
        return lastMsg;
    }
}
