/*
 * Log.java
 *
 * Created on 11 maggio 2004, 16.00
 */

package it.unive.dsi.pear.log;

/** This interface models a logger.
 * @author Francesco Placella
 */
public interface Log
{
    public interface Message
    {
        String getMessage();
        String getSource();
        int getLine();
        int getColumn();
    }
    
    public static final Log VOID_LOG = new VoidLog();
    public static final Log SIMPLE_LOG = new SimpleLog();    
    
    /** The default log level is set through the related system property if
     * defined, otherwise it's NORMAL.
     */
    public static final int DEFAULT_LOG_LEVEL =
        Integer.parseInt(System.getProperty("logLevel", String.valueOf(Log.NORMAL)));
    
    /** The logger should print nothing. */    
    public static final int MUTE = -1;
    /** The logger should print only the warnings. */    
    public static final int WARNINGS = 0;
    /** The logger should print only warnings and normal-level informations. */    
    public static final int NORMAL = 1;    
    /** The logger should print only warnings and up to verbose-level 
      * informations. */
    public static final int VERBOSE = 2;
    /** The logger should print only warnings and all the available 
     * informations. */
    public static final int FULL_OUTPUT = 3;
    /** The logger should print everything. */    
    public static final int DEBUG = 4;

    public void info(Object msg);
    public void info(Object msg, int logLevel);
    public void forcedInfo(Object msg);
    
    public void warning(Message warning);
    public boolean hasWarnings();
    public Message[] getWarnings(boolean clear);
    
    public void error(Throwable ex);
    public void error(Object msg);
    public void error(Object msg, Throwable ex);
    
    public void debug(Object msg);
    public void debug(String header, Object msg);
    
    public void out(Object msg);
    public String newline();
    
    // properties    
    
    /** property logLevel.
     *
     */
    public int getLogLevel();
    public void setLogLevel(int logLevel);

    /** property timeHeader.
     *
     */
    public boolean isTimeHeader();
    public void setTimeHeader(boolean longHeader);
    
    /** property warningsEnabled.
     *
     */
    public boolean isWarningsEnabled();
    public void setWarningsEnabled(boolean warningsEnabled);
    
    /** property lastMsg.
     *
     */
    public String getLastMsg();
}

/** This class implements a void log: it prints nothing. */
class VoidLog implements Log
{
    public void info(Object msg) {}
    public void info(Object msg, int logLevel) {}
    public void forcedInfo(Object msg) {}
    public void warning(Message warn) {}
    public void error(Throwable ex) { ex.printStackTrace(System.err); }
    public void error(Object msg) { System.err.println(msg); }
    public void error(Object msg, Throwable ex)
        { System.err.println(msg); ex.printStackTrace(System.err); }
    public void debug(Object msg) {}
    public void debug(String s, Object msg) {}
    public void out(Object msg) {}
    public String newline() { return "\n\t"; }
    public int getLogLevel() { return Log.DEFAULT_LOG_LEVEL; }
    public boolean isTimeHeader() { return false; }
    public boolean isWarningsEnabled() { return false; }
    public String getLastMsg() { return ""; }
    public Message[] getWarnings(boolean clear) { return null; }
    public boolean hasWarnings() { return false; }
    public void setLogLevel(int logLevel) {}
    public void setTimeHeader(boolean longHeader) {}
    public void setWarningsEnabled(boolean warningsEnabled) {}
}

/** This class implements a simple log: it prints everything. */
final class SimpleLog extends VoidLog
{
    public void info(Object msg) { System.out.println(msg); }
    public void info(Object msg, int logLevel) { System.out.println(msg); }
    public void forcedInfo(Object msg) { System.out.println(msg); }
    public void warning(Message warn) { System.out.println(warn); }
    public void debug(Object msg) { System.out.println(msg); }
    public void debug(String s, Object msg) { System.out.println(msg); }
    public void out(Object msg) { System.out.println(msg); }
}