/*
 * CompositeData.java
 *
 * Created on 20 febbraio 2004, 18.09
 */

package it.unive.dsi.pear.core.parser;

import java.util.*;

/** This is a straightforward implementation of the <CODE>Ciphertext</CODE>
 * interface: it also may model cleartext so it allows no key to be associated
 * to the message.
 * @author Francesco Placella
 */
public class CompositeData extends SimpleData
    implements Ciphertext
{
    protected static final int DEFAULT_BUF_SIZE = 30;
    
    private int abstractIndex = ABSTRACT_INDEX;
    private TaggedContent taggedContent = null;
    
    /** Holds value of property encrypted. */
    private boolean encrypted = false;
    
    /** Holds value of property key. */
    private Data key = null;
    
    /** Holds value of property content. */
    private Data[] content = Data.EMPTY_DATA;
    
    /** Holds value of property asymmetric. */
    private boolean asymmetric = false;
    
    /** Creates a new instance of <CODE>CompositeData</CODE>.
     * @param taggedContent the tag model to be followed by this message
     */
    public CompositeData(TaggedContent taggedContent)
        { this(null, taggedContent); }
    /** Creates a new instance of <CODE>CompositeData</CODE>.
     * @param info the name of the message
     * @param taggedContent the tag model to be followed by this message
     */    
    public CompositeData(String info, TaggedContent taggedContent)
    {
        super(info);
        this.taggedContent = taggedContent;
    }
 
    public String toString()
    {
        String sup = super.toString();
        
        StringBuffer res = new StringBuffer(DEFAULT_BUF_SIZE);
        if (sup != null) res.append(sup + " = ");
        res.append("{");
        if (this.isAsymmetric()) res.append("|");
        
        int size = content.length;        
        if (size > 0)
        {
            res.append(content[0]);
            for (int i = 1; i < size; ++i) res.append(", " + content[i]);
        }
        
        if (this.isAsymmetric()) res.append("|");
        res.append("}");
        if (key != null)
        {
            if (encrypted) res.append("_" + key);
            else res.append("(_" + key + ")");
        }
        
        return res.toString();
    }

    public boolean isAbstract()
    {
        return this.abstractIndex != ABSTRACT_INDEX;
    }
    
    public boolean putMatching(Map substMap, List iList)
    {
        Data[] rData = this.getContent();
        int rSize = rData.length;        

        ListIterator it = iList.listIterator();
        while (it.hasNext())
        {
            Object obj = it.next();
            if (obj instanceof CompositeData)
            {
                CompositeData cd = (CompositeData)obj;
                if (this.encrypted == cd.encrypted)
                {
                    List iData = new ArrayList(Arrays.asList(cd.getContent()));
                    for (int i = 0; i < rSize; ++i)
                        if (i != this.abstractIndex &&
                            !rData[i].putMatching(substMap, iData))
                            return false;

                    Data abstractData;
                    if (this.isAbstract() && 
                        (abstractData = rData[this.abstractIndex]) != null &&
                        !abstractData.putMatching(substMap, iData))
                        return false;

                    if (this.key != null && !this.keyMatches(cd.key, substMap))
                        return false;
                    
                    List cdList = Arrays.asList(new Data[]{ cd });
                    super.putMatching(substMap, cdList);
                    
                    it.set(null);
                    return true;
                }
                return false;
            }
        }

        return true;
    }
    
    public final boolean keyMatches(Data sd)
    {
        return this.getKeyType() == sd.getKeyType();
    }
    /** Checks if the key associated with the message matches the given one.
     * @param sd the key to be matched
     * @param substMap the substitution map is populated with the appropriate mappings
     * @return <CODE>true</CODE> if the matching is successful,
     * <CODE>false</CODE> otherwise.
     */    
    public boolean keyMatches(Data sd, Map substMap)
    {
        List keyList = Arrays.asList(new Data[]{ sd });
        return this.getKey().putMatching(substMap, keyList);
    }
    
    /** Getter for property encrypted.
     * @return Value of property encrypted.
     *
     */
    public boolean isEncrypted() {
        return this.encrypted;
    }    
    
    /** Setter for property encrypted.
     * @param encrypted New value of property encrypted.
     *
     */
    public void setEncrypted(boolean encrypted) {
        this.encrypted = encrypted;
    }    
    
    /** Getter for property key.
     * @return Value of property key.
     *
     */
    public Data getKey() {
        return this.key;
    }
    
    /** Setter for property key.
     * @param key New value of property key.
     *
     */
    public void setKey(String key) { this.key = new SimpleData(key); }
    public void setKey(Data key)
    {
        this.key = key;
        asymmetric = key instanceof AsymKeyData.KeyInstance;
    }
    
    
    /** Getter for property content.
     * @return Value of property content.
     *
     */
    public Data[] getContent() { return this.content; }
    
    /** Setter for property content.
     * @param content New value of property content.
     *
     */
    public void setContent(Data[] content)
    {
        int size = content.length;
        
        this.abstractIndex = ABSTRACT_INDEX;
        boolean tc = taggedContent != null;
        if (tc) taggedContent.reset();
        
        for (int i = 0; i < size; ++i)
        {
            Data data = content[i];

            if (AbstractData.ABSTRACT_DATA.equals(data) ||
                AbstractTaggedData.ABSTRACT_TAGGED_DATA.equals(data))
                if (this.abstractIndex == ABSTRACT_INDEX)
                    this.abstractIndex = i;
                else
                    throw new SemanticError(Errors.DUPLICATE_ABSTRACT_DATA);
            
            if (tc && data instanceof TaggedData)
            {
                String tag = ((TaggedData)content[i]).getTag();
                if (!this.taggedContent.matchTag(tag))
                    throw new SemanticError(Errors.getMsg(Errors.FORBIDDEN_TAG, tag));
            }
        }

        this.content = content != null ? content : Data.EMPTY_DATA;
    }

    public boolean contains(Object obj)
    {
        if (!(obj instanceof Data)) return false;
        
        int size = this.content.length;
        for (int i = 0; i < size; ++i)
            if (this.content[i].equals(obj)) return true;
        
        return false;
    }
    
    public final Object[] toArray() { return this.getContent(); }
    
    public int hashCode()
    {
        int hashCode = this.rootHashCode();
        if (key != null) hashCode += key.hashCode(); 
        return hashCode;
    }
    
    public boolean equals(Object obj)
    {
        if (!(obj instanceof CompositeData)) return false;
        
        CompositeData cd = (CompositeData)obj;
        if (this.getInfo() != null && !super.equals(obj)) return false;
        if (this.encrypted != cd.encrypted) return false;
        if (this.key != null && cd.key != null &&
            !this.key.equals((Object)cd.key)) return false;
        if (this.content.length != cd.content.length) return false;
        
        Set content = new HashSet(Arrays.asList(this.content));
        Set cdContent = new HashSet(Arrays.asList(cd.content));
        
        return content.equals(cdContent);
    }
    
    public void subst(Map substMap, boolean removeAbstract)
    {
        super.subst(substMap, removeAbstract);
        
        List contentList = new LinkedList();
        int size = content.length;
        boolean ra = false;
        for (int i = 0; i < size; ++i)
        {
            Data data = content[i];
            if (data instanceof AbstractData)
            {
                ra = ((AbstractData)data).subst(substMap, contentList);
                abstractIndex = ABSTRACT_INDEX;
            }
            else
            {
                data.subst(substMap, removeAbstract);
                contentList.add(data);
            }
        }
        
        if (contentList.size() >= size || (removeAbstract && ra))
        {
            this.content = (Data[])contentList.toArray(Data.EMPTY_DATA);
            substMap.remove(AbstractData.ABSTRACT_DATA);
        }
        
        Data key = this.key;
        if (key != null) key.subst(substMap, removeAbstract);
    }
    
    
    public Object clone()
    {
        CompositeData cd = (CompositeData)super.clone();
        
        Data[] content = cd.getContent();
        cd.content = (Data[])content.clone();

        int size = content.length;
        for (int i = 0; i < size; ++i)
        {
            Data data = content[i];
            if (data != null) cd.content[i] = (Data)data.clone();
        }

        if (cd.key != null) cd.key = (SimpleData)cd.key.clone();       
        
        return cd;
    }
    
    public boolean containsTagged()
    {
        Data[] content = this.getContent();
        int size = content.length;
        for (int i = 0; i < size; ++i)
        {
            Data data = content[i];
            if (data.containsTagged()) return true;
        }
        return false;
    }
    
    public int getTagNumber()
    {
        int tagNumber = 0;
        
        Data[] content = this.getContent();
        int size = content.length;
        for (int i = 0; i < size; ++i)
        {
            Data data = content[i];
            if (data instanceof AbstractTaggedData) return 0;
            tagNumber += data.getTagNumber();            
        }
        
        return tagNumber;
    }
    
    public Collection getUsedTags()
    {
        Data[] content = this.getContent();
        int size = content.length;
        Set usedTags = new HashSet(size);
        for (int i = 0; i < size; ++i)
        {
            Data d = content[i];
            if (d instanceof TaggedData)
                usedTags.add(((TaggedData)d).getTag());
        }
        return usedTags;
    }
    
    public KeyType getKeyType() { return this.getKey().getKeyType(); }
    
    /** Getter for property asymmetric.
     * @return Value of property asymmetric.
     *
     */
    public boolean isAsymmetric() {
        return this.asymmetric;
    }
}
