/*
 * DefaultParser.java
 *
 * Created on 1 febbraio 2004, 16.55
 */

package it.unive.dsi.pear.core.parser;

import it.unive.dsi.pear.log.Log;

/** This class is a straightforward implementation of <CODE>Parser</CODE>.
 * @author Francesco Placella
 */
public abstract class DefaultParser implements Parser
{
    /** The related symbol table. */    
    protected final SymbolTable symbolTable;
    /** An implementation of the logger. */    
    protected Log log = Log.SIMPLE_LOG;
    /** The last token that generated a warning. */    
    protected GenericToken warnToken;
    
    /** Holds value of property definedTags. */
    private String[][] definedTags;
    
    /** Returns a new instance of the symbol table. */    
    protected abstract SymbolTable getSymbolTableInstance();
    /** Returns the current token. */    
    protected abstract GenericToken getCurrentToken();
    public abstract javax.swing.tree.TreeNode getParseTree()
        throws ParseException;
    
    /** Creates a new instance of <CODE>DefaultParser</CODE>. */    
    public DefaultParser()
    {
        this.symbolTable = this.getSymbolTableInstance();
    }

    
    // ------------------------------------------------------ Semantic warnings
    
    /** This class is an implementation of <CODE>Log.Message</CODE> used that 
     * models a warning.
     */    
    private class Warning implements Log.Message
    {
        private final int[] coords;
        private final String message;
        private final String source;
        
        /** Creates a new warning from the related 
         * <CODE>SemanticException</CODE>. */
        private Warning(SemanticException se)
        {
            GenericToken token = se.getToken();
            if (token == null)
            {
                token = warnToken != null ? warnToken : getCurrentToken();
                String image = se.getImage();
                source = image != null ? image : token.toString();
            }
            else source = token.toString();
            coords = new int[] { token.getBeginLine(), token.getBeginColumn() };
            message = se.getMessage();
            warnToken = null;
        }
        
        public int getLine() { return coords[0]; }
        public int getColumn() { return coords[1]; }
        public String getMessage() { return message; }
        public String getSource() { return source; }

        public String toString()
        {
            return message + " at " + coords[0] + ":" + coords[1];
        }
    }
    

    public final void warning(final SemanticException se)
    {
        log.warning(new Warning(se));
    }
    
    /* the methods defined below are wrappers for the corresponding symbol
     * table methods: they handle the semantic exceptions raised by the 
     * symbol table converting them into warnings.
     */
    
    public String checkDeclaration(
        final SymbolTable symbolTable, final String symbol)
    {
        try { return symbolTable.checkDeclaration(symbol); }
        catch (SemanticException se) { this.warning(se); }
        return null;
    }

    protected final String checkType(
        final String symbol, final EnumType expectedType)
    {
        this.checkType(this.symbolTable, symbol, new EnumType[]{ expectedType });
        return symbol;
    }

    protected final Data checkType(
        final Data symbol, final EnumType expectedType)
    {
        try { symbolTable.checkType(symbol, new EnumType[]{ expectedType }); }
        catch (SemanticException se) { this.warning(se); }
        return symbol;
    }
    
    protected final String checkType(
        final SymbolTable symbolTable,
        final String symbol,
        final EnumType expectedType)
    {
        this.checkType(symbolTable, symbol, new EnumType[]{ expectedType });
        return symbol;
    }
    
    protected final String checkType(
        final String symbol, final EnumType[] expectedType)

    {
        this.checkType(this.symbolTable, symbol, expectedType);
        return symbol;
    }
    
    protected final EnumType checkType(
        final SymbolTable symbolTable,
        final String symbol,
        final EnumType[] expectedType)
    {
        try { return this.symbolTable.checkType(symbol, expectedType); }
        catch (SemanticException se) { this.warning(se); }
        return null;
    }
    
    protected final String checkDeclaration(final String symbol)
    {
        try { return this.symbolTable.checkDeclaration(symbol); }
        catch (SemanticException se) { this.warning(se); }
        return null;
    }
    
    protected final void throwException(final SemanticException se)
    {
        this.warning(se);
    }

    protected final void put(
        final GenericToken token, final EnumType expectedType)
    {
        try { this.symbolTable.put(token, expectedType); }
        catch (SemanticException se) { this.warning(se); }
    }
    
    protected final void put(final String symbol, final EnumType expectedType)
    {
        try { this.symbolTable.put(new SimpleData(symbol), expectedType); }
        catch (SemanticException se) { this.warning(se); }
    }

    protected final void put(final Data symbol, final EnumType expectedType)
    {
        try { this.symbolTable.put(symbol, expectedType); }
        catch (SemanticException se) { this.warning(se); }
    }
    
    protected final EnumType putListElement(
        final Data symbol, final EnumType symbolType)
    {
        try { return this.symbolTable.putListElement(symbol, symbolType); }
        catch (SemanticException se) { this.warning(se); }
        return null;
    }

    protected final EnumType putListElement(
        final GenericToken token, final EnumType symbolType)
    {
        try { return this.symbolTable.putListElement(token, symbolType); }
        catch (SemanticException se) { this.warning(se); }
        return null;
    }
    
    // ------------------------------------------------------- properties
    
    /** Getter for property taggedContent.
     * @return Value of property taggedContent.
     *
     */
    public TaggedContent getTaggedContent() {
        return new TaggedContent(this.definedTags);
    }
    
    /** Setter for property taggedContent.
     * @param definedTags
     */
    public void setDefinedTags(String[][] definedTags) {
        this.definedTags = definedTags;
    }
    
    /** Getter for property definedTags.
     * @return Value of property definedTags.
     *
     */
    public String[][] getDefinedTags() {
        return this.definedTags;
    }
    
    /** Setter for property log.
     * @param log New value of property log.
     *
     */
    public void setLog(it.unive.dsi.pear.log.Log log) {
        this.log = log;
    }
    
    /** Getter for property log.
     * @return Value of property log.
     *
     */
    public it.unive.dsi.pear.log.Log getLog() {
        return log;
    }

    
    /** Getter for property symbolTable.
     * @return Value of property symbolTable.
     *
     */
    public SymbolTable getSymbolTable() {
        return symbolTable;
    }
}
