/*
 * DefaultSymbolTable.java
 *
 * Created on 18 gennaio 2004, 20.27
 */

package it.unive.dsi.pear.core.parser;

import java.util.*;

/** This class is a straightforward implementation of <CODE>SymbolTable</CODE>.
 * @author Francesco Placella
 */
public class DefaultSymbolTable extends HashMap
    implements SymbolTable
{
    private final EnumType anyType;
    private final EnumType[] types;
    private final Class enumTypeClass;
    private List symbolList = new LinkedList();    
    
    /** Creates a new instance of the symbol table.
     * @param types the types accepted by the symbol table
     */    
    public DefaultSymbolTable(final EnumType[] types) { this(types, null); }
    
    /** Creates a new instance of the symbol table.
     * @param types the types accepted by the symbol table
     * @param anyType the generic "any" type
     */    
    public DefaultSymbolTable(final EnumType[] types, EnumType anyType)
    {
        super();
        
        int length = types.length;
        this.types = new EnumType[length];
        this.anyType = anyType;

        if (length > 0)
        {
            this.enumTypeClass = types[0].getClass();
            
            System.arraycopy(types, 0, this.types, 0, length);
            Arrays.sort(this.types);
        }
        else this.enumTypeClass = null;
    }

    public final EnumType putListElement(
        final GenericToken token, final EnumType symbolType)
        throws SemanticException
    {
        try
        {
            String tokimg = token.toString();
            return this.putListElement(new SimpleData(tokimg), symbolType);
        }
        catch (SemanticException se)
        {
            throw new SemanticException(se.getMessage(), token);
        }
    }
    
    public final EnumType putListElement(
        final String symbol, final EnumType symbolType)
        throws SemanticException
    {
        return this.putListElement(new SimpleData(symbol), symbolType);
    }
    
    public EnumType putListElement(
        final Data symbol, final EnumType symbolType)
        throws SemanticException
    {
        EnumType res = this.put(symbol.getInfo(), symbolType);
        symbolList.add(symbol);
        return res;
    }

    public void checkListElement(final String symbol, final EnumType symbolType)
        throws SemanticException
    {
        if (this.containsKey(symbol)) symbolList.add(new SimpleData(symbol));
        else
        {
            symbolList.clear();
            throw new SemanticException(
                Errors.getMsg(Errors.UNDECLARED_IDENTIFIER, symbol), symbol);
        }
    }

    public String checkDeclaration(final String symbol)
        throws SemanticException
    {
        EnumType type = this.get(symbol);
        if (type == null)
            throw new SemanticException(
                Errors.getMsg(Errors.UNDECLARED_IDENTIFIER, symbol), symbol);

        return symbol;
    }
    
    public final String checkType(
        final String symbol, final EnumType expectedType)
        throws SemanticException
    {
        this.checkType(symbol, new EnumType[]{ expectedType });
        return symbol;
    }

    public final EnumType checkType(
        final String symbol, final EnumType[] expectedType)
        throws SemanticException
    {
        return this.checkType(new SimpleData(symbol), expectedType);
    }
    
    public EnumType checkType(final Data symbol, final EnumType[] expectedType)
        throws SemanticException
    {
        EnumType type = this.get(symbol);
        String info = symbol.toString();
        if (type == null)
            throw new SemanticException(
                Errors.getMsg(Errors.UNDECLARED_IDENTIFIER, info), info);
        else
        {
            int size = expectedType.length;
            for (int i = 0; i < size; ++i)
                if (type.equals(expectedType[i]) ||
                    type.equals(this.anyType)) return expectedType[i];

            throw new SemanticException(
                Errors.getMsg(
                    Errors.WRONG_TYPE,
                    info,
                    Arrays.asList(expectedType).toString()), info);
        }
    }
    
    public Data[] getSymbolList()
    {
        Data[] res = (Data[])symbolList.toArray(Data.EMPTY_DATA);
        symbolList.clear();
        return res;
    }
    
    public final EnumType put(
        final GenericToken token, final EnumType symbolType)
        throws SemanticException
    {
        try { return this.put(new SimpleData(token.toString()), symbolType); }
        catch (SemanticException se)
        {
            throw new SemanticException(se.getMessage(), token);
        }
    }
    
    public final EnumType put(final String symbol, final EnumType symbolType)
        throws SemanticException
    {
        return this.put(new SimpleData(symbol), symbolType);
    }

    public EnumType put(final Data symbol, final EnumType symbolType)
        throws SemanticException
    {
        String info = symbol.getInfo();
        
        if (symbolType.getClass() != enumTypeClass)
            throw new SemanticException(
                Errors.getMsg(Errors.WRONG_TYPE, enumTypeClass.getName()), info);
        
        EnumType type = this.get(symbol);
        if (type != null)
        {
            throw new SemanticException(
                Errors.getMsg(Errors.ALREADY_DECLARED_IDENTIFIER, info), info);
        }
        else super.put(symbol, type = symbolType);
        
        return type;
    }
    
    public final EnumType get(final String symbol)
    {
        return this.get(new SimpleData(symbol));
    }
    
    public final EnumType get(final Data symbol)
    {
        return (EnumType)super.get(symbol);
    }
    
    public EnumType changeType(final Data symbol, final EnumType type)
    {
        return (EnumType)super.put(symbol, type);
    }
}
