/*
 * InstructionImpl.java
 *
 * Created on 20 febbraio 2004, 17.15
 */

package it.unive.dsi.pear.core.parser;

import java.util.*;

import it.unive.dsi.pear.util.*;

/** This class is a straightforward implementation of <CODE>Instruction</CODE>.
 * @author Francesco Placella
 */
public class InstructionImpl extends RootData implements Instruction
{
    private static final String NIL = "0";
    
    public static final int DEFAULT_BUF_SIZE = 30;
    
    /** Creates a new instance of <CODE>Instruction</CODE>.
     * @param name the instruction name
     * @param data the instruction arguments
     */
    public InstructionImpl(String name, Data[] data)
    {
        this.name = name;
        
        this.setData(data = data != null ? data : Data.EMPTY_DATA);
        
        int size = data.length;
        for (int i = 0; i < size; ++i)
        {
            Data d = data[i];
            if (d instanceof AbstractData)
            {
                this.abstractIndex = i;
                break;
            }
        }
    }
    
    /** Holds value of property data. */
    private Data[] data;
    
    /** Holds value of property abstractInstr. */
    private int abstractIndex = ABSTRACT_INDEX;
    
    /** Holds value of property name. */
    private String name;
    
    public String toString()
    {
        if (NIL.equals(this.getName())) return NIL;
        
        if (this.data != null)
        {
            StringBuffer res = new StringBuffer(DEFAULT_BUF_SIZE);
            res.append(this.getName() + "(");
            
            int size = this.data.length;            
            if (size > 0)
            {
                res.append(this.data[0]);
                for (int i = 1; i < size; ++i) res.append(", " + this.data[i]);
            }
            
            res.append(")");
            return res.toString();
        }
        else return this.getName() + "()";
    }
    
    public final Map appliesTo(Instruction instr)
    {
        return this.appliesTo(instr, null, null);
    }
    
    public Map appliesTo(Instruction instr, String ruleId, Data instrId)
    {
        // the empty instruction aplies to the empty rule
        if (instr.isNil()) return CloneableMap.EMPTY_MAP;
        
        List iData = Arrays.asList(new Data[]{ instr });
        Map substMap = new CloneableMap(true, this.getData().length);
        if (ruleId != null && instrId != null) substMap.put(ruleId, instrId);
        
        if (!this.putMatching(substMap, iData)) return null;

        return substMap;
    }

    public boolean putMatching(Map substMap, List iList)
    {
        Data[] rData = this.getData();
        int rSize = rData.length;
        
        ListIterator it = iList.listIterator();
        while (it.hasNext())
        {
            Object obj = it.next();
            if (obj instanceof Instruction)
            {
                Instruction instr = (Instruction)obj;
                
                if (!this.getName().equals(instr.getName())) return false;

                List iData = new ArrayList(Arrays.asList(instr.getData()));
                int iSize = iData.size();   

                // if the content of the current instruction is not abstract
                // and the number of arguments of the two instructions are 
                // different, the current instruction doesn't match the given
                // one
                if (!this.isAbstract() && iSize != rSize) return false;

                int size = (iSize < rSize) ? iSize : rSize;
                for (int i = 0; i < size; ++i)
                    if (i != this.abstractIndex &&
                        !rData[i].putMatching(substMap, iData)) return false;

                if (this.isAbstract() && 
                    !rData[this.abstractIndex].putMatching(substMap, iData))
                    return false;
                
                it.set(null);
                return true;
            }
        }
        
        return false;
    }

   
    /** Getter for property data.
     * @return Value of property data.
     *
     */
    public final Object[] toArray() { return this.getData(); }
    public Data[] getData() {
        return this.data;
    }
    
    /** Getter for property abstract
     * @return Value of property abstract.
     *
     */
    public boolean isAbstract()
    {
        return this.abstractIndex != ABSTRACT_INDEX;
    }
    
    /** Getter for property name.
     * @return Value of property name.
     *
     */ 
    public String getName() { return this.name; }
    public final String getInfo() { return this.getName(); }    
    
    /** Setter for property name.
     * @param name New value of property name.
     *
     */     
    public void setName(String name) { this.name = name; }
    public final void setInfo(String name) { this.setName(name); }

    public boolean contains(Data data)
    {
        int size = this.data.length;
        for (int i = 0; i < size; ++i)
            if (this.data[i].equals(data)) return true;
        
        return false;
    }
    
    public boolean equals(Object obj)
    {
        if (!(obj instanceof Instruction)) return false;
        
        Instruction instr = (Instruction)obj;
        if (!this.getInfo().equals(instr.getInfo())) return false;
        
        Data[] data = this.getData();
        Data[] instrData = instr.getData();
        
        int size = data.length;
        if (size != instrData.length) return false;

        for (int i = 0; i < size; ++i)
            if (!data[i].equals(instrData[i])) return false;

        return true;
    }
    
    public void subst(Map substMap, boolean removeAbstract)
    {
        List contentList = new LinkedList();
        int size = data.length;
        boolean ra = false;
        
        for (int i = 0; i < size; ++i)
        {
            Data data = this.data[i];
            if (data instanceof AbstractData)
            {
                ra = ((AbstractData)data).subst(substMap, contentList);
                abstractIndex = ABSTRACT_INDEX;
            }
            else
            {
                data.subst(substMap, removeAbstract);
                contentList.add(data);
            }
        }
        
        if (contentList.size() >= size || (removeAbstract && ra))
        {
            this.setData((Data[])contentList.toArray(Data.EMPTY_DATA));
            substMap.remove(AbstractData.ABSTRACT_DATA);
        }
    }
    
    public final Object clone()
    {
        InstructionImpl instr = (InstructionImpl)super.clone();
        
        Data[] args = instr.getData();
        instr.data = (Data[])args.clone();
        
        int size = args.length;
        for (int i = 0; i < size; ++i)
        {
            Data data = args[i];
            if (data != null) instr.data[i] = (Data)data.clone();
        }
        
        return instr;
    }

    /** Setter for property data.
     * @param data New value of property data.
     *
     */
    public void setData(Data[] data) { this.data = data; }
  
    public boolean containsTagged()
    {
        Data[] content = this.getData();
        int size = content.length;
        for (int i = 0; i < size; ++i)
        {
            Data data = content[i];
            if (data.containsTagged()) return true;
        }
        return false;
    }
    
    public int getTagNumber()
    {
        int tagNumber = 0;
        
        Data[] content = this.getData();
        int size = content.length;
        for (int i = 0; i < size; ++i)
        {
            Data data = content[i];
            tagNumber += data.getTagNumber();
        }
        
        return tagNumber;
    }

    public boolean isNil() { return NIL.equals(this.getName()); }
}