/*
 * KeyData.java
 *
 * Created on 15 marzo 2004, 23.11
 */

package it.unive.dsi.pear.core.parser;

import java.util.*;

/** This class models a symmetric key declaration, keep in mind that
 * <I>sym-key(A,B) = sym-key(B,A)</I>.
 * @author Francesco Placella
 */
public class KeyData extends CompositeData
{
    public static final String KEY_INSTR = "sym-key";
    
    /** Creates a new instance of <CODE>KeyData</CODE>.
     * @param key the key name
     */
    public KeyData(String key) { this(key, null); }
    /** Creates a new instance of <CODE>KeyData</CODE>.
     * @param key the key name
     * @param principals an array containing the two principals
     */    
    public KeyData(String key, SimpleData[] principals)
    {
        super(key, null);
        this.setPrincipals(principals);
    }
 
    public final void setEncrypted(boolean encrypted) {}
    public final Data getKey() { return this; }
    public final void setKey(String key) { this.setInfo(key); }
    public final void setContent(Data[] content)
        { this.setPrincipals(content); }
    protected final void forceContent(Data[] content)
        { super.setContent(content); }

    /** Returns the principals associated with the key. */    
    public Data[] getPrincipals()
    {
        Data[] content = super.getContent();
        return content != null && content.length > 0 ?
            ((Instruction)content[0]).getData() : null;
    }
    
    /** Sets the new principals for the key. */    
    public void setPrincipals(Data[] p)
    {
        super.setContent(new Data[]{ new InstructionImpl(KEY_INSTR, p) });
    }

    /** Checks if the given key declaration matches the current one.
     * @param key the key to be matched
     * @param substMap the substitution map is populated with the apprpriate
     * mappings
     * @param declSelf a principal associated with the current key declaration
     * @param self a principal associated with the key declaration to be matched
     */    
    public boolean appliesTo(
        KeyData key, Map substMap, String declSelf, String self)
    {
        if (key == null) return false;
            
        Data[] principals = this.getPrincipals();        
        Data[] keyPrincipals = key.getPrincipals();
        
        int size = principals.length;
        if (size != keyPrincipals.length) return false;

        for (int i = 0; i < size; ++i)
        {
            String principal = principals[i].getInfo();
            String keyPrincipal = keyPrincipals[i].getInfo();
            String subst = (String)substMap.get(principal);
            if (subst == null && !substMap.values().contains(keyPrincipal))
                substMap.put(principal, subst = keyPrincipal);
            else if (!keyPrincipal.equals(subst)) return false;
            if (subst != null && subst.equals(self) &&
                !principal.equals(declSelf)) return false;
        }

        return true;
    }

    public boolean equals(Object obj)
    {
        if (!(obj instanceof KeyData)) return false;
        
        KeyData kd = (KeyData)obj;
        String info = this.getInfo();
        if (info != null && !info.equals(kd.getInfo())) return false;
        
        Set principals = new HashSet(Arrays.asList(this.getPrincipals()));
        Set kdPrincipals = new HashSet(Arrays.asList(kd.getPrincipals()));
        
        return principals.equals(kdPrincipals);
    }
    
    protected String keyInstr() { return KEY_INSTR; }
    
    public String toString()
    {
        String sup = this.getInfo();
        
        StringBuffer res = new StringBuffer(DEFAULT_BUF_SIZE);
        if (sup != null) res.append(sup + " = ");
        
        res.append(this.keyInstr()).append("(");
        
        Data[] content = this.getPrincipals();
        int size = content.length;
        if (size > 0)
        {
            res.append(content[0]);
            for (int i = 1; i < size; ++i) res.append(", " + content[i]);
        }
        
        res.append(")");
        
        return res.toString();
    }
}
