/*
 * JDockableWindow.java
 *
 * Created on 23 aprile 2004, 17.02
 */

package it.unive.dsi.pear;

import java.awt.*;

import javax.swing.*;

/** This class implements a dockable internal frame. It needs a docking desktop.
 * @author Francesco Placella
 */
public class JDockableInternalFrame extends javax.swing.JInternalFrame
{
    private static final int UNDEFINED_DOCKED_POS = -1;
    static final float X_DIM_FACTOR = 0.45f;
    static final float Y_DIM_FACTOR = 0.2f;
    
    /** Holds value of property docked. */
    private boolean docked = false;
    private int dockedX = UNDEFINED_DOCKED_POS;
    private int dockedY = UNDEFINED_DOCKED_POS;
    private JDockingDesktop desktop;
    private Object defaultDockingArea;
    private Object dockingArea;
    private Rectangle undockedRect;
    private boolean undockedMaximum;
    
    /** Creates a new instance of JDockableWindow.
     * @param desktop the docking desktop
     * @param dockingArea an object identifying the default place where the 
     * frame has to be docked
     */
    public JDockableInternalFrame(JDockingDesktop desktop, Object dockingArea)
    {
        this.desktop = desktop;
        this.defaultDockingArea = dockingArea;
        this.setDefaultCloseOperation(JInternalFrame.HIDE_ON_CLOSE);
        this.setClosable(true);
        this.setResizable(true);
        this.setMovingProperties(true);
        this.setMinimumSize(new Dimension(0, 0));
        this.updatePreferredSize();
    }
    
    /** Getter for property docked.
     * @return Value of property docked
     */
    public boolean isDocked() {
        return docked;
    }
    
    /** Setter for property docked.
     * @param docked New value of property docked.
     *
     */
    private void setDocked(boolean docked) {
        this.docked = docked;
    }

    /** Moves, if not docked, and resizes this component.
     * Overrides <CODE>JInternalFrame.reshape</CODE>.
     * @param x an integer giving the component's new horizontal position
     * measured in pixels from the left of its container
     * @param y an integer giving the component's new vertical position,
     * measured in pixels from the bottom of its container
     * @param width an integer giving the component's new width in pixels
     * @param height an integer giving the component's new height in pixels
     */    
    public final void reshape(int x, int y, int width, int height)
    {
        if (this.isDocked())
        {
            this.setPreferredSize(new Dimension(width, height));
            this.updateDockedCoords();
            x = dockedX;
            y = dockedY;
        }
        super.reshape(x, y, width, height);
    }
    
    /** Docks the frame to the default docking area. */    
    public final void dock() { this.dock(this.defaultDockingArea); }
    
    /** Docks the frame to the requested docking area.
     * @param dockingArea an object identifying the requested docking area
     */    
    public void dock(Object dockingArea)
    {
        if (BorderLayout.CENTER.equals(dockingArea)) return;
        if (!desktop.canDock(dockingArea)) return;
        
        if (!this.isDocked())
        {
            undockedRect = this.getBounds();
            undockedMaximum = this.isMaximum();
        }
        this.setDocked(true);
        this.setMovingProperties(false);
        this.setSize(this.updatePreferredSize());
        try { this.setMaximum(false); } catch (Exception e) {}
        
        desktop.dock(this, dockingArea);
    }

    /** Undocks the frame from the current docking area. */    
    public void undock()
    {
        desktop.undock(this);
        
        this.setDocked(false);
        this.setMovingProperties(true);        
        
        try { this.setMaximum(this.undockedMaximum); } catch (Exception e) {}
        
        if (this.undockedRect != null)
        {
            Rectangle r = this.undockedRect;
            this.setBounds(r.x, r.y, r.width, r.height);
            this.moveToFront();
        }
    }
    
    /** Makes the component visible or invisible. Overrides
     * <CODE>jvax.swing.JComponent.setVisible</CODE>.
     * @param visible true to make the component visible; false to make it
     * invisible
     */    
    public void setVisible(boolean visible)
    {
        if (!visible && this.isDocked()) this.undock();
        super.setVisible(visible);
    }
    
    /** Unselects the current frame. */    
    public void unselect()
    {
        try { super.setSelected(false); }
        catch (java.beans.PropertyVetoException pve) {}
    }
    
    /** Selects or deselects the internal frame if it's showing.
     * Overrides <CODE>JInternalFrame.setSelected</CODE>
     * @param selected a boolean, where true means this internal frame should 
     * become selected (currently active) and false means it should become 
     * deselected
     * @throws PropertyVetoException when the attempt to set the property is 
     * vetoed by the internal frame
     */    
    public void setSelected(boolean selected)
        throws java.beans.PropertyVetoException
    {
        if (selected) desktop.updateSelection(this);
        super.setSelected(selected);
    }
    
    private void setMovingProperties(boolean value)
    {
        this.setMaximizable(value);
        this.setIconifiable(value);
    }
 
    /* updateDockedCoords updates the coordinates of the frame according to the 
     * current size of the docking desktop.
     */
    private final void updateDockedCoords()
    {
        int dHeight = desktop.getHeight();
        int dWidth = desktop.getWidth();
        
        dockedX = 0;
        dockedY = desktop.getDockingAreaSize(BorderLayout.NORTH).height;
        
        Dimension preferredSize = this.getPreferredSize();
        if (BorderLayout.SOUTH.equals(dockingArea))
            dockedY = dHeight - preferredSize.height;
        else if (BorderLayout.EAST.equals(dockingArea))
            dockedX = dWidth - preferredSize.width;
        else if (BorderLayout.NORTH.equals(dockingArea))
            dockedY = 0;
    }

    /* updatePreferredSize resizes the frame according to the current size of 
     * the docking desktop.
     */
    private Dimension updatePreferredSize()
    {
        int pWidth = (int)(X_DIM_FACTOR * desktop.getDockingWidth());
        int pHeight = (int)(Y_DIM_FACTOR * desktop.getDockingHeight());
        Dimension res = new Dimension(pWidth, pHeight);
        this.setPreferredSize(res);
        return res;
    }
    
    /** Getter for property dockingArea.
     * @return Value of property dockingArea.
     *
     */
    public java.lang.Object getDockingArea() {
        return dockingArea;
    }
    
    /** Setter for property dockingArea.
     * @param dockingArea New value of property dockingArea.
     *
     */
    public void setDockingArea(java.lang.Object dockingArea) {
        this.dockingArea = dockingArea;
    }
    
}
