/*
 * JDockingDesktop.java
 *
 * Created on 25 aprile 2004, 16.49
 */

package it.unive.dsi.pear;

import java.util.*;
import java.awt.*;

import javax.swing.*;

/** This class implements a docking desktop, that is a desktop where frames can
 * be docked on the four sides.
 * @author Francesco Placella
 */
public class JDockingDesktop extends javax.swing.JPanel
{
    private static final Dimension ZERO = new Dimension(0, 0);
    
    private JDesktopPane desktop;
    private HashMap docked = new HashMap();
    private JDockableInternalFrame selectedFrame;
    private Set iFrames = new HashSet();
    
    /** Creates a new instance of JDockingDesktop */
    public JDockingDesktop()
    {
        desktop = new JDesktopPane();
        desktop.setDragMode(JDesktopPane.OUTLINE_DRAG_MODE);
        this.setLayout(new java.awt.BorderLayout());        
        this.add(desktop, java.awt.BorderLayout.CENTER);
    }
    
    /** Adds a docking frame to the desktop.
     * @param dif the docking frame
     */    
    public final void addIFrame(JDockableInternalFrame dif)
    {
        if (!iFrames.contains(dif))
        {
            dif.setVisible(true);
            desktop.add(dif, javax.swing.JLayeredPane.DEFAULT_LAYER);
            iFrames.add(dif);
        }
        else if (dif.isDocked()) dif.undock();
    }
    
    /** Removes a docking frame from the desktop.
     * @param dif the docking frame
     */    
    public final void removeIFrame(JDockableInternalFrame dif)
    {
        if (iFrames.contains(dif))
        {
            iFrames.remove(dif);
            if (dif.isDocked()) this.remove(dif);
            else desktop.remove(dif);            
        }
    }
    
    /** Gets the docking width available.
     * @return the current desktop width
     */    
    public int getDockingWidth() { return desktop.getWidth(); }
    /** Gets the docking height available.
     * @return the current desktop height
     */    
    public int getDockingHeight() { return desktop.getHeight(); }
    
    /** Docks a docking frame to the specified docking area.
     * @param dif the docking frame
     * @param dockingArea an object identifying the docking area
     */    
    public void dock(JDockableInternalFrame dif, Object dockingArea)
    {
        if (this.canDock(dockingArea))
        {
            docked.remove(dif.getDockingArea());
            docked.put(dockingArea, dif);
            desktop.remove(dif);        
            this.add(dif, dockingArea);
            dif.setDockingArea(dockingArea);
        }
    }

    /** Undocks a docking frame putting it onto the desktop.
     * @param dif the docking frame
     */    
    public void undock(JDockableInternalFrame dif)
    {
        if (dif.isDocked() && iFrames.contains(dif))
        {
            this.remove(dif);
            desktop.add(dif, javax.swing.JLayeredPane.DEFAULT_LAYER);
            docked.remove(dif.getDockingArea());
            dif.setDockingArea(null);
        }
    }
    
    /** Verifies if the specified docking area is free.
     * @param dockingArea an object identifying the docking area
     * @return <CODE>true</CODE> if the docking area is free,
     * <CODE>false</CODE> otherwise
     */    
    public boolean canDock(Object dockingArea)
    {
        return !docked.containsKey(dockingArea);
    }
    
    /** Computes the size of the specified docking area.
     * @param dockingArea an object identifying the docking area
     * @return the size of the specified docking area or zero if it's free
     */    
    public Dimension getDockingAreaSize(Object dockingArea)
    {
        JDockableInternalFrame dif =
            (JDockableInternalFrame)docked.get(dockingArea);
        
        if (dif != null) return dif.getSize();
        return ZERO;
    }
    
    /** Selects the specified frame, unselecting previous one.
     * @param selectedFrame the frame to be selected.
     */    
    public void updateSelection(JDockableInternalFrame selectedFrame)
    {
        if (this.selectedFrame != null) this.selectedFrame.unselect();
        this.selectedFrame = selectedFrame;
    }

    /** Undocks all the docking frames. */    
    public void undockAll()
    {
        HashMap dc = (HashMap)docked.clone();
        Iterator it = dc.values().iterator();
        while (it.hasNext())
        {
            JDockableInternalFrame dif = (JDockableInternalFrame)it.next();
            if (dif.isDocked()) dif.undock();
        }
    }
    
    /** Getter for property selectedFrame.
     * @return Value of property selectedFrame.
     *
     */
    public JDockableInternalFrame getSelectedFrame() { return selectedFrame; }
}
