/*
 * JProtocolSourceArea.java
 *
 * Created on 28 aprile 2004, 17.05
 */

package it.unive.dsi.pear;

import java.util.*;
import java.io.*;
import java.awt.event.*;
import java.awt.*;

import javax.swing.*;

import it.unive.dsi.pear.log.Log;
import it.unive.dsi.pear.core.parser.LogError;
import it.unive.dsi.pear.core.parser.ParseException;
import it.unive.dsi.pear.core.parser.Protocol;
import it.unive.dsi.pear.core.parser.ProtocolNode;
import it.unive.dsi.pear.core.Parser;

/** This class implements a text area which is used to display and edit a given
 * protocol source.
 * @author Francesco Placella
 */
public class JProtocolSourceArea extends javax.swing.JPanel
    implements ActionListener, KeyListener, Runnable
{
    /** The name of the property whose value has changed. */    
    public static final String PROPERTY_NAME = "sourceText";
    
    private static final Font CONSOLE_FONT = Pear.getConsoleFont();
    private static final Font DEFAULT_FONT = Pear.getDefaultFont();
    private static final Insets TEXT_AREA_MARGINS = new Insets(2, 5, 2, 5);
    private static final Dimension BUTTON_SIZE = new Dimension(65, 25);
    private static final Color VALID = new Color(0,153,0);
    private static final Color WARNING_COLOR = Color.GRAY;

    private static final ImageIcon MODIFIED_ICON =
        new ImageIcon(JProtocolSourceArea.class.
            getResource("/resources/modified.gif"));
    
    private static final String MODIFIED_TEXT = "!";
    private static final String INDENT = " ";
    private static final String PREFIX = "";
    private static final String NEW_SOURCE = "";
    private static final String ALLOWED_CTRL_KEYS = "xXvV";
    private static final String CUT_CTRL_KEY = "x";
    
    private static final int BUFFER_SIZE = 512;
    private static final int SEP = 5;
    
    private static final Log.Message[] MSG_ARRAY = {};    
    
    private final java.util.List errors = new LinkedList();
    private final JTextArea sourceArea = new JTextArea();
    private final JScrollPane scrollPane = new JScrollPane(sourceArea);
    private final JPanel buttonPanel = new JPanel(new FlowLayout(FlowLayout.RIGHT, 0, 0));
    private final JPanel statePanel = new JPanel(new BorderLayout());
    private final JPanel taggingPanel = new JPanel(new FlowLayout(FlowLayout.LEFT, 0, 0));
    private final JLabel stateLabel = new JLabel();
    private final JLabel modifiedLabel = new JLabel();
    private final Color defaultColor = sourceArea.getForeground();
    private final Color selectedTextColor = sourceArea.getSelectedTextColor();
    private final Color selectionColor = sourceArea.getSelectionColor();
    private final JButton save = new JButton("save");
    private final JButton cancel = new JButton("don't");
    private final JTaggingBox taggingBox = new JTaggingBox();
    private final boolean readonly;

    private Parser parser;
    private Log log;
    private Protocol protocol;
    private Protocol tagging;
    private Boolean valid;
    private File sourceFile;
    private boolean editable = false;
    private boolean modified = false;
    private boolean previouslyModified = false;
    private boolean proceed = false;
    private boolean forceSave = false;
    private String source;
    private Font defaultFont;
    private Font warningFont;
    private Log.Message[] warnings;
    
    // ---------------------------------------------- costruttori
    
    /** Creates a new instance of the class.
     * @param parser an object allowing to parse and syntax-check the given
     * protocol
     * @param sourceFile a file containing the source of the protocol
     * @param log an object implementing a logger
     * @throws FileNotFoundException if the file does not exist, is a directory
     * rather than a regular file, or for some other reason cannot be opened for
     * reading.
     * @throws IOException if some other kind of I/O problem is detected
     * @throws ParseException if a syntax error is detected while parsing the
     * given source
     */
    
    public JProtocolSourceArea(Parser parser, File sourceFile, final Log log)
        throws FileNotFoundException, IOException, ParseException
    {
        super(new BorderLayout());
        readonly = false;
        this.loadSource(sourceFile);
        this.init(parser, log);
   }

    public JProtocolSourceArea(Parser parser, Reader sourceReader, final Log log)
        throws FileNotFoundException, IOException, ParseException
    {
        super(new BorderLayout());
        readonly = true;
        try { sourceArea.read(sourceReader, sourceFile); }
        finally { if (sourceReader != null) sourceReader.close(); }
        source = sourceArea.getText();
        this.init(parser, log);
    }

    private void init(Parser parser, final Log log)
        throws ParseException, IOException
    {
        this.parser = parser;
        this.log = log;
        
        defaultFont = CONSOLE_FONT;
        warningFont = this.getWarningFont();
        
        sourceArea.setMargin(TEXT_AREA_MARGINS);
        
        if (source != NEW_SOURCE)
        {
            sourceArea.setEditable(false);
            try
            {
                this.checkSource(this.getSourceReader(), true);
                this.resetText();
            }
            catch (ParseException ex)
            {
                this.setErrors(ex);
                this.setEditable(true);
            }
        }
        
        tagging = protocol;

        this.add(scrollPane);
        stateLabel.setHorizontalAlignment(SwingConstants.LEFT);
        modifiedLabel.setHorizontalAlignment(SwingConstants.CENTER);
        
        save.setPreferredSize(BUTTON_SIZE);
        save.addActionListener(this);
        
        cancel.setPreferredSize(BUTTON_SIZE);
        cancel.addActionListener(this);
                
        statePanel.add(stateLabel, BorderLayout.WEST);
        buttonPanel.add(save);
        buttonPanel.add(cancel);
        statePanel.setVisible(false);

        taggingPanel.add(taggingBox);
        taggingBox.setProtocol(protocol);
        taggingBox.addActionListener(this);
        
        Font f = CONSOLE_FONT;
        modifiedLabel.setFont(new Font(f.getName(), Font.BOLD, f.getSize()));
        scrollPane.setCorner(JScrollPane.LOWER_RIGHT_CORNER, modifiedLabel);
        
        sourceArea.addKeyListener(this);
    }
 
    
    // ---------------------------------------------- metodi pubblici
    
    /** Loads a source from the given file.
     * @param sourceFile the file to be loaded.
     * @throws FileNotFoundException if the file does not exist, is a directory
     * rather than a regular file, or for some other reason cannot be opened for
     * reading.
     * @throws IOException if some other kind of I/O problem is detected
     * @return a string representation of the source file.
     */    
    public String loadSource(File sourceFile)
        throws FileNotFoundException, IOException
    {
        this.sourceFile = sourceFile;
        
        if (sourceFile.exists())
        {
            BufferedReader sourceReader = null;

            try
            {
                sourceReader = new BufferedReader(new FileReader(sourceFile));
                sourceArea.read(sourceReader, sourceFile);
            }
            finally { if (sourceReader != null) sourceReader.close(); }
            
            return source = sourceArea.getText();
        }
        
        sourceArea.setText(source = NEW_SOURCE);
        this.setEditable(true);
        return source;
    }

    /** Writes to disk the current source.
     * @throws FileNotFoundException if the file does not exist, is a directory
     * rather than a regular file, or for some other reason cannot be opened for
     * reading.
     * @throws IOException if some other kind of I/O problem is detected
     */    
    public void saveSource()
        throws FileNotFoundException, IOException
    {
        BufferedWriter sourceWriter = null;

        try
        {
            sourceWriter = new BufferedWriter(new FileWriter(sourceFile));
            if (editable) sourceArea.write(sourceWriter);
            else
            {
                String p = sourceArea.getText();
                sourceArea.setText(source);
                sourceArea.write(sourceWriter);
                sourceArea.setText(p);
                sourceArea.setCaretPosition(0);
            }
        }
        finally { if (sourceWriter != null) sourceWriter.close(); }
    }
    
    /** Appends the given string to the associated text area.
     * @param str the string to be appended
     */    
    public final void append(String str) { sourceArea.append(str); }
    
    /** Performs a syntax-check of the given source.
     * @param sourceReader a <CODE>Reader</CODE> for the source to be parsed
     * @param keep if <CODE>true</CODE> the parse tree obtained by the parser,
     * if the check is successful, will be kept, otherwise discarded.
     * @throws ParseException if a syntax error is detected while parsing the
     * given source
     * @throws IOException if some kind of I/O problem is detected
     * @return an array of messages generated by the parser
     */
    public Log.Message[] checkSource(Reader sourceReader, boolean keep)
        throws ParseException, IOException
    {
        try
        {
            errors.clear();
            Protocol pn = parser.parseProtocol(sourceReader);
            if (keep) this.setProtocol(pn);
        }
        finally { if (sourceReader != null) sourceReader.close(); }

        return warnings = log.getWarnings(true);
    }
    
    /** Sets the font of the components contained by the current instance to one
     * specified.
     * @param font the font to be set.
     */    
    public void setComponentFont(Font font)
    {
        super.setFont(font);
        stateLabel.setFont(font);
        save.setFont(font);
        cancel.setFont(font);
        taggingBox.setFont(font);
    }
    
    /** Selects the given protocol node: different colors will be used whether the
     * related source is editable or not.
     * @param instr the protocol node to be selected
     */    
    public void select(ProtocolNode instr)
    {
        if (instr != null)
        {
            StringBuffer sb = new StringBuffer(BUFFER_SIZE);
            Protocol pn = this.getProtocol();
            int startSel = this.protocolDump(pn, instr, sb, 0, PREFIX);
            int endSel = startSel + instr.toString().length();
            sourceArea.setText(sb.toString());
            sourceArea.requestFocus();
            sourceArea.setCaretPosition(startSel);
            sourceArea.moveCaretPosition(endSel);
        }
    }
    
    /** Sets the text color to the default one. */    
    public void resetColor()
    {
        sourceArea.setForeground(defaultColor);
    }
    
    /** The text area is filled with the textual representation of the protocol and
     * displaying mode is set.
     */    
    public void resetText()
    {
        Protocol protocol = this.getProtocol();
        if (protocol == null) protocol = tagging = this.getActualProtocol();
        
        StringBuffer sb = new StringBuffer(BUFFER_SIZE);
        this.protocolDump(protocol, null, sb, 0, PREFIX);
        
        boolean warnings = this.hasWarnings();
        sourceArea.setFont(warnings ? warningFont : defaultFont);
        sourceArea.setForeground(warnings ? WARNING_COLOR : defaultColor);
        sourceArea.setText(sb.toString());
        sourceArea.setCaretPosition(0);
        sourceArea.setSelectionColor(sourceArea.getBackground());
        sourceArea.setSelectedTextColor(Color.BLUE);
    }    
    
    /** Writes to disk the current source after asking confirmation in case its
     * syntax is not valid.
     * @return <CODE>true</CODE> if the saving has been performed,
     * <CODE>false</CODE> otherwise.
     */    
    public synchronized boolean save()
    {
        boolean res = true;
        
        try
        {
            if (!forceSave)
            {
                try { this.checkSource(this.getSourceReader(), true); }
                catch (ParseException ex)
                {
                    this.setErrors(ex);
                    
                    forceSave = true;
                    this.showSaveDialog(" protocol source is not valid");

                    return false;
                }
                catch (Throwable t) {  log.error(t); }
            }

            statePanel.setVisible(false);
            this.saveSource();
            previouslyModified = this.isModified();
            this.setModified(false);
            proceed = true;            
            this.notifyAll();
        }
        catch (Exception ex)
        {
            log.error(ex);
            res = false;
        }
        
        return res;
    }
    
    /** Displays the rule names applied to the related instructions. */    
    public void appendRules()
    {
        Protocol protocol = this.getProtocol();
        int column = this.getMaxLength(protocol, PREFIX) + SEP;
        StringBuffer sb = new StringBuffer(BUFFER_SIZE);
        this.ruleDump(protocol, sb, column, PREFIX);
        sourceArea.setCaretPosition(0);
        sourceArea.setText(sb.toString());
    }
    
    /** Checks if there are warnings associated with the current source.
     * @return <CODE>true</CODE> if there are warnings, 
     * <CODE>false</CODE> otherwise.
     */    
    public boolean hasWarnings()
    {
        return warnings != null && warnings.length > 0;
    }
    
    
    // ---------------------------------------------- metodi privati
    
    /* getSourceReader returns a Reader for the current source. */
    private Reader getSourceReader()
    {
        if (editable) source = sourceArea.getText();
        return new BufferedReader(new StringReader(source));
    }
    
    /* getWarningFont return the font used to display protocol sources that have
     * warnings.
     */
    private final Font getWarningFont()
    {
        int size = defaultFont.getSize();
        return new Font(defaultFont.getName(), Font.ITALIC, size);
    }
    
    /* showSaveDialog displays a dialog docked into the current text area asking
     * for the confirmation of throws message passed as argument.
     */
    private void showSaveDialog(String message)
    {
        this.remove(taggingPanel);
        this.add(statePanel, BorderLayout.NORTH);

        sourceArea.setEditable(false);
        stateLabel.setText(message);
        statePanel.add(buttonPanel, BorderLayout.EAST);
        statePanel.setVisible(true);
    }
    
    /* protocolDump creates a string representation of the protocol and returns
     * the index of the specified protocol node.
     */
    private int protocolDump(ProtocolNode node, ProtocolNode instr, StringBuffer sb, int index, String pre)
    {
        sb.append(pre);
        index = sb.length();
        
        int res = node == instr ? index : -1;
        sb.append(node.toString()).append("\n");

        int size = node.getChildCount();
        for (int i = 0; i < size; ++i)
        {
            ProtocolNode rn = (ProtocolNode)node.getChildAt(i); 
            int ret = this.protocolDump(rn, instr, sb, index, pre + INDENT);
            if (ret != -1) res = ret;
        }
        
        return res;
    }
    
    /* getMaxLength returns the max length of the string representation of a
     * single instruction: this informatio is used to display the name of the
     * associated rule.
     */
    private int getMaxLength(ProtocolNode node, String prefix)
    {
        int res = 0;
        if (node.isInstructionNode())
            res += node.toString().length() + prefix.length();

        int size = node.getChildCount();
        for (int i = 0; i < size; ++i)
        {
            ProtocolNode rn = (ProtocolNode)node.getChildAt(i);
            int len = this.getMaxLength(rn, prefix + INDENT);
            if (res < len) res = len;
        }
        return res;
    }
    
    /* ruleDump creates a string representation of the protocol appending
     * the applied rules after the related instructions.
     */
    private void ruleDump(
        ProtocolNode node, StringBuffer sb, int column, String prefix)
    {
        sb.append(prefix);        
        String str = node.toString();
        Object deco = node.getDecoration();

        if (deco != null)
        {
            String rna = deco.toString();
            String sep = node.getSeparator();
            boolean nwl = sep.equals("\n"); 
            String rap = nwl ? str.replace('\n',  ' ') : str;
            int rnl = rna.length();
            int ral = rap.length();
            sb.ensureCapacity(sb.length() + ral + column + rnl + 3);
            sb.append(rap);
            int limit = column - (prefix.length() + ral);
            if (limit > 0)
            {
                char[] spaces = new char[limit];
                Arrays.fill(spaces, ' ');
                sb.append(spaces);
            }
            sb.append("(").append(rna).append(")");
            if (nwl) sb.append("\n");
        }
        else sb.append(str);
        sb.append("\n");
        
        int size = node.getChildCount();
        for (int i = 0; i < size; ++i)
        {
            ProtocolNode rn = (ProtocolNode)node.getChildAt(i);
            this.ruleDump(rn, sb, column, prefix + INDENT);
        }
    }
    
    
    // ---------------------------------------------- properties
   
    /** Getter for property sourceFile.
     * @return Value of property sourceFile.
     *
     */
    public java.io.File getSourceFile() {
        return sourceFile;
    }
    
    /** Setter for property sourceFile.
     * @param sourceFile New value of property sourceFile.
     *
     */
    public void setSourceFile(java.io.File sourceFile) {
        this.sourceFile = sourceFile;
    }
    
    /** Getter for property sourceArea.
     * @return Value of property sourceArea.
     *
     *
    public javax.swing.JTextArea getTextArea() {
        return sourceArea;
    }
     */
    
    /** Getter for property valid.
     * @return Value of property valid.
     *
     */
    public boolean isValid() {
        return valid.booleanValue();
    }
    
    /** Setter for property valid.
     * @param valid New value of property valid.
     *
     */
    public void setValid(boolean valid)
    {
        this.valid = new Boolean(valid);
        Color color = valid ? color = VALID : Color.RED;
        sourceArea.setForeground(color);
    }

    /** Getter for property valid.
     * @return Value of property valid.
     *
     */
    public java.lang.Boolean getValid() {
        return valid;
    }

    /** Returns the parse tree of the selected tagging of the current protocol.
     * @return Value of property protocol.
     */
    public Protocol getProtocol() {
        return tagging;
    }
    /** Returns the parse tree of the given protocol.
     * @return the parse tree of the given protocol
     */    
    public Protocol getActualProtocol() {
        return protocol;
    }

    
    /** Getter for property editable.
     * @return Value of property editable.
     *
     */
    public boolean isEditable() {
        return this.editable;
    }
    
    /** Setter for property editable.
     * @param editable New value of property editable.
     *
     */
    public void setEditable(boolean editable)
    {
        if (this.editable == editable) return;
        
        taggingBox.setEnabled(!editable);
        sourceArea.getCaret().setVisible(editable);

        if (editable)
        {
            try
            {
                if (source == null) this.loadSource(this.getSourceFile());
                else sourceArea.setText(source);
                sourceArea.setSelectionColor(selectionColor);
                sourceArea.setSelectedTextColor(selectedTextColor);
                sourceArea.setFont(defaultFont);
                sourceArea.setForeground(defaultColor);
                sourceArea.setCaretPosition(0);
                sourceArea.requestFocus();
                
                this.editable = editable;
                sourceArea.setEditable(editable);
            }
            catch (Exception ex) { log.error(ex); }
        }
        else
        {
            boolean mod = previouslyModified || this.isModified();
            if (this.checkSource(mod).length == 0)
            {
                if (mod) this.remove(taggingPanel);
                sourceArea.setEditable(this.editable = editable);
                this.resetText();
                previouslyModified = false;
            }
        }
    }
    
    /** Performs a syntax-check of the current source.
     * @param keep if <CODE>true</CODE> the parse tree obtained by the parser,
     * if the check is successful, will be kept, otherwise discarded.
     * @return an array of messages generated by the parser
     */
    public Log.Message[] checkSource(boolean keep)
    {
        Log.Message[] res = {};
        
        try { this.checkSource(this.getSourceReader(), keep); }
        catch (ParseException ex)
        {
            this.setErrors(ex);
            res = this.getErrors();
        }
        catch (Throwable ex) { log.error(ex); }
        
        return res;
    }
    
    /** Return a string representation of the current source.
     * @return a string representation of the current source
     */    
    public String getSource() { return source; }
    
    /** Returns an array containing the warnings generated by the parser.
     * @return an array containing the warnings generated by the parser
     */    
    public Log.Message[] getWarnings() { return warnings; }
    
    JTextArea getSourceArea() { return sourceArea; }
    
    /** Closes the current protocol source, asking for confirmation in case the
     * source has been modified.
     */    
    public void close()
    {
        if (modified)
        {
            Thread t = new Thread(this);
            t.start();
            try { t.join(); } catch (InterruptedException ie) {}
        }
    }
    
    /** Handles the action performed event: it's generated by the tagging 
     * selector, and the save confirmation dialog.
     * @param evt the event to be handled
     */    
    public synchronized void actionPerformed(ActionEvent evt)
    {
        statePanel.setVisible(false);
        Object source = evt.getSource();
        
        if (source == taggingBox)
        {
            tagging = taggingBox.getSelectedTagging();
            sourceArea.setForeground(defaultColor);
            sourceArea.setText(tagging.getProtocolImage());
            sourceArea.setCaretPosition(0);
        }
        else if (source == save)
            try { this.save(); } catch (Throwable e) { log.error(e); }
        else // don't save
        {
            statePanel.setVisible(false);
            forceSave = false;
            proceed = true;
            this.notifyAll();
        }
        
        sourceArea.setEditable(editable);
     }
    
    /** Displays the dialog that asks for confirmation to close the current 
     * source.
     */
    public synchronized void run()
    {
        proceed = false;
        this.showSaveDialog(" source has been modified");
        while (!proceed)
            try { this.wait(); } catch (InterruptedException ie) {}
        proceed = false;        
    }
    
    /** Handles the key released event.
     * @param e the event to be handled
     */    
    public void keyReleased(KeyEvent e) {}
    
    /** Handles the key pressed event
     * @param e the event to be handled
     */    
    public void keyPressed(KeyEvent e)
    {
        // handles cut and paste modifying
        String c = KeyEvent.getKeyText(e.getKeyCode());
        int m = e.getModifiers();
        if (m == KeyEvent.CTRL_MASK && ALLOWED_CTRL_KEYS.indexOf(c) != -1)
        {
            if (c.equalsIgnoreCase(CUT_CTRL_KEY) &&
                sourceArea.getSelectionStart() == sourceArea.getSelectionEnd())
                return;
            
            this.setModified(true);
        }
    }
    
    /** Handles the key typed event
     * @param e the event to be handled
     */    
    public void keyTyped(KeyEvent e)
    {
        // handles normal modifying
        int mod = e.getModifiers();
        if ((mod == 0 || mod == KeyEvent.SHIFT_MASK) && !e.isActionKey())
            this.setModified(true);
    }
    
    /** Checks if the current source has been modified since the last time a
     * writing to disk has been performed.
     * @return <CODE>true</CODE> if the source has enne modified,
     * <CODE>false</CODE> otherwise.
     */    
    public boolean isModified() { return modified; }
    
    /* setModified notifies the change of the source text. */
    private void setModified(boolean modified)
    {
        if (modified) previouslyModified = true;
        if (readonly) return;
        if (this.modified = modified) modifiedLabel.setIcon(MODIFIED_ICON);
        else modifiedLabel.setIcon(null);
        super.firePropertyChange(PROPERTY_NAME, !modified, modified);
    }

    /** Checks if the current source has errors.
     * @return <CODE>true</CODE> if there are errors,
     * <CODE>false</CODE> otherwise.
     */    
    public boolean hasErrors() { return errors.size() > 0; }
    
    /** Returns an array containing the errors found for the current source.
     * @return an array containing the errors found for the current source
     */    
    public Log.Message[] getErrors()
    {
        return (Log.Message[])errors.toArray(MSG_ARRAY);
    }
    
    /* setErrors records the errors present in the argument. */
    private final void setErrors(ParseException pe)
    {
        errors.add(new LogError(pe));
    }
    
    /** Setter for property protocol.
     * @param protocol New value of property protocol.
     *
     */
    public void setProtocol(Protocol protocol)
    {
        if (protocol != this.protocol)
        {
            this.protocol = tagging = protocol;
            taggingBox.setProtocol(protocol);
            if (!this.isEditable()) this.resetText();
        }    
    }
    
    /** Updates the content of the tagging box hiding it if no taggings are 
     * available.
     */
    public void updateTaggings()
    {
        if (taggingBox.updateTaggings() > 0)
        {
            this.add(taggingPanel, BorderLayout.NORTH);
            taggingBox.selectProtocol();
            if (this.isEditable()) taggingBox.setEnabled(false);
            taggingPanel.setVisible(true);
        }
        else
            this.remove(taggingPanel);
        
        tagging = protocol;
    }

    /** Tells the source area that the related source is (not) being validated.
     * @param validating <CODE>true</CODE> if the protocol source is being 
     * validated, <CODE>false</CODE> otherwise.
     */    
    public void setValidating(boolean validating)
    {
        if (validating) this.setEditable(false);
        taggingBox.setEnabled(!validating);
    }
    
    /** Requests that this Component get the input focus, and that this 
     * Component's top-level ancestor become the focused Window. Overrides
     * <CODE>java.awt.Component</CODE>.
     */    
    public final void requestFocus()
    {
        super.requestFocus();
        sourceArea.requestFocus();    
    }
    
    public boolean isReadonly() { return readonly; }
}