/*
 * Pear.java
 *
 * Created on 23 aprile 2004, 14.05
 */

package it.unive.dsi.pear;

import it.unive.dsi.pear.core.Configuration;
import it.unive.dsi.pear.core.Inference;
import it.unive.dsi.pear.core.StepLevel;
import it.unive.dsi.pear.core.Validation;
import it.unive.dsi.pear.core.ValidationOutput;
import it.unive.dsi.pear.core.parser.ParseException;
import it.unive.dsi.pear.core.parser.Protocol;
import it.unive.dsi.pear.core.parser.SemanticError;
import it.unive.dsi.pear.core.parser.ValidationRules;
import it.unive.dsi.pear.log.Log;
import it.unive.dsi.pear.plugins.tabber.TabPlugin;
import it.unive.dsi.pear.util.ArrayIterator;
import it.unive.dsi.pear.util.ExtensionBasedFileFilter;
import it.unive.dsi.pear.util.MultiMap;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dialog;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Insets;
import java.awt.Rectangle;
import java.awt.Toolkit;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.BufferedOutputStream;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.io.Reader;
import java.io.StringReader;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;

import javax.swing.Icon;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JCheckBoxMenuItem;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JList;
import javax.swing.JTextArea;
import javax.swing.JTree;
import javax.swing.SpinnerNumberModel;
import javax.swing.ToolTipManager;
import javax.swing.UIManager;
import javax.swing.tree.DefaultTreeCellRenderer;
import javax.swing.tree.TreePath;


/** This class implements the GUI, putting the various graphic components 
 * together: it handles validation and inference actions, that is displaying the
 * validation rule tree and container content. Morever it uses the appropriate 
 * graphic components to provide logging and file handling.
 * @author Francesco Placella
 */
public class Pear extends JFrame 
    implements java.beans.PropertyChangeListener, javax.swing.event.InternalFrameListener
{
    private static final Font CONSOLE_FONT;
    private static final Font DEFAULT_FONT;
    
    public static final Font getConsoleFont() { return CONSOLE_FONT; }
    public static final Font getDefaultFont() { return DEFAULT_FONT; }
    
    static
    {
        int consoleFontSize = Integer.parseInt(System.getProperty("consoleFontSize", "11"));
        int fontSize = Integer.parseInt(System.getProperty("fontSize", "11"));
        
        CONSOLE_FONT = new Font("Monospaced", 0, consoleFontSize);
        DEFAULT_FONT = new Font("Dialog", 0, fontSize);
    }
    
    private static final File INITIAL_DIR = new File("..");
    
    private static final Insets TEXT_AREA_MARGINS = new Insets(2, 5, 2, 5);
    private static final Dimension DIALOG_BUTTON_SIZE = new Dimension(65, 25);
    
    private static final float DIM_FACTOR = 0.85f;
    private static final float DIALOG_DIM_FACTOR = 0.60f;
    private static final float MESSAGE_X_DIM_FACTOR = 0.40f;
    private static final float MESSAGE_Y_DIM_FACTOR = 0.15f;
    
    private static final Insets BUTTON_MARGINS = new Insets(2, 2, 2, 2);
    private static final Dimension BUTTON_SIZE = new Dimension(26, 26);
    
    private static final int VALIDATION_TAB = 0;
    private static final int CONSOLE_TAB = 1;
    private static final int WAIT_AFTER_VALIDATION = 1000;
    private static final int PATIENCE_MAX_LIMIT = 100;
    private static final int PATIENCE_MIN_LIMIT = 0;
    private static final int PATIENCE_DEFAULT = 1;
    private static final int LEVEL_FACTOR =
        StepLevel.getMaxLevel() / Log.FULL_OUTPUT + 1;
    
    private static final javax.swing.filechooser.FileFilter ROSPI_FILTER =
        new ExtensionBasedFileFilter("RoSpi");
    private static final javax.swing.filechooser.FileFilter VR_FILTER =
        new ExtensionBasedFileFilter("vr");
    
    private static final String VR_TITLE = "Validation Rules";
    private static final String TMP_PREFIX = "[built-in] ";
    private static final String NO_RESOURCES = "no resources available";
    
    private final File PEAR_HOME = new File(Configuration.getCurrent().getHomeDir());
    
    private volatile StepLevel curStepLevel = StepLevel.STRAIGHT_ON_LEVEL;
    private int curTabIndex = 0;
    private boolean validating = false;
    private boolean infering = false;
    private boolean closing = false;
    private boolean defaultRules = true;
    private File lastDir = INITIAL_DIR;
    
    private File errLogFile;
    private PrintStream errStream;
    private File outLogFile;
    private PrintStream outStream;
    
    private String rulesName;
    private Protocol infProtocol;
    private Validation validationCore;
    private Inference inferenceCore;
    private JDockingDesktop desktopPanel = new JDockingDesktop();
    private JDockableInternalFrame containers;
    private JDockableInternalFrame console;
    private JDockableInternalFrame source;
    private JDockableInternalFrame rules;
    private JTree ruleTree;
    private TreePath ruleRootPath;
    private TreePath ruleLastPath;
    private JProtocolSourceArea currentSource;
    private JProtocolSourceArea infSource;
    private JProtocolSourceArea selectedSource;
    private JNotifyingTextArea errConsole;
    private Collection builtinRules;
    private Collection builtinProtocols;
    private JTaggingBox taggingBox = new JTaggingBox();
    private boolean rulesEnabled = true;
    
    private final Configuration conf = Configuration.getCurrent();
    private final Map openedFiles = new HashMap();
    private final Map vruleMap = new MultiMap();
    private final Map descMap = new HashMap();
    private final ValidationRules defRules;
    private final JFileChooser fileChooser = new JFileChooser();
    private final JTextAreaStream tas;
    private final JTextAreaStream errTas;
    private final Dimension screenDim;
    private final Log log;
    private final Icon openIcon =  new ImageIcon(getClass().getResource("/resources/source.gif"));
    private final Icon openRulesIcon =  new ImageIcon(getClass().getResource("/resources/rules.gif"));
    private final DefaultTreeCellRenderer renderer =
        new DefaultTreeCellRenderer();
    
    private class ToolTipCellRenderer
        extends javax.swing.plaf.basic.BasicComboBoxRenderer
    {
        public final Component getListCellRendererComponent(
            JList list, Object value, int index, boolean isSelected, boolean cellHasFocus)
        {
            if (value != null)
            {
                String v = value.toString();
                String desc = (String)descMap.get(value);
                
                if (desc == null) this.setToolTipText(v);
                else
                {
                    StringBuffer tip = new StringBuffer((v.length() + desc.length()) * 2);
                    tip.append("<HTML>").append(v).append("<BR>").append(desc.replaceAll("\\n", "<BR>")).append("</HTML>");
                    this.setToolTipText(tip.toString());
                }
            }
            
            Component c = super.getListCellRendererComponent(list, value, index, isSelected, cellHasFocus);
            
            return c;
        }
    }
    
    /** Creates new instance of Pear */
    public Pear()
    {
        log = new it.unive.dsi.pear.log.DefaultLog(Pear.class);
        
        try
        {
            PEAR_HOME.mkdir();
            errLogFile = File.createTempFile("pear_err", ".log", PEAR_HOME);
            outLogFile = new File(PEAR_HOME + "/pear_out.log" );
            errStream = new PrintStream(new FileOutputStream(errLogFile));
            outStream = new PrintStream(new BufferedOutputStream(new FileOutputStream(outLogFile)));
            System.setOut(outStream);
            System.setErr(errStream);
        }
        catch (IOException ioe) { log.error(ioe); }
        
        ToolTipManager.sharedInstance().setDismissDelay(Integer.MAX_VALUE);
        
        this.setFont(DEFAULT_FONT);
        ImageIcon pearIcon =
            new ImageIcon(getClass().getResource("/resources/pear.gif"));
        this.setIconImage(pearIcon.getImage());

        this.initComponents();
        
        errConsole = new JNotifyingTextArea();

        Color col =  (Color)UIManager.getDefaults().get("TextField.inactiveBackground");
        errConsole.setBackground(col);
        errConsole.setEditable(false);
        errConsole.setFont(CONSOLE_FONT);
        errConsole.setMargin(TEXT_AREA_MARGINS);
        
        errScrollPane.setViewportView(errConsole);
        
        tas = new JTextAreaStream(System.out, logConsole);
        errTas = new JTextAreaStream(System.err, errConsole);
        System.setOut(tas);
        System.setErr(errTas);

        try
        {
            builtinRules = conf.getBuiltinValidationRules();
            builtinProtocols = conf.getBuiltinProtocols();
        }
        catch (IOException ioe) { log.error(ioe); }
        
        if (builtinRules.isEmpty() && builtinProtocols.isEmpty())
            toolBarPanel.remove(resourceToolBar);
        
        taggingBox.setFont(DEFAULT_FONT);
        taggingBox.setToolTipText("select a tagging");
        taggingBox.setFocusable(false);
        taggingBox.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                taggingBoxActionPerformed(evt);
            }
        });
        infPanel.add(taggingBox, 0);
        
        screenDim = Toolkit.getDefaultToolkit().getScreenSize();
        Rectangle screen = new Rectangle(0, 0, screenDim.width, screenDim.height);
        
        Dimension size = this.getScaledDim(DIM_FACTOR);
        this.setBounds(this.getCenteredRectangle(size, screen));
        
        size = this.getScaledDim(DIALOG_DIM_FACTOR);
        warningDialog.setSize(size);
        inferenceDialog.setSize(size);
        
        size = this.getScaledDim(MESSAGE_X_DIM_FACTOR);
        size.height = 100;
        exitDialog.setSize(size);
        
        size = this.getScaledDim(MESSAGE_Y_DIM_FACTOR);
        size.height = 100;
        notifyDialog.setSize(size);
        
        pearIcon = new ImageIcon(getClass().getResource("/resources/pera.gif"));
        size.width = pearIcon.getIconWidth() + 400;
        size.height = pearIcon.getIconHeight() + 85;
        aboutDialog.setSize(size);
        aboutText.getCaret().setVisible(false);
        
        this.initTabbedPane();
        
        Reader in = null;
        try
        {
            validationCore = new it.unive.dsi.pear.core.ValidationCore(log);
            validationCore.setFullInfo(true);
            ValidationRules vr = validationCore.getValidationRules();
            inferenceCore = new it.unive.dsi.pear.core.InferenceCore(vr, log);
            inferenceCore.setEqualLimit(PATIENCE_DEFAULT);
            in = validationCore.getDefRulesReader();
        }
        catch (Throwable t)
        {
            log.error(t);
            
            try
            {
                validationCore = new it.unive.dsi.pear.core.ValidationCore(log);
            }
            catch (Throwable e) { e.printStackTrace(System.err); }
        }
        defRules = validationCore.getValidationRules();
        
        if (log.getWarnings(true).length > 0)
            stateBar.setText(" there have been warnings");
        
        log.getWarnings(true);

        this.loadRuleSource(in);
        this.initRuleTree(defRules);
        this.initResourceBrowser();
        
        fileChooser.setFont(DEFAULT_FONT);
        fileChooser.addChoosableFileFilter(ROSPI_FILTER);
        fileChooser.addChoosableFileFilter(VR_FILTER);
        
        for (int i = StepLevel.VALUES.length - 1; i >= 0; --i)
            stepLevelBox.addItem(StepLevel.VALUES[i]);
        curStepLevel = this.getCurStepLevel();
        
        this.setVisible(true);
        
        int xSize = desktopPanel.getDockingWidth();
        int ySize = desktopPanel.getDockingHeight();
        
        int sw = xSize - (int)(JDockableInternalFrame.X_DIM_FACTOR * xSize);
        int sh = ySize - (int)(JDockableInternalFrame.Y_DIM_FACTOR * ySize);
        
        Rectangle bounds = new Rectangle(0, 0, sw, sh);
        console = this.createIFrame("Console", bounds, "/resources/console.gif");
        source = this.createIFrame("Protocol Source", bounds, "/resources/source.gif");
        bounds = new Rectangle(sw + 1, 0, xSize - sw - 1, sh);
        rules = this.createIFrame("Validation Rules", bounds, "/resources/rules.gif");
        bounds = new Rectangle(0, sh + 1, xSize, ySize - sh - 1);
        containers = this.createIFrame("Containers", bounds, "/resources/containers.gif");
        
        rules.getContentPane().add(ruleScrollPane);
        containers.getContentPane().add(containerScrollPane);
        source.getContentPane().add(srcTabbedPane);
        source.setDefaultCloseOperation(DO_NOTHING_ON_CLOSE);
        source.addInternalFrameListener(this);
        
        this.disposeIFrames();
        
        console.getContentPane().add(logScrollPane);

        errConsole.addPropertyChangeListener(new PropertyChangeListener() {
            public void propertyChange(PropertyChangeEvent evt) {
                String name = evt.getPropertyName();
                if (name.equals(JNotifyingTextArea.PROPERTY_NAME))
                    tabbedPane.setSelectedComponent(errScrollPane);
            }
        });
    }
    
    private void initTabbedPane()
    {
        tabbedPane.setFont(DEFAULT_FONT);
        
        tabbedPane.addTab("desktop", desktopPanel);
        tabbedPane.addTab("console", consolePanel);
        tabbedPane.addTab("errors", errScrollPane);
        
        Collection c = conf.getTabbers();
        if (c != null)
        {
            Iterator it = c.iterator();
            while (it.hasNext())
            {
                TabPlugin tp = (TabPlugin)it.next();
                tabbedPane.addTab(tp.getTabTitle(), null, tp.getTab(), tp.getToolTip());
                tp.setLog(log);
            }
        }
    }
    
    /* initResourceBrowser loads and checks for compatibility the resources 
     * specified in the builtin-resources.ini and builtin-protocols.ini 
     * configuration files */
    private void initResourceBrowser()
    {
        new Thread(new Runnable()
        {
            public void run ()
            {
                try
                {
                    String defRulesName = validationCore.getDefRulesName();
                    Collection l = new HashSet(builtinRules.size() + 1);
                    l.add(defRulesName);
                    l.addAll(builtinRules);
                    builtinRules = l;

                    if (!builtinRules.isEmpty())
                    {
                        Validation vc = new it.unive.dsi.pear.core.ValidationCore(log);
                        
                        Iterator it = builtinRules.iterator();
                        while (it.hasNext())
                        {
                            try
                            {
                                String vresource = (String)it.next();
                                InputStream in = conf.getPluginResource(vresource);
                                Reader rd = new BufferedReader(new InputStreamReader(in));
                                ValidationRules vr = vc.parseRules(rd);
                                vc.setValidationRules(vr);
                                String d = vr.getDescription();
                                if (d != null) descMap.put(vresource, d);

                                Iterator jt = builtinProtocols.iterator();
                                while (jt.hasNext()) 
                                {
                                    String presource = (String)jt.next();
                                    in = conf.getPluginResource(presource);
                                    rd = new BufferedReader(new InputStreamReader(in));

                                    try 
                                    {
                                        Protocol p = vc.parseProtocol(rd);
                                        Collection c = p.getAssociatedRules();
                                        if (c == null || c.contains(vresource)) vruleMap.put(vresource, presource);
                                        d = p.getDescription();
                                        if (d != null) descMap.put(presource, d);
                                    }
                                    catch (ParseException pe) {}
                                    catch (SemanticError se) {}
                                }
                            }
                            catch (ParseException pe) { pe.printStackTrace(); }
                        }
                    }
                    
                    builtinRules = new ArrayList(builtinRules);
                    Collections.sort((java.util.List)builtinRules);
                    fillResourceBrowser(builtinRules, rulesName = defRulesName);
                    resourceButton.setEnabled(true);
                }
                catch (Exception e) { e.printStackTrace(); }
            }
        }).start();
    }
    
    private final void fillResourceBrowser(Collection resources)
    {
        this.fillResourceBrowser(resources, rulesName);
    }
    
    private void fillResourceBrowser(Collection resources, String selectedItem)
    {
        java.awt.event.ActionListener[] als = resourceBrowser.getActionListeners();
        java.awt.event.ActionListener al = als.length > 0 ? als[0] : null;
        if (al != null) resourceBrowser.removeActionListener(al);
        
        boolean empty = resources == null || resources.isEmpty();
        Object item = resourceBrowser.getSelectedItem();
        resourceBrowser.removeAllItems();
        resourceBrowser.setEnabled(!empty);
        
        if (empty) resourceBrowser.addItem(NO_RESOURCES);
        else 
        {
            Iterator it = resources.iterator();
            while (it.hasNext()) 
            {
                String name = (String)it.next();
                name = name.substring(name.lastIndexOf('/') + 1, name.length());
                resourceBrowser.addItem(name);
            }
        }
        
        resourceBrowser.setSelectedItem(selectedItem);
        if (al != null) resourceBrowser.addActionListener(al);
    }
    
    private final Dimension getScaledDim(float factor)
    {
        return getScaledDim(factor, factor);
    }
    private final Dimension getScaledDim(float xFactor, float yFactor) 
    {
        int w = (int)(screenDim.width * xFactor);
        int h = (int)(screenDim.height * yFactor);
        return new Dimension(w, h);
    }
    
    private final Rectangle getCenteredRectangle(Dimension size, Rectangle parent) 
    {
        int x = (parent.width - size.width) / 2;
        int y = (parent.height - size.height) / 2;
        return new Rectangle(parent.x + x, parent.y + y, size.width, size.height);
    }
    
    private void showDialog(Dialog d) 
    {
        d.setBounds(this.getCenteredRectangle(d.getSize(), this.getBounds()));
        d.setVisible(true);
    }
    
    /* initRuleTree builds a JTree according to the shape of the validation rule
     * parse tree passed as argument; it also loads the validation rule source
     * into the corresponding text area.
     */
    private void initRuleTree(ValidationRules vr)
    {
        try 
        {
            Class c = this.getClass();
            ImageIcon ii = new ImageIcon(c.getResource("/resources/leaf.gif"));
            renderer.setFont(DEFAULT_FONT);
            renderer.setLeafIcon(ii);
            ii = new ImageIcon(getClass().getResource("/resources/node.gif"));
            renderer.setOpenIcon(ii);
            renderer.setClosedIcon(ii);
            
            ruleTree = new JTree(vr);
            ruleTree.setCellRenderer(renderer);
            ruleScrollPane.setViewportView(ruleTree);
            
            Object[] ruleRootNode = new Object[]{ vr };
            ruleRootPath = new TreePath(ruleRootNode);
            ruleLastPath = ruleRootPath;
        }
        catch (Throwable ex) { log.error(ex); }
    }
    
    private final void loadRuleSource(Reader in) 
    {
        try 
        {
            try { ruleSource.read(in, null); }
            finally { if (in != null) in.close(); }
        }
        catch (IOException ioe) { ioe.printStackTrace(); }
    }
    
    /* setValidationRules sets the current validation rules for the validation
     * and inference core.
     */
    private void setValidationRules(ValidationRules vr, Reader in)
    {
        try
        {
            this.loadRuleSource(in);
            
            if (vr == null)
            {
                in = new BufferedReader(new StringReader(ruleSource.getText()));
                vr = validationCore.parseRules(in);
                log.info("validation rules changed");
                if (log.hasWarnings())
                {
                    stateBar.setText(" there have been warnings");
                    log.getWarnings(true);
                }
            }
            
            validationCore.setValidationRules(vr);
            vr = validationCore.getValidationRules();
            inferenceCore.setValidationRules(vr);
            this.initRuleTree(vr);
        }
        catch (Throwable ex) { log.error(ex); }
    }
    
    /* createIFrame creates and initializes a new JDockableInternalFrame. */
    private JDockableInternalFrame createIFrame(
    String title, Rectangle bounds, String iconSrc) {
        JDockableInternalFrame dif =
        new JDockableInternalFrame(desktopPanel, BorderLayout.WEST);
        dif.setFont(DEFAULT_FONT);
        dif.setTitle(title);
        dif.setFrameIcon(null);
        dif.setPreferredSize(new Dimension(bounds.width, bounds.height));
        dif.setVisible(true);
        dif.setBounds(bounds);
        dif.setFrameIcon(new ImageIcon(getClass().getResource(iconSrc)));
        desktopPanel.addIFrame(dif);
        return dif;
    }
    
    
    /** This method is called from within the constructor to
     * initialize the form.
     * WARNING: Do NOT modify this code. The content of this method is
     * always regenerated by the Form Editor.
     */
    private void initComponents() {//GEN-BEGIN:initComponents
        consolePanel = new javax.swing.JPanel();
        logScrollPane = new javax.swing.JScrollPane();
        logConsole = new javax.swing.JTextArea();
        consoleSouthPanel = new javax.swing.JPanel();
        logSpinnerPanel = new javax.swing.JPanel();
        logLevel = new javax.swing.JLabel();
        logSpinner = new javax.swing.JSpinner();
        clearPanel = new javax.swing.JPanel();
        clearConsole = new javax.swing.JButton();
        ruleScrollPane = new javax.swing.JScrollPane();
        containerScrollPane = new javax.swing.JScrollPane();
        containerContent = new javax.swing.JTextArea();
        errScrollPane = new javax.swing.JScrollPane();
        ruleSrcScrollPane = new javax.swing.JScrollPane();
        ruleSource = new javax.swing.JTextArea();
        srcTabbedPane = new javax.swing.JTabbedPane();
        ruleTabbedPane = new javax.swing.JTabbedPane();
        exitDialog = new javax.swing.JDialog();
        exitLabel = new javax.swing.JLabel();
        exitPanel = new javax.swing.JPanel();
        cancelButton = new javax.swing.JButton();
        exitButton = new javax.swing.JButton();
        warningDialog = new javax.swing.JDialog();
        warnPanel = new javax.swing.JPanel();
        warningBox = new javax.swing.JComboBox();
        warnScrollPane = new javax.swing.JScrollPane();
        sourceArea = new javax.swing.JTextArea();
        warnSouthPanel = new javax.swing.JPanel();
        warnCloseButton = new javax.swing.JButton();
        inferenceDialog = new javax.swing.JDialog();
        infPanel = new javax.swing.JPanel();
        infOptPanel = new javax.swing.JPanel();
        patiencePanel = new javax.swing.JPanel();
        patienceLabel = new javax.swing.JLabel();
        patienceSlider = new javax.swing.JSlider();
        checkBoxPanel = new javax.swing.JPanel();
        infCommitHints = new javax.swing.JCheckBox();
        infReorder = new javax.swing.JCheckBox();
        infScrollPane = new javax.swing.JScrollPane();
        infSourceArea = new javax.swing.JTextArea();
        infSouthPanel = new javax.swing.JPanel();
        buttonPanel = new javax.swing.JPanel();
        infRunButton = new javax.swing.JButton();
        infStopButton = new javax.swing.JButton();
        infSaveButton = new javax.swing.JButton();
        infCloseButton = new javax.swing.JButton();
        aboutDialog = new javax.swing.JDialog();
        aboutNorthPanel = new javax.swing.JPanel();
        aboutLogo = new javax.swing.JLabel();
        aboutPanel = new javax.swing.JPanel();
        aboutText = new javax.swing.JTextArea();
        aboutSouthPanel = new javax.swing.JPanel();
        closeAboutDialog = new javax.swing.JButton();
        notifyDialog = new javax.swing.JDialog();
        notifyPanel = new javax.swing.JPanel();
        notifyButton = new javax.swing.JButton();
        notifyLabel = new javax.swing.JLabel();
        stateBar = new javax.swing.JLabel();
        tabbedPane = new javax.swing.JTabbedPane();
        toolBarPanel = new javax.swing.JPanel();
        fileToolBar = new javax.swing.JToolBar();
        openButton = new javax.swing.JButton();
        saveButton = new javax.swing.JButton();
        closeButton = new javax.swing.JButton();
        resourceToolBar = new javax.swing.JToolBar();
        resourceButton = new javax.swing.JButton();
        resourceBrowser = new javax.swing.JComboBox();
        resourceBrowser.setRenderer(new ToolTipCellRenderer());
        resourceBrowser.addItem("initializing...");
        ruleToolBar = new javax.swing.JToolBar();
        openRulesButton = new javax.swing.JButton();
        defRulesButton = new javax.swing.JButton();
        editToolBar = new javax.swing.JToolBar();
        warningsButton = new javax.swing.JButton();
        editButton = new javax.swing.JButton();
        nextErrorButton = new javax.swing.JButton();
        inferenceButton = new javax.swing.JButton();
        actionsToolBar = new javax.swing.JToolBar();
        prevStepButton = new javax.swing.JButton();
        stopButton = new javax.swing.JButton();
        validateButton = new javax.swing.JButton();
        nextStepButton = new javax.swing.JButton();
        stepLevelBox = new javax.swing.JComboBox();
        logCheckBox = new javax.swing.JCheckBox();
        menuBar = new javax.swing.JMenuBar();
        fileMenu = new javax.swing.JMenu();
        open = new javax.swing.JMenuItem();
        close = new javax.swing.JMenuItem();
        validate = new javax.swing.JMenuItem();
        exit = new javax.swing.JMenuItem();
        viewMenu = new javax.swing.JMenu();
        viewConsole = new javax.swing.JMenuItem();
        viewContainers = new javax.swing.JMenuItem();
        viewSource = new javax.swing.JMenuItem();
        rulesMenu = new javax.swing.JMenu();
        viewRuleFrame = new javax.swing.JMenuItem();
        ruleSeparator = new javax.swing.JSeparator();
        viewRules = new javax.swing.JCheckBoxMenuItem();
        viewTree = new javax.swing.JCheckBoxMenuItem();
        windowMenu = new javax.swing.JMenu();
        dockTo = new javax.swing.JMenu();
        east = new javax.swing.JMenuItem();
        north = new javax.swing.JMenuItem();
        south = new javax.swing.JMenuItem();
        west = new javax.swing.JMenuItem();
        undock = new javax.swing.JMenuItem();
        dockSeparator = new javax.swing.JSeparator();
        dispose = new javax.swing.JMenuItem();
        helpMenu = new javax.swing.JMenu();
        contents = new javax.swing.JMenuItem();
        about = new javax.swing.JMenuItem();

        consolePanel.setLayout(new java.awt.BorderLayout());

        consolePanel.setFont(DEFAULT_FONT);
        logScrollPane.setFont(DEFAULT_FONT);
        logConsole.setBackground(javax.swing.UIManager.getDefaults().getColor("TextField.inactiveBackground"));
        logConsole.setEditable(false);
        logConsole.setFont(CONSOLE_FONT);
        logConsole.setMargin(TEXT_AREA_MARGINS);
        logScrollPane.setViewportView(logConsole);

        consolePanel.add(logScrollPane, java.awt.BorderLayout.CENTER);

        consoleSouthPanel.setLayout(new java.awt.GridLayout(1, 0));

        logSpinnerPanel.setLayout(new java.awt.FlowLayout(java.awt.FlowLayout.LEFT, 5, 2));

        logSpinnerPanel.setFont(DEFAULT_FONT);
        logLevel.setFont(DEFAULT_FONT);
        logLevel.setText("Log level:");
        logSpinnerPanel.add(logLevel);

        logSpinner.setFont(DEFAULT_FONT);
        logSpinner.setModel(new SpinnerNumberModel(
            new Integer(Log.DEFAULT_LOG_LEVEL),
            new Integer(Log.MUTE),
            new Integer(Log.DEBUG),
            new Integer(1)));
        logSpinner.setPreferredSize(new java.awt.Dimension(30, 20));
        logSpinner.addChangeListener(new javax.swing.event.ChangeListener() {
            public void stateChanged(javax.swing.event.ChangeEvent evt) {
                logSpinnerStateChanged(evt);
            }
        });

        logSpinnerPanel.add(logSpinner);

        consoleSouthPanel.add(logSpinnerPanel);

        clearPanel.setLayout(new java.awt.FlowLayout(java.awt.FlowLayout.RIGHT, 0, 0));

        clearPanel.setFont(DEFAULT_FONT);
        clearConsole.setFont(DEFAULT_FONT);
        clearConsole.setText("clear");
        clearConsole.setFocusable(false);
        clearConsole.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                clearConsoleActionPerformed(evt);
            }
        });

        clearPanel.add(clearConsole);

        consoleSouthPanel.add(clearPanel);

        consolePanel.add(consoleSouthPanel, java.awt.BorderLayout.SOUTH);

        ruleScrollPane.setFont(DEFAULT_FONT);
        containerScrollPane.setFont(DEFAULT_FONT);
        containerContent.setEditable(false);
        containerContent.setFont(DEFAULT_FONT);
        containerContent.setForeground(new java.awt.Color(0, 0, 153));
        containerContent.setMargin(new java.awt.Insets(2, 5, 2, 5));
        containerScrollPane.setViewportView(containerContent);

        errScrollPane.setFont(DEFAULT_FONT);
        ruleSrcScrollPane.setFont(DEFAULT_FONT);
        ruleSource.setEditable(false);
        ruleSource.setFont(CONSOLE_FONT);
        ruleSource.setMargin(TEXT_AREA_MARGINS);
        ruleSrcScrollPane.setViewportView(ruleSource);

        srcTabbedPane.setTabLayoutPolicy(javax.swing.JTabbedPane.SCROLL_TAB_LAYOUT);
        srcTabbedPane.setTabPlacement(javax.swing.JTabbedPane.BOTTOM);
        srcTabbedPane.setFocusable(false);
        srcTabbedPane.setFont(DEFAULT_FONT);
        srcTabbedPane.addChangeListener(new javax.swing.event.ChangeListener() {
            public void stateChanged(javax.swing.event.ChangeEvent evt) {
                srcTabbedPaneStateChanged(evt);
            }
        });

        ruleTabbedPane.setTabPlacement(javax.swing.JTabbedPane.BOTTOM);
        ruleTabbedPane.setFocusable(false);
        ruleTabbedPane.setFont(DEFAULT_FONT);
        exitDialog.setTitle("Modified sources");
        exitDialog.setFont(DEFAULT_FONT);
        exitDialog.setModal(true);
        exitDialog.setName("exitDialog");
        exitDialog.setResizable(false);
        exitLabel.setFont(DEFAULT_FONT);
        exitLabel.setHorizontalAlignment(javax.swing.SwingConstants.CENTER);
        exitLabel.setText("There are modified sources: you should close them before exiting.");
        exitDialog.getContentPane().add(exitLabel, java.awt.BorderLayout.CENTER);

        exitPanel.setFont(DEFAULT_FONT);
        cancelButton.setFont(DEFAULT_FONT
        );
        cancelButton.setText("cancel");
        cancelButton.setFocusable(false);
        cancelButton.setMargin(BUTTON_MARGINS);
        cancelButton.setPreferredSize(DIALOG_BUTTON_SIZE);
        cancelButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                dialogCloseButtonActionPerformed(evt);
            }
        });

        exitPanel.add(cancelButton);

        exitButton.setFont(DEFAULT_FONT);
        exitButton.setText("exit");
        exitButton.setFocusable(false);
        exitButton.setMargin(BUTTON_MARGINS);
        exitButton.setPreferredSize(DIALOG_BUTTON_SIZE);
        exitButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                exitButtonActionPerformed(evt);
            }
        });

        exitPanel.add(exitButton);

        exitDialog.getContentPane().add(exitPanel, java.awt.BorderLayout.SOUTH);

        warningDialog.setTitle("Warnings");
        warningDialog.setFont(DEFAULT_FONT);
        warningDialog.setName("Warnings");
        warnPanel.setLayout(new java.awt.FlowLayout(java.awt.FlowLayout.RIGHT, 0, 1));

        warnPanel.setFont(DEFAULT_FONT);
        warningBox.setFont(DEFAULT_FONT);
        warningBox.setFocusable(false);
        warningBox.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                warningBoxActionPerformed(evt);
            }
        });

        warnPanel.add(warningBox);

        warningDialog.getContentPane().add(warnPanel, java.awt.BorderLayout.NORTH);

        warnScrollPane.setFont(DEFAULT_FONT);
        sourceArea.setBackground(javax.swing.UIManager.getDefaults().getColor("TextField.inactiveBackground"));
        sourceArea.setEditable(false);
        sourceArea.setFont(CONSOLE_FONT);
        sourceArea.setMargin(TEXT_AREA_MARGINS);
        warnScrollPane.setViewportView(sourceArea);

        warningDialog.getContentPane().add(warnScrollPane, java.awt.BorderLayout.CENTER);

        warnSouthPanel.setLayout(new java.awt.FlowLayout(java.awt.FlowLayout.RIGHT, 0, 1));

        warnSouthPanel.setFont(DEFAULT_FONT);
        warnCloseButton.setFont(DEFAULT_FONT);
        warnCloseButton.setText("close");
        warnCloseButton.setFocusable(false);
        warnCloseButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                dialogCloseButtonActionPerformed(evt);
            }
        });

        warnSouthPanel.add(warnCloseButton);

        warningDialog.getContentPane().add(warnSouthPanel, java.awt.BorderLayout.SOUTH);

        inferenceDialog.setTitle("Tag Inference");
        inferenceDialog.setFont(DEFAULT_FONT);
        inferenceDialog.setName("Tag Inference");
        infPanel.setLayout(new java.awt.FlowLayout(java.awt.FlowLayout.LEFT, 1, 1));

        infPanel.setFont(DEFAULT_FONT);
        infOptPanel.setLayout(new java.awt.FlowLayout(java.awt.FlowLayout.LEFT, 20, 1));

        infOptPanel.setFont(DEFAULT_FONT);
        patiencePanel.setLayout(new java.awt.FlowLayout(java.awt.FlowLayout.CENTER, 2, 0));

        patienceLabel.setFont(DEFAULT_FONT);
        patienceLabel.setText("patience");
        patiencePanel.add(patienceLabel);

        patienceSlider.setFont(DEFAULT_FONT);
        patienceSlider.setMajorTickSpacing(20);
        patienceSlider.setMaximum(PATIENCE_MAX_LIMIT);
        patienceSlider.setMinimum(PATIENCE_MIN_LIMIT);
        patienceSlider.setMinorTickSpacing(5);
        patienceSlider.setSnapToTicks(true);
        patienceSlider.setToolTipText("set the maximum time to wait after the first tagging is found, before terminating inference process");
        patienceSlider.setValue(PATIENCE_DEFAULT);
        patienceSlider.addChangeListener(new javax.swing.event.ChangeListener() {
            public void stateChanged(javax.swing.event.ChangeEvent evt) {
                patienceSliderStateChanged(evt);
            }
        });

        patiencePanel.add(patienceSlider);

        infOptPanel.add(patiencePanel);

        checkBoxPanel.setLayout(new java.awt.FlowLayout(java.awt.FlowLayout.CENTER, 0, 0));

        infCommitHints.setFont(DEFAULT_FONT);
        infCommitHints.setText("using commit hints");
        infCommitHints.setToolTipText("enable (disable) commit hints use: it may slow down the process");
        infCommitHints.setFocusable(false);
        infCommitHints.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                infCommitHintsActionPerformed(evt);
            }
        });

        checkBoxPanel.add(infCommitHints);

        infReorder.setFont(DEFAULT_FONT);
        infReorder.setText("reorder protocol");
        infReorder.setToolTipText("enable (disable) protocol reordering (it improves performances)");
        infReorder.setFocusable(false);
        infReorder.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                infReorderActionPerformed(evt);
            }
        });

        checkBoxPanel.add(infReorder);

        infOptPanel.add(checkBoxPanel);

        infPanel.add(infOptPanel);

        inferenceDialog.getContentPane().add(infPanel, java.awt.BorderLayout.NORTH);

        infScrollPane.setFont(DEFAULT_FONT);
        infSourceArea.setBackground(javax.swing.UIManager.getDefaults().getColor("TextField.inactiveBackground"));
        infSourceArea.setEditable(false);
        infSourceArea.setFont(CONSOLE_FONT);
        infSourceArea.setMargin(TEXT_AREA_MARGINS);
        infScrollPane.setViewportView(infSourceArea);

        inferenceDialog.getContentPane().add(infScrollPane, java.awt.BorderLayout.CENTER);

        infSouthPanel.setLayout(new java.awt.FlowLayout(java.awt.FlowLayout.RIGHT, 0, 1));

        buttonPanel.setLayout(new java.awt.FlowLayout(java.awt.FlowLayout.LEFT, 5, 0));

        infRunButton.setFont(DEFAULT_FONT);
        infRunButton.setText("run");
        infRunButton.setToolTipText("run tagging inference");
        infRunButton.setFocusable(false);
        infRunButton.setPreferredSize(DIALOG_BUTTON_SIZE);
        infRunButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                infRunButtonActionPerformed(evt);
            }
        });

        buttonPanel.add(infRunButton);

        infStopButton.setFont(DEFAULT_FONT);
        infStopButton.setText("stop");
        infStopButton.setToolTipText("stop tagging inference");
        infStopButton.setFocusable(false);
        infStopButton.setPreferredSize(DIALOG_BUTTON_SIZE);
        infStopButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                infStopButtonActionPerformed(evt);
            }
        });

        buttonPanel.add(infStopButton);

        infSaveButton.setFont(DEFAULT_FONT);
        infSaveButton.setText("save");
        infSaveButton.setToolTipText("save current tagging as...");
        infSaveButton.setFocusable(false);
        infSaveButton.setPreferredSize(DIALOG_BUTTON_SIZE);
        infSaveButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                infSaveButtonActionPerformed(evt);
            }
        });

        buttonPanel.add(infSaveButton);

        infSouthPanel.add(buttonPanel);

        infCloseButton.setFont(DEFAULT_FONT);
        infCloseButton.setText("close");
        infCloseButton.setToolTipText("close dialog");
        infCloseButton.setFocusable(false);
        infCloseButton.setPreferredSize(DIALOG_BUTTON_SIZE);
        infCloseButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                dialogCloseButtonActionPerformed(evt);
            }
        });

        infSouthPanel.add(infCloseButton);

        inferenceDialog.getContentPane().add(infSouthPanel, java.awt.BorderLayout.SOUTH);

        aboutDialog.setTitle("PEAR");
        aboutDialog.setFont(DEFAULT_FONT);
        aboutDialog.setModal(true);
        aboutDialog.setName("exitDialog");
        aboutDialog.setResizable(false);
        aboutDialog.getContentPane().add(aboutNorthPanel, java.awt.BorderLayout.NORTH);

        aboutLogo.setFont(DEFAULT_FONT);
        aboutLogo.setHorizontalAlignment(javax.swing.SwingConstants.RIGHT);
        aboutLogo.setIcon(new javax.swing.ImageIcon(getClass().getResource("/resources/pera.gif")));
        aboutLogo.setText("  ");
        aboutLogo.setHorizontalTextPosition(javax.swing.SwingConstants.LEADING);
        aboutDialog.getContentPane().add(aboutLogo, java.awt.BorderLayout.WEST);

        aboutPanel.setLayout(new java.awt.GridBagLayout());

        aboutText.setBackground(javax.swing.UIManager.getDefaults().getColor("Label.background"));
        aboutText.setEditable(false);
        aboutText.setFont(DEFAULT_FONT);
        aboutText.setText("Protocol Extensible AnalyzeR - also known as ProtOK\ndeveloped by Francesco Placella\nUniversit\u00e0 di Ca' Foscari - Dipartimento d'Informatica, Venezia");
        aboutText.setBorder(null);
        aboutPanel.add(aboutText, new java.awt.GridBagConstraints());

        aboutDialog.getContentPane().add(aboutPanel, java.awt.BorderLayout.CENTER);

        aboutSouthPanel.setLayout(new java.awt.FlowLayout(java.awt.FlowLayout.CENTER, 0, 5));

        closeAboutDialog.setFont(DEFAULT_FONT
        );
        closeAboutDialog.setText("ok");
        closeAboutDialog.setFocusable(false);
        closeAboutDialog.setPreferredSize(DIALOG_BUTTON_SIZE);
        closeAboutDialog.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                dialogCloseButtonActionPerformed(evt);
            }
        });

        aboutSouthPanel.add(closeAboutDialog);

        aboutDialog.getContentPane().add(aboutSouthPanel, java.awt.BorderLayout.SOUTH);

        notifyDialog.setFont(DEFAULT_FONT);
        notifyDialog.setModal(true);
        notifyDialog.setName("notifyDialog");
        notifyPanel.setLayout(new java.awt.FlowLayout(java.awt.FlowLayout.CENTER, 0, 2));

        notifyPanel.setFont(DEFAULT_FONT);
        notifyButton.setFont(DEFAULT_FONT);
        notifyButton.setText("ok");
        notifyButton.setFocusable(false);
        notifyButton.setPreferredSize(DIALOG_BUTTON_SIZE);
        notifyButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                notifyButtonActionPerformed(evt);
            }
        });

        notifyPanel.add(notifyButton);

        notifyDialog.getContentPane().add(notifyPanel, java.awt.BorderLayout.SOUTH);

        notifyLabel.setFont(DEFAULT_FONT);
        notifyLabel.setHorizontalAlignment(javax.swing.SwingConstants.CENTER);
        notifyLabel.setBorder(new javax.swing.border.EmptyBorder(new java.awt.Insets(10, 20, 10, 20)));
        notifyLabel.setHorizontalTextPosition(javax.swing.SwingConstants.CENTER);
        notifyDialog.getContentPane().add(notifyLabel, java.awt.BorderLayout.CENTER);

        setDefaultCloseOperation(javax.swing.WindowConstants.DO_NOTHING_ON_CLOSE);
        setTitle("PEAR");
        setFont(DEFAULT_FONT);
        setName("PEAR");
        addWindowListener(new java.awt.event.WindowAdapter() {
            public void windowClosing(java.awt.event.WindowEvent evt) {
                exitForm(evt);
            }
        });

        stateBar.setFont(DEFAULT_FONT);
        stateBar.setHorizontalAlignment(javax.swing.SwingConstants.LEFT);
        stateBar.setBorder(new javax.swing.border.EtchedBorder());
        stateBar.setFocusable(false);
        stateBar.setMaximumSize(new java.awt.Dimension(24, 20));
        stateBar.setMinimumSize(new java.awt.Dimension(24, 20));
        stateBar.setPreferredSize(new java.awt.Dimension(24, 20));
        stateBar.setRequestFocusEnabled(false);
        stateBar.setHorizontalTextPosition(javax.swing.SwingConstants.LEFT);
        getContentPane().add(stateBar, java.awt.BorderLayout.SOUTH);

        tabbedPane.setTabPlacement(javax.swing.JTabbedPane.BOTTOM);
        tabbedPane.setFocusable(false);
        tabbedPane.setFont(DEFAULT_FONT);
        tabbedPane.setMinimumSize(new java.awt.Dimension(16, 16));
        tabbedPane.setPreferredSize(new java.awt.Dimension(16, 16));
        tabbedPane.addChangeListener(new javax.swing.event.ChangeListener() {
            public void stateChanged(javax.swing.event.ChangeEvent evt) {
                tabSelected(evt);
            }
        });

        getContentPane().add(tabbedPane, java.awt.BorderLayout.CENTER);

        toolBarPanel.setLayout(new java.awt.FlowLayout(java.awt.FlowLayout.LEFT, 2, 0));

        toolBarPanel.setBorder(new javax.swing.border.EtchedBorder());
        toolBarPanel.setFocusable(false);
        toolBarPanel.setFont(DEFAULT_FONT);
        fileToolBar.setRollover(true);
        fileToolBar.setFont(DEFAULT_FONT);
        openButton.setFont(DEFAULT_FONT);
        openButton.setIcon(new javax.swing.ImageIcon(getClass().getResource("/resources/open.gif")));
        openButton.setToolTipText("open file");
        openButton.setFocusable(false);
        openButton.setMargin(BUTTON_MARGINS);
        openButton.setMaximumSize(BUTTON_SIZE);
        openButton.setMinimumSize(BUTTON_SIZE);
        openButton.setPreferredSize(BUTTON_SIZE);
        openButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                openActionPerformed(evt);
            }
        });

        fileToolBar.add(openButton);

        saveButton.setFont(DEFAULT_FONT);
        saveButton.setIcon(new javax.swing.ImageIcon(getClass().getResource("/resources/save.gif")));
        saveButton.setToolTipText("save file");
        saveButton.setFocusable(false);
        saveButton.setMargin(BUTTON_MARGINS);
        saveButton.setMaximumSize(BUTTON_SIZE);
        saveButton.setMinimumSize(BUTTON_SIZE);
        saveButton.setPreferredSize(BUTTON_SIZE);
        saveButton.setEnabled(false);
        saveButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                saveButtonActionPerformed(evt);
            }
        });

        fileToolBar.add(saveButton);

        closeButton.setFont(DEFAULT_FONT);
        closeButton.setIcon(new javax.swing.ImageIcon(getClass().getResource("/resources/close.gif")));
        closeButton.setToolTipText("close file");
        closeButton.setFocusable(false);
        closeButton.setMargin(BUTTON_MARGINS);
        closeButton.setMaximumSize(BUTTON_SIZE);
        closeButton.setMinimumSize(BUTTON_SIZE);
        closeButton.setPreferredSize(BUTTON_SIZE);
        closeButton.setEnabled(false);
        closeButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                closeActionPerformed(evt);
            }
        });

        fileToolBar.add(closeButton);

        toolBarPanel.add(fileToolBar);

        resourceToolBar.setRollover(true);
        resourceButton.setFont(DEFAULT_FONT);
        resourceButton.setIcon(openRulesIcon);
        resourceButton.setToolTipText("cycle through resource types");
        resourceButton.setFocusable(false);
        resourceButton.setMargin(BUTTON_MARGINS);
        resourceButton.setMaximumSize(BUTTON_SIZE);
        resourceButton.setMinimumSize(BUTTON_SIZE);
        resourceButton.setPreferredSize(BUTTON_SIZE);
        resourceButton.setEnabled(false);
        resourceButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                resourceButtonActionPerformed(evt);
            }
        });

        resourceToolBar.add(resourceButton);

        resourceBrowser.setFont(DEFAULT_FONT);
        resourceBrowser.setToolTipText("the buit-in resource browser");
        resourceBrowser.setFocusable(false);
        resourceBrowser.setPreferredSize(new Dimension(150, BUTTON_SIZE.height));
        resourceBrowser.setEnabled(false);
        resourceBrowser.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                resourceBrowserActionPerformed(evt);
            }
        });

        resourceToolBar.add(resourceBrowser);

        toolBarPanel.add(resourceToolBar);

        ruleToolBar.setRollover(true);
        ruleToolBar.setFont(DEFAULT_FONT);
        openRulesButton.setFont(DEFAULT_FONT);
        openRulesButton.setIcon(new javax.swing.ImageIcon(getClass().getResource("/resources/openrules.gif")));
        openRulesButton.setToolTipText("select a file defining the validation rules");
        openRulesButton.setFocusable(false);
        openRulesButton.setMargin(BUTTON_MARGINS);
        openRulesButton.setMaximumSize(BUTTON_SIZE);
        openRulesButton.setMinimumSize(BUTTON_SIZE);
        openRulesButton.setPreferredSize(BUTTON_SIZE);
        openRulesButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                openRulesButtonActionPerformed(evt);
            }
        });

        ruleToolBar.add(openRulesButton);

        defRulesButton.setFont(DEFAULT_FONT);
        defRulesButton.setIcon(new javax.swing.ImageIcon(getClass().getResource("/resources/defrules.gif")));
        defRulesButton.setToolTipText("select default validation rules");
        defRulesButton.setFocusable(false);
        defRulesButton.setMargin(BUTTON_MARGINS);
        defRulesButton.setMaximumSize(BUTTON_SIZE);
        defRulesButton.setMinimumSize(BUTTON_SIZE);
        defRulesButton.setPreferredSize(BUTTON_SIZE);
        defRulesButton.setEnabled(false);
        defRulesButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                defRulesButtonActionPerformed(evt);
            }
        });

        ruleToolBar.add(defRulesButton);

        toolBarPanel.add(ruleToolBar);

        editToolBar.setRollover(true);
        editToolBar.setFont(DEFAULT_FONT);
        warningsButton.setFont(DEFAULT_FONT);
        warningsButton.setIcon(new javax.swing.ImageIcon(getClass().getResource("/resources/warnings.gif")));
        warningsButton.setToolTipText("display warning locations");
        warningsButton.setFocusable(false);
        warningsButton.setMargin(BUTTON_MARGINS);
        warningsButton.setMaximumSize(BUTTON_SIZE);
        warningsButton.setMinimumSize(BUTTON_SIZE);
        warningsButton.setPreferredSize(BUTTON_SIZE);
        warningsButton.setEnabled(false);
        warningsButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                warningsButtonActionPerformed(evt);
            }
        });

        editToolBar.add(warningsButton);

        editButton.setFont(DEFAULT_FONT);
        editButton.setIcon(new javax.swing.ImageIcon(getClass().getResource("/resources/edit.gif")));
        editButton.setToolTipText("switch beetween source editing mode and source displaying mode");
        editButton.setFocusable(false);
        editButton.setMargin(BUTTON_MARGINS);
        editButton.setMaximumSize(BUTTON_SIZE);
        editButton.setMinimumSize(BUTTON_SIZE);
        editButton.setPreferredSize(BUTTON_SIZE);
        editButton.setEnabled(false);
        editButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                editButtonActionPerformed(evt);
            }
        });

        editToolBar.add(editButton);

        nextErrorButton.setFont(DEFAULT_FONT);
        nextErrorButton.setIcon(new javax.swing.ImageIcon(getClass().getResource("/resources/nexterror.gif")));
        nextErrorButton.setToolTipText("select next error");
        nextErrorButton.setFocusable(false);
        nextErrorButton.setMargin(BUTTON_MARGINS);
        nextErrorButton.setMaximumSize(BUTTON_SIZE);
        nextErrorButton.setMinimumSize(BUTTON_SIZE);
        nextErrorButton.setPreferredSize(BUTTON_SIZE);
        nextErrorButton.setEnabled(false);
        nextErrorButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                nextErrorButtonActionPerformed(evt);
            }
        });

        editToolBar.add(nextErrorButton);

        inferenceButton.setFont(DEFAULT_FONT);
        inferenceButton.setIcon(new javax.swing.ImageIcon(getClass().getResource("/resources/inference.gif")));
        inferenceButton.setToolTipText("display the tag inference dialog");
        inferenceButton.setFocusable(false);
        inferenceButton.setMargin(BUTTON_MARGINS);
        inferenceButton.setMaximumSize(BUTTON_SIZE);
        inferenceButton.setMinimumSize(BUTTON_SIZE);
        inferenceButton.setPreferredSize(BUTTON_SIZE);
        inferenceButton.setEnabled(false);
        inferenceButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                inferenceButtonActionPerformed(evt);
            }
        });

        editToolBar.add(inferenceButton);

        toolBarPanel.add(editToolBar);

        actionsToolBar.setRollover(true);
        actionsToolBar.setFont(DEFAULT_FONT);
        prevStepButton.setFont(DEFAULT_FONT);
        prevStepButton.setIcon(new javax.swing.ImageIcon(getClass().getResource("/resources/prevstep.gif")));
        prevStepButton.setToolTipText("go back to previous instruction");
        prevStepButton.setFocusable(false);
        prevStepButton.setMargin(BUTTON_MARGINS);
        prevStepButton.setMaximumSize(BUTTON_SIZE);
        prevStepButton.setMinimumSize(BUTTON_SIZE);
        prevStepButton.setPreferredSize(BUTTON_SIZE);
        prevStepButton.setEnabled(false);
        prevStepButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                prevStepButtonActionPerformed(evt);
            }
        });

        actionsToolBar.add(prevStepButton);

        stopButton.setFont(DEFAULT_FONT);
        stopButton.setIcon(new javax.swing.ImageIcon(getClass().getResource("/resources/stop.gif")));
        stopButton.setToolTipText("stop validation");
        stopButton.setFocusable(false);
        stopButton.setMargin(BUTTON_MARGINS);
        stopButton.setMaximumSize(BUTTON_SIZE);
        stopButton.setMinimumSize(BUTTON_SIZE);
        stopButton.setPreferredSize(BUTTON_SIZE);
        stopButton.setEnabled(false);
        stopButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                stopButtonActionPerformed(evt);
            }
        });

        actionsToolBar.add(stopButton);

        validateButton.setFont(DEFAULT_FONT);
        validateButton.setIcon(new javax.swing.ImageIcon(getClass().getResource("/resources/validate.gif")));
        validateButton.setToolTipText("run validation");
        validateButton.setFocusable(false);
        validateButton.setMargin(BUTTON_MARGINS);
        validateButton.setMaximumSize(BUTTON_SIZE);
        validateButton.setMinimumSize(BUTTON_SIZE);
        validateButton.setPreferredSize(BUTTON_SIZE);
        validateButton.setEnabled(false);
        validateButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                validateActionPerformed(evt);
            }
        });

        actionsToolBar.add(validateButton);

        nextStepButton.setFont(DEFAULT_FONT);
        nextStepButton.setIcon(new javax.swing.ImageIcon(getClass().getResource("/resources/nextstep.gif")));
        nextStepButton.setToolTipText("next validation step");
        nextStepButton.setFocusable(false);
        nextStepButton.setMargin(BUTTON_MARGINS);
        nextStepButton.setMaximumSize(BUTTON_SIZE);
        nextStepButton.setMinimumSize(BUTTON_SIZE);
        nextStepButton.setPreferredSize(BUTTON_SIZE);
        nextStepButton.setEnabled(false);
        nextStepButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                nextStepButtonActionPerformed(evt);
            }
        });

        actionsToolBar.add(nextStepButton);

        stepLevelBox.setFont(DEFAULT_FONT);
        stepLevelBox.setToolTipText("select current step granularity");
        stepLevelBox.setFocusable(false);
        stepLevelBox.setEnabled(false);
        stepLevelBox.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                stepLevelBoxActionPerformed(evt);
            }
        });

        actionsToolBar.add(stepLevelBox);

        logCheckBox.setFont(DEFAULT_FONT);
        logCheckBox.setSelected(true);
        logCheckBox.setText("log level");
        logCheckBox.setToolTipText("automatically set appropriate log level according to the selected step granularity");
        logCheckBox.setFocusable(false);
        logCheckBox.setEnabled(false);
        logCheckBox.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                logCheckBoxActionPerformed(evt);
            }
        });

        actionsToolBar.add(logCheckBox);

        toolBarPanel.add(actionsToolBar);

        getContentPane().add(toolBarPanel, java.awt.BorderLayout.NORTH);

        menuBar.setBorder(null);
        menuBar.setFont(DEFAULT_FONT);
        fileMenu.setMnemonic('f');
        fileMenu.setText("File");
        fileMenu.setFont(DEFAULT_FONT);
        open.setFont(DEFAULT_FONT);
        open.setMnemonic('o');
        open.setText("Open");
        open.setToolTipText("");
        open.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                openActionPerformed(evt);
            }
        });

        fileMenu.add(open);

        close.setFont(DEFAULT_FONT);
        close.setMnemonic('c');
        close.setText("Close");
        close.setEnabled(false);
        close.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                closeActionPerformed(evt);
            }
        });

        fileMenu.add(close);

        validate.setFont(DEFAULT_FONT);
        validate.setMnemonic('v');
        validate.setText("Validate");
        validate.setEnabled(false);
        validate.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                validateActionPerformed(evt);
            }
        });

        fileMenu.add(validate);

        exit.setFont(DEFAULT_FONT);
        exit.setMnemonic('x');
        exit.setText("Exit");
        exit.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                exitActionPerformed(evt);
            }
        });

        fileMenu.add(exit);

        menuBar.add(fileMenu);

        viewMenu.setMnemonic('v');
        viewMenu.setText("View");
        viewMenu.setFont(DEFAULT_FONT);
        viewConsole.setFont(DEFAULT_FONT);
        viewConsole.setIcon(new javax.swing.ImageIcon(getClass().getResource("/resources/console.gif")));
        viewConsole.setMnemonic('n');
        viewConsole.setText("Console");
        viewConsole.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                viewConsoleActionPerformed(evt);
            }
        });

        viewMenu.add(viewConsole);

        viewContainers.setFont(DEFAULT_FONT);
        viewContainers.setIcon(new javax.swing.ImageIcon(getClass().getResource("/resources/containers.gif")));
        viewContainers.setMnemonic('a');
        viewContainers.setText("Containers");
        viewContainers.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                viewContainersActionPerformed(evt);
            }
        });

        viewMenu.add(viewContainers);

        viewSource.setFont(DEFAULT_FONT);
        viewSource.setIcon(new javax.swing.ImageIcon(getClass().getResource("/resources/source.gif")));
        viewSource.setMnemonic('p');
        viewSource.setText("Protocol Source");
        viewSource.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                viewSourceActionPerformed(evt);
            }
        });

        viewMenu.add(viewSource);

        rulesMenu.setIcon(new javax.swing.ImageIcon(getClass().getResource("/resources/rules.gif")));
        rulesMenu.setText("Validation Rules");
        rulesMenu.setFont(DEFAULT_FONT);
        viewRuleFrame.setFont(DEFAULT_FONT);
        viewRuleFrame.setMnemonic('w');
        viewRuleFrame.setText("Window");
        viewRuleFrame.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                viewRuleFrameActionPerformed(evt);
            }
        });

        rulesMenu.add(viewRuleFrame);

        rulesMenu.add(ruleSeparator);

        viewRules.setFont(DEFAULT_FONT);
        viewRules.setMnemonic('s');
        viewRules.setText("Source");
        viewRules.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                viewRulesActionPerformed(evt);
            }
        });

        rulesMenu.add(viewRules);

        viewTree.setFont(DEFAULT_FONT);
        viewTree.setMnemonic('t');
        viewTree.setSelected(true);
        viewTree.setText("Tree");
        viewTree.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                viewTreeActionPerformed(evt);
            }
        });

        rulesMenu.add(viewTree);

        viewMenu.add(rulesMenu);

        menuBar.add(viewMenu);

        windowMenu.setMnemonic('w');
        windowMenu.setText("Window");
        windowMenu.setFont(DEFAULT_FONT);
        windowMenu.addMouseListener(new java.awt.event.MouseAdapter() {
            public void mousePressed(java.awt.event.MouseEvent evt) {
                windowMenuMousePressed(evt);
            }
        });

        dockTo.setMnemonic('d');
        dockTo.setText("Dock to");
        dockTo.setFont(DEFAULT_FONT);
        east.setFont(DEFAULT_FONT);
        east.setMnemonic('r');
        east.setText("Right");
        east.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                eastActionPerformed(evt);
            }
        });

        dockTo.add(east);

        north.setFont(DEFAULT_FONT);
        north.setMnemonic('t');
        north.setText("Top");
        north.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                northActionPerformed(evt);
            }
        });

        dockTo.add(north);

        south.setFont(DEFAULT_FONT);
        south.setMnemonic('b');
        south.setText("Bottom");
        south.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                southActionPerformed(evt);
            }
        });

        dockTo.add(south);

        west.setFont(DEFAULT_FONT);
        west.setMnemonic('l');
        west.setText("Left");
        west.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                westActionPerformed(evt);
            }
        });

        dockTo.add(west);

        windowMenu.add(dockTo);

        undock.setFont(DEFAULT_FONT);
        undock.setMnemonic('u');
        undock.setText("Undock");
        undock.setEnabled(false);
        undock.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                undockActionPerformed(evt);
            }
        });

        windowMenu.add(undock);

        windowMenu.add(dockSeparator);

        dispose.setFont(DEFAULT_FONT);
        dispose.setMnemonic('i');
        dispose.setText("Dispose");
        dispose.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                disposeActionPerformed(evt);
            }
        });

        windowMenu.add(dispose);

        menuBar.add(windowMenu);

        helpMenu.setMnemonic('h');
        helpMenu.setText("Help");
        helpMenu.setFont(DEFAULT_FONT);
        contents.setFont(DEFAULT_FONT);
        contents.setMnemonic('c');
        contents.setText("Contents");
        contents.setEnabled(false);
        helpMenu.add(contents);

        about.setFont(DEFAULT_FONT);
        about.setMnemonic('a');
        about.setText("About");
        about.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                aboutActionPerformed(evt);
            }
        });

        helpMenu.add(about);

        menuBar.add(helpMenu);

        setJMenuBar(menuBar);

        pack();
    }//GEN-END:initComponents

    private void setResourceBrowser(boolean rulesEnabled)
    {
        resourceButton.setIcon(rulesEnabled ? openRulesIcon : openIcon);
        
        if (rulesEnabled)
            this.fillResourceBrowser(builtinRules);
        else
        {
            Collection c = vruleMap.isEmpty() ?
                builtinProtocols :
                (Collection)vruleMap.get(resourceBrowser.getSelectedItem());
            this.fillResourceBrowser(c);
        }
        
        boolean b = openedFiles.size() == 0 || !rulesEnabled;
        String tip = b ? "the built-in resource browser" : "close all protocols to enable the resource browser";
        resourceBrowser.setToolTipText(tip);
        resourceBrowser.setEnabled(b);
    }
    
    private void resourceButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_resourceButtonActionPerformed

        this.setResourceBrowser(rulesEnabled = !rulesEnabled);

    }//GEN-LAST:event_resourceButtonActionPerformed

    private void resourceBrowserActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_resourceBrowserActionPerformed
        
        final String name = rulesName = (String)resourceBrowser.getSelectedItem();
        final InputStream in = conf.getPluginResource(name);
        
        if (in != null)
        {
            new Thread(new Runnable()
            {
                public void run()
                {
                    Reader rd = new BufferedReader(new InputStreamReader(in));
                    if (Configuration.isRuleSystem(name)) openRules(rd, TMP_PREFIX + name);
                    else addProtocolSource(rd, name);
                }
            }).start();
        }
    }//GEN-LAST:event_resourceBrowserActionPerformed

    private void notifyButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_notifyButtonActionPerformed
        notifyDialog.setVisible(false);
    }//GEN-LAST:event_notifyButtonActionPerformed

    private void aboutActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_aboutActionPerformed
        this.showDialog(aboutDialog);
    }//GEN-LAST:event_aboutActionPerformed

    /* patienceSliderStateChanged updtaes the equal limit of the inference core.
     */
    private void patienceSliderStateChanged(javax.swing.event.ChangeEvent evt) {//GEN-FIRST:event_patienceSliderStateChanged
        if (!patienceSlider.getValueIsAdjusting())
            inferenceCore.setEqualLimit(patienceSlider.getValue());
    }//GEN-LAST:event_patienceSliderStateChanged

    private void clearConsoleActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_clearConsoleActionPerformed
        tas.clear();
    }//GEN-LAST:event_clearConsoleActionPerformed

    /* prevStepButtonActionPerformed tells the validation core to backtrack to 
     * the previous back stap available.
     */
    private void prevStepButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_prevStepButtonActionPerformed
        if (validating) validationCore.stepBack();
    }//GEN-LAST:event_prevStepButtonActionPerformed

    /* logCheckBoxActionPerformed enables or disables the log level spinner 
     * according to the value of the check box.
     */
    private void logCheckBoxActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_logCheckBoxActionPerformed
        logSpinner.setEnabled(!logCheckBox.isSelected());
        this.resetLogLevel();
    }//GEN-LAST:event_logCheckBoxActionPerformed

    /* defRulesButtonActionPerformed sets the default validation rule set as the
     * current one.
     */
    private void defRulesButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_defRulesButtonActionPerformed
        try 
        {
            Reader in = validationCore.getDefRulesReader();
            this.setValidationRules(defRules, in);
            defaultRules = true;
            defRulesButton.setEnabled(false);
            rules.setTitle(VR_TITLE);
            
            java.awt.event.ActionListener[] als = resourceBrowser.getActionListeners();
            java.awt.event.ActionListener al = als.length > 0 ? als[0] : null;
            if (al != null) resourceBrowser.removeActionListener(al);
            resourceBrowser.setSelectedItem(rulesName = validationCore.getDefRulesName());
            if (al != null) resourceBrowser.addActionListener(al);
        }
        catch (Exception ex) { log.error(ex); }
    }//GEN-LAST:event_defRulesButtonActionPerformed

    private void openRules(Reader in, String name)
    {
        try { rules.setSelected(true); } catch (Exception e) {}
        rules.setTitle(VR_TITLE + " - " + name);
        this.setValidationRules(null, new BufferedReader(in));
        defaultRules = false;
        defRulesButton.setEnabled(true);
    }
    
    /* openRulesButtonActionPerformed opens a new validation rule source, then
     * parsing it and setting it as current one.
     */
    private void openRulesButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_openRulesButtonActionPerformed
        fileChooser.setFileFilter(VR_FILTER);
        final File vrFile = this.openFile(fileChooser);
        
        if (vrFile != null)
        {
            new Thread(new Runnable()
            {
                public void run()
                {
                    try
                    {
                        Reader in = null;
                        in = new FileReader(vrFile);
                        in = new BufferedReader(in);
                        openRules(in, vrFile.getName());
                        lastDir = new File(vrFile.getPath());
                    }
                    catch (Throwable t) { log.error(t); }
                }
            }).start();
        }
    }//GEN-LAST:event_openRulesButtonActionPerformed
    
    /* nextErrorButtonActionPerformed runs a syntax check on the current 
     * protocol source, selecting the first error found.
     */    
    private void nextErrorButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_nextErrorButtonActionPerformed
        
        Log.Message[] errors = selectedSource.checkSource(false);
        if (errors.length > 0)
        {
            Log.Message error = errors[0];
            this.selectWarning(selectedSource.getSourceArea(), error);
            stateBar.setText(" " + error);
            this.showNotify("Error found", error.toString());
        }
        else
        {
            nextErrorButton.setEnabled(false);
            stateBar.setText(" protocol source is correct");
        }
        
        warningsButton.setEnabled(selectedSource.hasWarnings());

    }//GEN-LAST:event_nextErrorButtonActionPerformed

    /* infSaveButtonActionPerformed shows a file chooser dialog to let the user
     * choose the name and path of the tagging to be saved.
     */
    private void infSaveButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_infSaveButtonActionPerformed
        if (lastDir != null)
            fileChooser.setCurrentDirectory(lastDir);
        
        int fcCode = fileChooser.showSaveDialog(inferenceDialog);
        if (fcCode == JFileChooser.APPROVE_OPTION)
        {
            Writer sourceWriter = null;
            try
            {
                try
                {
                    File outFile = fileChooser.getSelectedFile();
                    sourceWriter = new BufferedWriter(new FileWriter(outFile));
                    infSourceArea.write(sourceWriter);
                    String name = outFile.getName();
                    stateBar.setText(" current tagging saved as: " + name);
                }
                finally { if (sourceWriter != null) sourceWriter.close(); }
            }
            catch (IOException ioe) { log.error(ioe); }
        }
    }//GEN-LAST:event_infSaveButtonActionPerformed

    /* taggingBoxActionPerformed displays the source corresponding to the
     * selected tagging and enables the appropriate buttons.
     */
    private void taggingBoxActionPerformed(java.awt.event.ActionEvent evt)
    {
        Protocol tagging = taggingBox.getSelectedTagging();
        infSourceArea.setText(tagging.getProtocolImage());
        infSourceArea.setCaretPosition(0);
        
        boolean protocolSelected = taggingBox.isProtocolSelected();
        infSaveButton.setEnabled(!protocolSelected);
        infRunButton.setEnabled(!validating && protocolSelected);
    }
    
    /* infRunButtonActionPerformed runs the tag inference as a separate thread,
     * this way you can keep using the tool to do something else.
     */    
    private void infRunButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_infRunButtonActionPerformed
        
        new Thread(new Runnable()
        {
            public final void run()
            {
                infStopButton.setEnabled(infering = true);
                infRunButton.setEnabled(false);
                infSaveButton.setEnabled(false);
                infCommitHints.setEnabled(false);
                infReorder.setEnabled(false);
                setValidationControlState(false);

                inferenceDialog.setTitle("Tag Inference (working)");
                String name = infProtocol.getProtocolName();
                log.info("[" + name + "] starting tagging inference");
                
                int logLevel = log.getLogLevel();
                if (logCheckBox.isSelected()) log.setLogLevel(Log.MUTE);
                
                long time = System.currentTimeMillis();
                Protocol[] taggings = 
                    inferenceCore.inferTaggings(infProtocol);
                time = System.currentTimeMillis() - time;
                
                if (logCheckBox.isSelected()) log.setLogLevel(logLevel);
                if (inferenceCore.isStopped())
                {
                    inferenceDialog.setTitle("Tag Inference (stopped)");
                    log.info("[" + name + "] tagging inference stopped");
                }
                else
                {
                    setDialogTaggings(infProtocol, time);
                    String msg = "] tagging inference complete - elapsed ";
                    log.info("[" + name + msg + time + " ms");
                }
                
                boolean enabled = !inferenceCore.isUsingCommitHintsDefault();
                infCommitHints.setEnabled(enabled);
                infReorder.setEnabled(true);
                infSource.updateTaggings();
                infRunButton.setEnabled(taggingBox.isProtocolSelected());
                infStopButton.setEnabled(infering = false);
                setValidationControlState(true);
                
                showNotify("Tag inference complete", "taggings found: " + infProtocol.getTaggings().length);
            }
        }).start();
    }//GEN-LAST:event_infRunButtonActionPerformed

    /* infStopButtonActionPerformed tells the inference core to stop. */
    private void infStopButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_infStopButtonActionPerformed
        infStopButton.setEnabled(false);
        inferenceCore.setStopped(true);
        infSourceArea.setText("");
    }//GEN-LAST:event_infStopButtonActionPerformed

    private void infCommitHintsActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_infCommitHintsActionPerformed
        inferenceCore.setUsingCommitHints(infCommitHints.isSelected());
    }//GEN-LAST:event_infCommitHintsActionPerformed
    
    /* inferenceButtonActionPerformed displays the inference dialog, that shows
     * the inference action buttons, the selected protocols source, the 
     * corresponding taggings and a couple of options.
     */
    private void inferenceButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_inferenceButtonActionPerformed

        if (!infering)
        {
            infSource = selectedSource;
            infProtocol = infSource.getActualProtocol();
            Protocol[] taggings = infProtocol.getTaggings();
            
            if (taggings == null)
            {
                inferenceDialog.setTitle("Tag Inference");
                
                boolean b = inferenceCore.isUsingCommitHints();
                infCommitHints.setSelected(b);
                b = inferenceCore.isUsingCommitHintsDefault();
                infCommitHints.setEnabled(!b);

                b = inferenceCore.isUsingProtocolReordering();
                infReorder.setSelected(b);

                infSourceArea.setText(infProtocol.getProtocolImage());
                infSourceArea.setCaretPosition(0);

                taggingBox.setProtocol(infProtocol);

                infStopButton.setEnabled(false);
                infSaveButton.setEnabled(false);
                infRunButton.setEnabled(!validating);
            }
            else
                this.setDialogTaggings(infProtocol, 0);
        }
        
        this.showDialog(inferenceDialog);        
    }//GEN-LAST:event_inferenceButtonActionPerformed
            
    private void infReorderActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_infReorderActionPerformed
        inferenceCore.setUsingProtocolReordering(infReorder.isSelected());
    }//GEN-LAST:event_infReorderActionPerformed
    
    /* setAppropriateLogLevel computes and sets the appropriate log level 
     * according to the current step level.
     */
    private final void setAppropriateLogLevel()
    {
        if (logCheckBox.isSelected() && srcTabbedPane.getTabCount() > 0)
            log.setLogLevel(curStepLevel.getLevel() / LEVEL_FACTOR + 1);        
    }
    
    /* stepLevelBoxActionPerformed updates the current step level. */
    private void stepLevelBoxActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_stepLevelBoxActionPerformed
        curStepLevel = this.getCurStepLevel();
        validationCore.setCurStepLevel(curStepLevel);
        this.setAppropriateLogLevel();
    }//GEN-LAST:event_stepLevelBoxActionPerformed
    
    private void exitButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_exitButtonActionPerformed
        if (errStream != null) errStream.close();
        if (errLogFile.length() == 0) errLogFile.delete();
        if (outStream != null) outStream.close();
        System.exit(0);
    }//GEN-LAST:event_exitButtonActionPerformed

    /* dialogCloseButtonActionPerformed hides the dialog containing the button
     * that generated the event.
     */
    private void dialogCloseButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_dialogCloseButtonActionPerformed
        JButton button = (JButton)evt.getSource();
        JDialog dialog = (JDialog)button.getRootPane().getParent();
        dialog.setVisible(false);
    }//GEN-LAST:event_dialogCloseButtonActionPerformed
    
    private void warningsButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_warningsButtonActionPerformed
        this.displayMessageDialog(selectedSource.getWarnings(), "Warnings");
    }//GEN-LAST:event_warningsButtonActionPerformed
    
    private void warningBoxActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_warningBoxActionPerformed
        this.selectWarning();
    }//GEN-LAST:event_warningBoxActionPerformed
    
    /* saveButtonActionPerformed writes to disk the current protocol source,
     * notifying the detection of warning if any.
     */    
    private void saveButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_saveButtonActionPerformed
        if (selectedSource.save())
        {
            warningsButton.setEnabled(selectedSource.hasWarnings());
            String fileName = selectedSource.getSourceFile().getName();
            stateBar.setText(" saved file: " + fileName);
        }
    }//GEN-LAST:event_saveButtonActionPerformed
    
    /* srcTabbedPaneStateChanged updates the state of the toolbar buttons
     * according to the action performed on the tabbed pane
     */    
    private void srcTabbedPaneStateChanged(javax.swing.event.ChangeEvent evt) {//GEN-FIRST:event_srcTabbedPaneStateChanged
        selectedSource =
            (JProtocolSourceArea)srcTabbedPane.getSelectedComponent();
        
        boolean b = selectedSource != null;
        boolean warnings = false;
        
        if (b)
        {
            selectedSource.requestFocus();
            warnings = this.manageParserMessages(selectedSource.isEditable());
            this.updateTabber();
        }
        
        saveButton.setEnabled(b && selectedSource.isModified());
        
        this.updateControlStateButtons();
        this.updateSourceFrameTitle();
    }//GEN-LAST:event_srcTabbedPaneStateChanged
    
    private void updateTabber()
    {
        Protocol p;
        Object obj = tabbedPane.getSelectedComponent();
        if (obj instanceof TabPlugin && (p = selectedSource.getProtocol()) != null)
        {
            TabPlugin tp = (TabPlugin)obj;
            tp.tabSelected(p);
        }
    }            
    
    /* updateSourceFrameTitle changes the title of the source editor window
     * according to the selected tab mode (editing/displaying).
     */
    private boolean updateSourceFrameTitle()
    {
        if (selectedSource == null) return false;
        boolean editable = selectedSource.isEditable();
        
        String title = "Protocol Source";
        if (editable) title += selectedSource.isReadonly() ? " (readonly mode)" : " (editing mode)";
        source.setTitle(title);
        
        return editable;
    }
    
    /* editButtonActionPerformed switches between source editing mode and
     * source displaying mode (if source is valid).
     */
    private void editButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_editButtonActionPerformed
        boolean editable = !selectedSource.isEditable();
        
        selectedSource.setEditable(editable);
        saveButton.setEnabled(selectedSource.isModified());
        
        this.updateControlStateButtons();
        this.updateSourceFrameTitle();        
        this.manageParserMessages(editable);
    }//GEN-LAST:event_editButtonActionPerformed

    /* manageParserMessages shows the dialog corresponding to the kind of 
     * messages found and enables the appropiate buttons.
     */
    private boolean manageParserMessages(boolean editable)
    {
        boolean warnings = selectedSource.hasWarnings();
        boolean hasErrors = selectedSource.hasErrors();
        nextErrorButton.setEnabled(hasErrors);
        
        if (hasErrors)
        {
            String fn = selectedSource.getSourceFile().getName();
            Log.Message[] errors = selectedSource.getErrors();
            int size = errors.length;
            for (int i = 0; i < size; ++i)
                errConsole.silentAppend("[" + fn + "] " + errors[i] + "\n");
            
            this.displayMessageDialog(errors, "Errors");
        }
        else
        {
            warningsButton.setEnabled(warnings);
            if (warnings && !editable)
                stateBar.setText(" there have been warnings");
        }
        
        return warnings;
    }
    
    /* stopButtonActionPerformed tells the validationCore to stop as soon as
     * possible.
     */
    private void stopButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_stopButtonActionPerformed
        validationCore.setStopped(true);
    }//GEN-LAST:event_stopButtonActionPerformed
    
    /* nextStepButtonActionPerformed tells the validationi core to proceed to 
     * the next validation step, running the validation if necessary.
     */
    private void nextStepButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_nextStepButtonActionPerformed
        if (validating)
        {
            validationCore.setCurStepLevel(this.getCurStepLevel());
            validationCore.proceed();
        }
        else this.startValidation((StepLevel)stepLevelBox.getSelectedItem());
    }//GEN-LAST:event_nextStepButtonActionPerformed
    
    private void viewRuleFrameActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_viewRuleFrameActionPerformed
        this.showFrame(rules);
    }//GEN-LAST:event_viewRuleFrameActionPerformed
    
    private void viewTreeActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_viewTreeActionPerformed
        this.showRules(
        viewTree, "Tree", ruleScrollPane,
        viewRules, "Source", ruleSrcScrollPane);
    }//GEN-LAST:event_viewTreeActionPerformed
    
    private void viewRulesActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_viewRulesActionPerformed
        this.showRules(
            viewRules, "Source", ruleSrcScrollPane,
            viewTree, "Tree", ruleScrollPane);
    }//GEN-LAST:event_viewRulesActionPerformed
    
    private void disposeActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_disposeActionPerformed
        desktopPanel.undockAll();
        this.showFrame(console);
        this.showFrame(containers);
        this.showFrame(source);
        this.showFrame(rules);
        this.disposeIFrames();
    }//GEN-LAST:event_disposeActionPerformed
    
    private void viewSourceActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_viewSourceActionPerformed
        this.showFrame(source);
    }//GEN-LAST:event_viewSourceActionPerformed
    
    private void viewContainersActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_viewContainersActionPerformed
        this.showFrame(containers);
    }//GEN-LAST:event_viewContainersActionPerformed
    
    private void viewConsoleActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_viewConsoleActionPerformed
        this.showFrame(console);
    }//GEN-LAST:event_viewConsoleActionPerformed
    
    /* windowMenuMousePressed enables or disables the menu items of the view 
     * menu according to the docking state of the corresponding windows.
     */
    private void windowMenuMousePressed(java.awt.event.MouseEvent evt) {//GEN-FIRST:event_windowMenuMousePressed
        undock.setEnabled(desktopPanel.getSelectedFrame().isDocked());
        north.setEnabled(desktopPanel.canDock(BorderLayout.NORTH));
        west.setEnabled(desktopPanel.canDock(BorderLayout.WEST));
        south.setEnabled(desktopPanel.canDock(BorderLayout.SOUTH));
        east.setEnabled(desktopPanel.canDock(BorderLayout.EAST));
    }//GEN-LAST:event_windowMenuMousePressed
    
    /* if validation or inference are running, they are stopped before closing
     * the corresponding protocol source.
     */
    private void closeActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_closeActionPerformed
        if (selectedSource == currentSource)
        {
            this.collapseLastPath();
            containerContent.setText("");
            if (validating) this.stopButtonActionPerformed(null);
        }
        
        if (infering && selectedSource == infSource)
                this.infStopButtonActionPerformed(null);
        
        this.closeSelectedSource();
    }//GEN-LAST:event_closeActionPerformed
    
    /* before exiting exitActionPerformed checks if each open protcol source is
     * modified: if so it doesn't remove it from the source editor window and
     * asks a confirm to the exit command.
     */
    private void exitActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_exitActionPerformed
        boolean modified = false;
        
        for (int i = srcTabbedPane.getTabCount() - 1; i >= 0; --i)
        {
            JProtocolSourceArea psa =
                (JProtocolSourceArea)srcTabbedPane.getComponentAt(i);
            
            boolean m = psa.isModified();
            if (!m) this.removeProtocolSource(psa);
            modified = modified || m;
        }
        
        if (modified) this.showDialog(exitDialog);
        else this.exitButtonActionPerformed(null);
    }//GEN-LAST:event_exitActionPerformed
    
    /* validateActionPerformed runs a new validation session or completes the 
     * current one.
     */
    private void validateActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_validateActionPerformed
        this.setValidationControlState(false);
        this.collapseLastPath();
        
        if (validating)
        {
            curStepLevel = StepLevel.STRAIGHT_ON_LEVEL;
            validationCore.setCurStepLevel(curStepLevel);
            this.setAppropriateLogLevel();
            validationCore.proceed();
        }
        else this.startValidation(StepLevel.STRAIGHT_ON_LEVEL);
    }//GEN-LAST:event_validateActionPerformed
    
    /* openFile displays a open file chooser dialog with initial directory equal
     * to the last valid one, returning the file choosed.
     */
    private File openFile(JFileChooser fileChooser)
    {
        if (lastDir != null)
            fileChooser.setCurrentDirectory(lastDir);
        
        if (fileChooser.showOpenDialog(this) == JFileChooser.APPROVE_OPTION)
            return fileChooser.getSelectedFile();
        else 
            return lastDir = null;
    }
    
    /* openActionPerformed opens a new protocol source, then creating the 
     * corresponding tab in the sourc editor window.
     */    
    private void openActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_openActionPerformed
        
        fileChooser.setFileFilter(ROSPI_FILTER);
        final File src = this.openFile(fileChooser);
        
        if (src != null)
        {
            new Thread(new Runnable()
            {
                public void run() { addProtocolSource(src); }
            }).start();
        }

    }//GEN-LAST:event_openActionPerformed
    
    /* logSpinnerStateChanged updates the log level according to the one 
     * associated to the spinner.
     */
    private void logSpinnerStateChanged(javax.swing.event.ChangeEvent evt) {//GEN-FIRST:event_logSpinnerStateChanged
        Integer logLevel = (Integer)logSpinner.getValue();
        log.setLogLevel(logLevel.intValue());
    }//GEN-LAST:event_logSpinnerStateChanged
    
    /* tabSelected puts the logConsole component into the appropriate parent
     * component according to the selected tab (desktop or console).
     */
    private void tabSelected(javax.swing.event.ChangeEvent evt) {//GEN-FIRST:event_tabSelected
        int index = tabbedPane.getSelectedIndex();
        if (curTabIndex != index)
        {
            curTabIndex = index;
            Component selectedComponent = tabbedPane.getSelectedComponent();
            if (selectedComponent == desktopPanel) 
            {
                console.getContentPane().add(logScrollPane);
                if (selectedSource != null) selectedSource.requestFocus();
            }
            else if (selectedComponent == consolePanel)
                consolePanel.add(logScrollPane, BorderLayout.CENTER);
            else if (selectedComponent instanceof TabPlugin)
            {
                TabPlugin tp = (TabPlugin)selectedComponent;
                tp.tabSelected(selectedSource != null ? selectedSource.getProtocol() : null);
            }
        }
    }//GEN-LAST:event_tabSelected
    
    private void undockActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_undockActionPerformed
        desktopPanel.getSelectedFrame().undock();
    }//GEN-LAST:event_undockActionPerformed
    
    private void westActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_westActionPerformed
        desktopPanel.getSelectedFrame().dock(BorderLayout.WEST);
    }//GEN-LAST:event_westActionPerformed
    
    private void southActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_southActionPerformed
        desktopPanel.getSelectedFrame().dock(BorderLayout.SOUTH);
    }//GEN-LAST:event_southActionPerformed
    
    private void northActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_northActionPerformed
        desktopPanel.getSelectedFrame().dock(BorderLayout.NORTH);
    }//GEN-LAST:event_northActionPerformed
    
    private void eastActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_eastActionPerformed
        desktopPanel.getSelectedFrame().dock(BorderLayout.EAST);
    }//GEN-LAST:event_eastActionPerformed
    
    /** Exit the Application */
    private void exitForm(java.awt.event.WindowEvent evt) {//GEN-FIRST:event_exitForm
        this.exitActionPerformed(null);
    }//GEN-LAST:event_exitForm
    
    /** Runs the application, trying to set the java look'n'feel.
     * @param args A file with the name equal to the value of the first argument
     * is looked for and, if found, opened.
     * @throws InitializationException Any exception raised during the 
     * application initialization is caught and set as cause of the 
     * initialization one.
     */
    public static void main(String args[])
    {
        String lookAndFeel = UIManager.getCrossPlatformLookAndFeelClassName();
        try { UIManager.setLookAndFeel(lookAndFeel); }
        catch (Exception e) { e.printStackTrace(System.err); }

        JFrame.setDefaultLookAndFeelDecorated(true);
        JDialog.setDefaultLookAndFeelDecorated(true);

        try
        {
            Pear pear = new Pear();

            int size = args.length;
            if (size > 0) 
            {
                for (int i = 0; i < size; ++i)
                {
                    File curFile = new File(args[i]);
                    if (curFile.exists()) pear.addProtocolSource(curFile);
                }
            }

            pear.setVisible(true);
        }
        catch (Throwable t)
        {
            Log.SIMPLE_LOG.error(t);
            System.exit(1);
        }
    }
    
    /* addProtocolSource adds a new tab in the source editor window
     * containig the source and the parse tree (if the source is valid) of the
     * protcol encoded in the file argument.
     */
    private void addProtocolSource(java.io.File file)
    {
        if (validationCore.getValidationRules() == null) return;
        
        JProtocolSourceArea psa = (JProtocolSourceArea)openedFiles.get(file);
        
        if (psa == null)
        {
            this.stateBar.setText(" selected file: " + file.getName());
            try
            {
                psa = new JProtocolSourceArea(validationCore, file, log);
                lastDir = new File(file.getPath());
                this.initProtocolSource(psa, file);
            }
            catch (Throwable e) { log.error(e); }
        }
        else srcTabbedPane.setSelectedComponent(selectedSource = psa);
    }
    
    private void addProtocolSource(Reader rd, String name)
    {
        if (validationCore.getValidationRules() == null) return;
        
        File file = new File(TMP_PREFIX + name);
        JProtocolSourceArea psa = (JProtocolSourceArea)openedFiles.get(file);
        
        if (psa == null)
        {
            this.stateBar.setText(" selected file: " + file.getName());
            try
            {
                psa = new JProtocolSourceArea(validationCore, rd, log);
                psa.setSourceFile(file);
                this.initProtocolSource(psa, file);
            }
            catch (Throwable e) { log.error(e); }
        }
        else srcTabbedPane.setSelectedComponent(selectedSource = psa);
    }
    
    private void initProtocolSource(JProtocolSourceArea psa, File file)
    {
        psa.setComponentFont(DEFAULT_FONT);
        srcTabbedPane.add(psa, file.getName());
        srcTabbedPane.setSelectedIndex(srcTabbedPane.getTabCount() - 1);
        openedFiles.put(file, psa);
        this.updateControlStateButtons();
        psa.addPropertyChangeListener(this);
        try { source.setSelected(true); } catch (Exception ex) {}
    }
    
    
    /** Handles the change of the current protocol source.
     * @param evt The event to be handled.
     */
    public void propertyChange(java.beans.PropertyChangeEvent evt) 
    {
        if (evt.getPropertyName().equals(JProtocolSourceArea.PROPERTY_NAME)) 
        {
            boolean modified = selectedSource.isModified();
            saveButton.setEnabled(modified);
            nextErrorButton.setEnabled(modified);
        }
    }
    
    private void setValidationControlState(boolean enabled) 
    {
        if (!validating)
        {
            enabled = enabled && !closing;
            validateButton.setEnabled(enabled);
            validate.setEnabled(enabled);
            nextStepButton.setEnabled(enabled);
            stepLevelBox.setEnabled(enabled);
            logCheckBox.setEnabled(enabled);
        }
    }
    
    private void setFileControlState(boolean enabled) 
    {
        close.setEnabled(enabled);
        closeButton.setEnabled(enabled);
        saveButton.setEnabled(enabled);
    }
    
    private void setSourceControlState(boolean enabled) 
    {
        warningsButton.setEnabled(enabled);
        editButton.setEnabled(enabled);
        nextErrorButton.setEnabled(enabled);
        if (!validating) inferenceButton.setEnabled(enabled);
    }
    
    /* updateControlStateButtons enables or disables the toolbar buttons
     * according to the current state of the application.
     */
    private final void updateControlStateButtons()
    {
        boolean enabled = srcTabbedPane.getTabCount() > 0;
        close.setEnabled(enabled);
        closeButton.setEnabled(enabled);
        openRulesButton.setEnabled(!enabled);
        defRulesButton.setEnabled(!enabled && !defaultRules);
        
        resourceBrowser.setEnabled(!enabled || !rulesEnabled);
                
        logSpinner.setEnabled(!enabled || !logCheckBox.isSelected());
        
        boolean b = selectedSource != null;
        boolean editable = !validating || selectedSource != currentSource;
        editButton.setEnabled(enabled && editable && b);
        
        warningsButton.setEnabled(b && selectedSource.hasWarnings());
        nextErrorButton.setEnabled(b && selectedSource.hasErrors());
        saveButton.setEnabled(b && selectedSource.isModified());
        
        b = b && selectedSource.getProtocol() != null;
        inferenceButton.setEnabled(b && selectedSource != currentSource);
        
        b = b && !(infering && selectedSource == infSource) && enabled;
        this.setValidationControlState(b);
        prevStepButton.setEnabled(b && validationCore.isBackStepAvailable());
    }
    
    /* showFrame brings on the foreground the selcted window. */
    private void showFrame(JDockableInternalFrame dif) {
        try {
            dif.setVisible(true);
            dif.setIcon(false);
            dif.setSelected(true);
        }
        catch (java.beans.PropertyVetoException e) {}
    }
    
    /* disposeIFrames resets the window positions to the default ones. */
    private void disposeIFrames() {
        try {
            containers.dock(BorderLayout.SOUTH);
            rules.dock(BorderLayout.EAST);
            console.setIcon(true);
            source.setMaximum(true);
            source.setSelected(true);
        }
        catch (java.beans.PropertyVetoException e) {}
    }
    
    /* showRules displays or hides the validation rule tre and its source
     * according to the values set for the corresponding check-boxes.
     */
    private void showRules(
    JCheckBoxMenuItem cur, String title, JComponent component,
    JCheckBoxMenuItem other, String otherTitle, JComponent otherComponent) {
        if (!rules.isVisible()) this.showFrame(rules);
        
        boolean otherState = other.getState();
        java.awt.Container cp = rules.getContentPane();
        
        if (cur.getState()) {
            // if the other checkbox is selected too, tabbedPanel must be used
            if (otherState) {
                cp.remove(otherComponent);
                cp.add(ruleTabbedPane);
                ruleTabbedPane.addTab(otherTitle, otherComponent);
                ruleTabbedPane.addTab(title, component);
            }
            else cp.add(component);
        }
        else {
            if (otherState) {
                ruleTabbedPane.removeAll();
                cp.remove(ruleTabbedPane);
                cp.add(otherComponent);
            }
            else cp.remove(component);
        }
    }
    
    /* startValidation runs a validation session with throwse requested step
     * level, setting the related buttons to the appropriate state; the
     * validation process and the data-collecting process are run as separate
     * threads in order not to block the main thread.
     */
    private void startValidation(StepLevel curStepLevel) {
        validationCore.setCurStepLevel(this.curStepLevel = curStepLevel);
        validating = true;
        saveButton.setEnabled(false);
        infRunButton.setEnabled(false);
        stopButton.setEnabled(true);
        this.setSourceControlState(false);
        this.setAppropriateLogLevel();
        
        currentSource = selectedSource;
        currentSource.setValidating(true);
        
        new Thread(new Runnable() {
            public final void run() {
                try {
                    Protocol pn = currentSource.getProtocol();
                    validationCore.validateProtocol(pn);
                }
                catch (Validation.StoppedException se) {}
                catch (Throwable e) { log.error(e.getMessage(), e); }
            }
        }).start();
        
        new Thread(new Runnable() {
            public final void run() {
                try { validateCurrentSource(); }
                catch (Throwable t) { log.error(t); }
            }
        }).start();
    }
    
    private final StepLevel getCurStepLevel() {
        return (StepLevel)stepLevelBox.getSelectedItem();
    }
    
    /* validateCurrentSource collects all the informations provided by the
     * validation core and diplays them with the appropriate component.
     */
    private void validateCurrentSource() 
    {
        currentSource.resetColor();
        ValidationOutput vo = new ValidationOutput();
        boolean valid = true;
        
        try 
        {
            boolean stepAvailable = true;
            while (stepAvailable) 
            {
                stepAvailable = validationCore.getStep(vo);
                
                // show the current state
                stateBar.setText(" " + vo.getMessage());
                
                if (curStepLevel != StepLevel.STRAIGHT_ON_LEVEL)
                {
                    // show the content of the containers and the image of
                    // the substitution map
                    this.showContainers(vo);
                    
                    // select the current rule in the validation rule tree
                    int ruleLevel = StepLevel.RULE_LEVEL.getLevel();
                    if (curStepLevel.getLevel() >= ruleLevel)
                    {
                        this.collapseLastPath();
                        Object[] path = vo.getTreePath();
                        if (path.length > 0)
                        {
                            ruleLastPath = new TreePath(path);
                            ruleTree.setSelectionPath(ruleLastPath);
                        }
                    }
                    
                    // select the part of the protocol just validated
                    currentSource.setProtocol(vo.getProtocol());
                    currentSource.select(vo.getInstruction());
                    
                    // activate back step button
                    boolean bsa = validationCore.isBackStepAvailable();
                    prevStepButton.setEnabled(bsa);
                }
            }
            
            valid = vo.isValid();
            currentSource.setValid(valid);
            String notify;
            
            if (valid)
            {
                currentSource.appendRules();
                notify = "protocol has been successfully validated";
                
                this.updateTabbers();
                this.updateTabber();
            }
            else 
            {
                notify = "validation failed at: " + validationCore.getErrMsg();
                stateBar.setText(" " + notify);
                Object[] path = validationCore.getErrPath();
                
                if (path != null) 
                {
                    ruleLastPath = new TreePath(path);
                    ruleTree.setSelectionPath(ruleLastPath);
                }
                
                if (curStepLevel == StepLevel.STRAIGHT_ON_LEVEL) 
                {
                    this.showContainers(vo);
                    currentSource.select(vo.getInstruction());
                }
            }
            
            this.showNotify("Validation complete", notify);
        }
        catch (Validation.StoppedException se) 
        {
            this.collapseLastPath();
            containerContent.setText("");
            currentSource.resetText();
            stateBar.setText(" " + vo.getMessage());
        }
        finally 
        {
            currentSource.setValidating(validating = false);
            currentSource = null;
            
            if (logCheckBox.isSelected()) this.resetLogLevel();
            
            infRunButton.setEnabled(taggingBox.isProtocolSelected());
            stopButton.setEnabled(false);
            prevStepButton.setEnabled(false);
            this.setValidationControlState(false);
/*
            try { Thread.sleep(WAIT_AFTER_VALIDATION); }
            catch (InterruptedException ie) {}
*/
            this.updateControlStateButtons();
        }
    }
    
    private void updateTabbers()
    {
        Iterator it = new ArrayIterator(tabbedPane.getComponents());
        while (it.hasNext())
        {
            Object obj = it.next();
            if (obj instanceof TabPlugin)
            {
                TabPlugin tp = (TabPlugin)obj;
                if (currentSource != null)
                    tp.protocolDecorated(currentSource.getProtocol());
            }
        }
    }
    
    private void showNotify(String title, String message)
    {
        notifyDialog.setTitle(title);
        notifyLabel.setText(message);
        notifyDialog.pack();
        this.showDialog(notifyDialog);
    }
    
    /* resetLogLevel sets the log level to the value of the log level spinner.
     */
    private final void resetLogLevel()
    {
        log.setLogLevel(((Integer)logSpinner.getValue()).intValue());
    }
    
    /* showContainers displays the content of the containers and the
     * substitution map.
     */
    private void showContainers(ValidationOutput vo)
    {
        containerContent.setText(null);
        Iterator it = vo.getContainers().iterator();
        while (it.hasNext())
            containerContent.append(it.next().toString() + "\n");
        
        Map pm = vo.getSubstMap();
        if (pm != null && !pm.isEmpty())
            containerContent.append("\npattern matching: " + pm.entrySet());
    }
    
    /* collapseLastPath hides the branches of the validation rule tree below
     * the name of the current rule.
     */
    private void collapseLastPath()
    {
        ruleTree.clearSelection();
        if (ruleLastPath != ruleRootPath)
        {
            Object[] path = ruleLastPath.getPath();
            TreePath tp = new TreePath(new Object[]{ path[0], path[1] });
            ruleTree.collapsePath(tp);
        }
    }
    
    /* selectWarning selects the part of the protocol source that generated the
     * warning, both in the waring window and in the source editor one,
     * if editable.
     */
    private void selectWarning()
    {
        Log.Message warning = (Log.Message)warningBox.getSelectedItem();
        this.selectWarning(sourceArea, warning);
        if (selectedSource.isEditable())
            this.selectWarning(selectedSource.getSourceArea(), warning);
    }
    
    /* selectWarning selects the part of the protocol source that generated the
     * warning in the source area passed as argument.
     */
    private void selectWarning(JTextArea sourceArea, Log.Message warning) {
        String source = sourceArea.getText();
        if (source.length() > 0)
        {
            try
            {
                int line = warning.getLine() - 1;
                int pos = sourceArea.getLineStartOffset(line);
                pos += warning.getColumn() - 1;
                
                String token = warning.getSource();
                if (token != null && token.length() > 0 &&
                source.charAt(pos) != token.charAt(0))
                {
                    int index = source.indexOf(token, pos);
                    if (index == -1 || sourceArea.getLineOfOffset(index) != line)
                        index = source.lastIndexOf(token, pos);
                    if (index != -1 && sourceArea.getLineOfOffset(index) == line)
                        pos = index;
                    else pos = -1;
                }
                
                if (pos != -1) 
                {
                    int end = pos + token.length();
                    sourceArea.requestFocus();
                    sourceArea.setCaretPosition(pos);
                    sourceArea.moveCaretPosition(end);
                }
            }
            catch (Exception ex) { log.error(ex); }
        }
    }
    
    /* closeSelectedSource closes the current protocol source and is run as
     * separate thread, because the operation may be blocking.
     */
    private void closeSelectedSource() 
    {
        new Thread(new Runnable()
        {
            public final void run()
            {
                closing = true;
                
                setValidationControlState(false);
                setFileControlState(false);
                selectedSource.close();
                removeProtocolSource(selectedSource);
                
                // if the selected source is the last one on which inference has
                // been run, the inference data is no more valid.
                if (selectedSource == infSource)
                {
                    infProtocol = null;
                    infSource = null;
                }
                
                closing = false;

                updateControlStateButtons();
            }
        }).start();
    }
    
    /* displayMessageDialog displays a dialog showing the selected source and
     * allowing to select the messages passed as arguments.
     */
    private void displayMessageDialog(Log.Message[] messages, String title) {
        sourceArea.setText(selectedSource.getSource());
        
        // removing the actionListener prevents the JComboBox from reacting to
        // the successive editing of its data
        java.awt.event.ActionListener al = warningBox.getActionListeners()[0];
        warningBox.removeActionListener(al);
        warningBox.removeAllItems();
        int size = messages.length;
        for (int i = 0; i < size; ++i) warningBox.addItem(messages[i]);
        warningBox.addActionListener(al);
        
        this.selectWarning();
        
        warningDialog.setTitle(title + " (" + size + ")");
        this.showDialog(warningDialog);
    }
    
    /* removeProtocolSource removes from the source editor window the tab
     * corresponding to the selected source.
     */
    private void removeProtocolSource(JProtocolSourceArea psa)
    {
        openedFiles.remove(psa.getSourceFile());
        srcTabbedPane.remove(psa);
        selectedSource =
        (JProtocolSourceArea)srcTabbedPane.getSelectedComponent();
        
        if (openedFiles.size() == 0) this.updateControlStateButtons();
    }
    
    /* setDialogTaggings adds the taggings found by the inference core to the
     * inference window, setting the appropriate title.
     */
    private void setDialogTaggings(Protocol protocol, long time)
    {
        int size = taggingBox.setProtocol(protocol);
        
        if (size > 0)
        {
            taggingBox.setSelectedIndex(1);
            infSaveButton.setEnabled(true);
        }
        
        String title = "Tag Inference (" + size + ")";
        if (time > 0)  title += " - elapsed " + time + " ms";
        inferenceDialog.setTitle(title);
    }
    
    // InternalFrameListener implementation
    
    public final void internalFrameActivated(javax.swing.event.InternalFrameEvent e) {}
    public final void internalFrameClosed(javax.swing.event.InternalFrameEvent e) {}
    public final void internalFrameDeactivated(javax.swing.event.InternalFrameEvent e) {}
    public final void internalFrameDeiconified(javax.swing.event.InternalFrameEvent e) {}
    public final void internalFrameIconified(javax.swing.event.InternalFrameEvent e) {}
    public final void internalFrameOpened(javax.swing.event.InternalFrameEvent e) {}
    
    public void internalFrameClosing(javax.swing.event.InternalFrameEvent e)
    {
        Object src = e.getSource();
        if (src == source)
        {
            if (openedFiles.isEmpty()) source.setVisible(false);
            else this.closeActionPerformed(null);        
        }
    }
    
    
    // Variables declaration - do not modify//GEN-BEGIN:variables
    private javax.swing.JMenuItem about;
    private javax.swing.JDialog aboutDialog;
    private javax.swing.JLabel aboutLogo;
    private javax.swing.JPanel aboutNorthPanel;
    private javax.swing.JPanel aboutPanel;
    private javax.swing.JPanel aboutSouthPanel;
    private javax.swing.JTextArea aboutText;
    private javax.swing.JToolBar actionsToolBar;
    private javax.swing.JPanel buttonPanel;
    private javax.swing.JButton cancelButton;
    private javax.swing.JPanel checkBoxPanel;
    private javax.swing.JButton clearConsole;
    private javax.swing.JPanel clearPanel;
    private javax.swing.JMenuItem close;
    private javax.swing.JButton closeAboutDialog;
    private javax.swing.JButton closeButton;
    private javax.swing.JPanel consolePanel;
    private javax.swing.JPanel consoleSouthPanel;
    private javax.swing.JTextArea containerContent;
    private javax.swing.JScrollPane containerScrollPane;
    private javax.swing.JMenuItem contents;
    private javax.swing.JButton defRulesButton;
    private javax.swing.JMenuItem dispose;
    private javax.swing.JSeparator dockSeparator;
    private javax.swing.JMenu dockTo;
    private javax.swing.JMenuItem east;
    private javax.swing.JButton editButton;
    private javax.swing.JToolBar editToolBar;
    private javax.swing.JScrollPane errScrollPane;
    private javax.swing.JMenuItem exit;
    private javax.swing.JButton exitButton;
    private javax.swing.JDialog exitDialog;
    private javax.swing.JLabel exitLabel;
    private javax.swing.JPanel exitPanel;
    private javax.swing.JMenu fileMenu;
    private javax.swing.JToolBar fileToolBar;
    private javax.swing.JMenu helpMenu;
    private javax.swing.JButton infCloseButton;
    private javax.swing.JCheckBox infCommitHints;
    private javax.swing.JPanel infOptPanel;
    private javax.swing.JPanel infPanel;
    private javax.swing.JCheckBox infReorder;
    private javax.swing.JButton infRunButton;
    private javax.swing.JButton infSaveButton;
    private javax.swing.JScrollPane infScrollPane;
    private javax.swing.JTextArea infSourceArea;
    private javax.swing.JPanel infSouthPanel;
    private javax.swing.JButton infStopButton;
    private javax.swing.JButton inferenceButton;
    private javax.swing.JDialog inferenceDialog;
    private javax.swing.JCheckBox logCheckBox;
    private javax.swing.JTextArea logConsole;
    private javax.swing.JLabel logLevel;
    private javax.swing.JScrollPane logScrollPane;
    private javax.swing.JSpinner logSpinner;
    private javax.swing.JPanel logSpinnerPanel;
    private javax.swing.JMenuBar menuBar;
    private javax.swing.JButton nextErrorButton;
    private javax.swing.JButton nextStepButton;
    private javax.swing.JMenuItem north;
    private javax.swing.JButton notifyButton;
    private javax.swing.JDialog notifyDialog;
    private javax.swing.JLabel notifyLabel;
    private javax.swing.JPanel notifyPanel;
    private javax.swing.JMenuItem open;
    private javax.swing.JButton openButton;
    private javax.swing.JButton openRulesButton;
    private javax.swing.JLabel patienceLabel;
    private javax.swing.JPanel patiencePanel;
    private javax.swing.JSlider patienceSlider;
    private javax.swing.JButton prevStepButton;
    private javax.swing.JComboBox resourceBrowser;
    private javax.swing.JButton resourceButton;
    private javax.swing.JToolBar resourceToolBar;
    private javax.swing.JScrollPane ruleScrollPane;
    private javax.swing.JSeparator ruleSeparator;
    private javax.swing.JTextArea ruleSource;
    private javax.swing.JScrollPane ruleSrcScrollPane;
    private javax.swing.JTabbedPane ruleTabbedPane;
    private javax.swing.JToolBar ruleToolBar;
    private javax.swing.JMenu rulesMenu;
    private javax.swing.JButton saveButton;
    private javax.swing.JTextArea sourceArea;
    private javax.swing.JMenuItem south;
    private javax.swing.JTabbedPane srcTabbedPane;
    private javax.swing.JLabel stateBar;
    private javax.swing.JComboBox stepLevelBox;
    private javax.swing.JButton stopButton;
    private javax.swing.JTabbedPane tabbedPane;
    private javax.swing.JPanel toolBarPanel;
    private javax.swing.JMenuItem undock;
    private javax.swing.JMenuItem validate;
    private javax.swing.JButton validateButton;
    private javax.swing.JMenuItem viewConsole;
    private javax.swing.JMenuItem viewContainers;
    private javax.swing.JMenu viewMenu;
    private javax.swing.JMenuItem viewRuleFrame;
    private javax.swing.JCheckBoxMenuItem viewRules;
    private javax.swing.JMenuItem viewSource;
    private javax.swing.JCheckBoxMenuItem viewTree;
    private javax.swing.JButton warnCloseButton;
    private javax.swing.JPanel warnPanel;
    private javax.swing.JScrollPane warnScrollPane;
    private javax.swing.JPanel warnSouthPanel;
    private javax.swing.JComboBox warningBox;
    private javax.swing.JDialog warningDialog;
    private javax.swing.JButton warningsButton;
    private javax.swing.JMenuItem west;
    private javax.swing.JMenu windowMenu;
    // End of variables declaration//GEN-END:variables
}
