/*
 * BackTrackingHandler.java
 *
 * Created on 21 marzo 2004, 15.29
 */

package it.unive.dsi.pear.util;

import java.util.*;

/** This class implements the backtracking handler.
 * @author Francesco Placella
 */
public class BackTrackingHandler
{
    /** An integer representing the null brand. */    
    public static final int NO_BRAND = -1;
    /** An integer representing an unlimited number of backtracking steps. */    
    public static final int UNLIMITED = Integer.MAX_VALUE;
    
    private int maxBTNumber;
    private int btNumber;
    
    private State curState;
    private final Stack stateStack = new Stack();    
    private final Map brandCount = new HashMap();
    
    /** Holds value of property enabled. */
    private boolean enabled = true;
    
    /** Holds value of property curBrand. */
    private int curBrand = NO_BRAND;
    
    /** It's throwed when the maximum number of backtracking steps allowed is
     * exceeded.
     */    
    public class LimitExcededException extends Exception
    {
        private int limit = maxBTNumber;
        
        /** Creates a new instance of LimitExcededException. */        
        public LimitExcededException()
            { super("backtracking limit exceded: " + maxBTNumber); }
        /** Returns the limit exceeded.
         * @return the limit exceeded
         */        
        public int getLimit() { return this.limit; }
    }
    
    /** This interface models the generic backtracking node. */    
    public interface State
    {
        public void backup(Object field);
        public void setFieldValue(Object field, CloneableElement value);
        public CloneableElement getFieldValue(Object field);        
        public CloneableElement getFieldValue(Object field, CloneableElement d);
        public CloneableElement retrieveFieldValue(Object field);
        public Object getBacktrackingEntity();
        public Iterator fieldIterator();
        public void setBacktrackingEntity(Object entity);
        public int getBrand();
        public void setBrand(int brand);
        public void resetBackup();        
        public State cloneState();
        
    }
    
    /** Creates a new instance of BackTrackingHandler. */
    public BackTrackingHandler()
    {
        this(null);
    }
    
    /** Creates a new instance of BackTrackingHandler.
     * @param startingState the initial state of the backtracking handler
     */    
    public BackTrackingHandler(State startingState)
    {
        this(startingState, UNLIMITED);
    }
    
    /** Creates a new instance of BackTrackingHandler.
     * @param startingState the initial state of the backtracking handler
     * @param maxBTNumber the maximum number of backtracking steps allowed
     */    
    public BackTrackingHandler(State startingState, int maxBTNumber)
    {
        this.maxBTNumber = maxBTNumber;
        this.setStartingState(startingState);
    }

    /** Saves the current state to the stack. */    
    public final void save() { this.save(this.getCurBrand()); }
    /** Saves the current state to the stack marking it with the given brand.
     * @param brand the brand to be associated to the state
     */    
    public void save(int brand)
    {
        if (enabled)
        {
            curState.setBrand(brand);
            stateStack.push(curState.cloneState());
            this.setBrandCount(this.getBrandCount() + 1);
        }
    }
    
    /** Clears the backtracking state. */    
    public final void clear()
    {
        stateStack.clear();
        brandCount.clear();

        // to permit unlimitedness, the test is perfomed
        // with "less or equal then" operator (see below)
        btNumber = 1;
    }    
    
    /** Returns the number of element in the backtracking stack.
     * @return the number of element in the backtracking stack
     */    
    public final int availableCount() { return stateStack.size(); }
    
    /** Restores the state on the top of the backtracking stack.
     * @throws LimitExcededException if the maximum number of backtacking step has been exceeded
     * @return the current state
     */    
    public State backTrack()
        throws LimitExcededException
    {
        if (enabled)
        {
            try
            {
                // if maxBTNumber == Integer.MAX_VALUE,
                // the test below is always true
                if (btNumber <= maxBTNumber)
                {
                    curState = (State)stateStack.pop();
                    this.setBrandCount(this.getBrandCount() - 1);
                    ++btNumber;
                }
                else throw new LimitExcededException();
            }
            catch (EmptyStackException ese) {}
        }
        return curState;
    }
    
    /** Restores the state on the <I>size - count</I>level of the backtracking stack.
     * @param count the number of backtracking steps to be performed
     * @throws LimitExcededException if the maximum number of backtacking step has been exceeded
     * @return the current state
     */    
    public State backTrack(int count)
        throws LimitExcededException
    {
        for (int i = 0; i < count; ++i) this.backTrack();
        return curState;
    }

    /** Checks if there are backtracking nodes available.
     * @return <CODE>true</CODE> if the stack's not empty,
     * <CODE>false</CODE> otherwise.
     */    
    public final boolean isAvailable() { return stateStack.size() > 0; }
    
    /** Removes from the backtracking stack all the nodes marked withe given brand.
     * @param brand the brand of the nodes to be removed
     */    
    public void removeBranded(int brand)
    {
        if (this.getBrandCount(brand) == 0) return;
        
        while (!stateStack.isEmpty())
        {
            State state = (State)stateStack.peek();
            if (state.getBrand() == brand) stateStack.pop();
            else break;
        }
        
        this.setBrandCount(0);
    }
    
    private final void setBrandCount(int i)
    {
        brandCount.put(new Integer(curState.getBrand()), new Integer(i));
    }

    /** Returns the number of nodes marked with the brand of the current state.
     * @return the number of nodes marked with the brand of the current state
     */    
    public final int getBrandCount()
    {
        return this.getBrandCount(curState.getBrand());
    }
    /** Returns the number of nodes marked with the given brand.
     * @param brand the brand of the nodes to be counted
     * @return the number of nodes marked with the given brand
     */    
    public final int getBrandCount(int brand)
    {
        Integer count = (Integer)brandCount.get(new Integer(brand));
        return count != null ? count.intValue() : 0;
    }

    
    
    /** Getter for property curState.
     * @return Value of property curState.
     *
     */
    public State getCurState() { return curState; }
    
    /** Setter for property curState.
     * @param curState New value of property curState.
     *
     */
    public void setCurState(State curState) {
        this.curState = curState;
    }
    
    /** Getter for property maxBTNumber.
     * @return Value of property maxBTNumber.
     *
     */
    public int getMaxBTNumber() {
        return maxBTNumber;
    }
    
    /** Setter for property maxBTNumber.
     * @param maxBTNumber New value of property maxBTNumber.
     *
     */
    public void setMaxBTNumber(int maxBTNumber) {
        this.maxBTNumber = maxBTNumber;
    }
    
    /** Getter for property enabled.
     * @return Value of property enabled.
     *
     */
    public boolean isEnabled() {
        return this.enabled;
    }
    
    /** Setter for property enabled.
     * @param enabled New value of property enabled.
     *
     */
    public void setEnabled(boolean enabled) {
        this.enabled = enabled;
    }

    /** Setter for property curState.
     * @param startingState the initial state
     */
    public void setStartingState(State startingState)
    {
        this.clear();
        this.curState = startingState;    
    }
    
    /** Getter for property stateStack.
     * @return Value of property stateStack.
     *
     */
    public java.util.Stack getStateStack() {
        return stateStack;
    }
    
    /** Getter for property curBrand.
     * @return Value of property curBrand.
     *
     */
    public int getCurBrand() {
        return this.curBrand;
    }
    
    /** Setter for property curBrand.
     * @param curBrand New value of property curBrand.
     *
     */
    public void setCurBrand(int curBrand) {
        this.curBrand = curBrand;
    }
}
