/*
 * DefaultTreeNode.java
 *
 * Created on 30 settembre 2004, 14.19
 */

package it.unive.dsi.pear.util;

import java.util.*;

import javax.swing.tree.TreeNode;

/** This class provides a straightforward implementation of the
 * <CODE>Node</CODE> interface.
 * @author Francesco Placella
 */
public class DefaultTreeNode 
    implements CloneableElement, Node
{
    /** The node parent. */    
    protected TreeNode parent;
    /** The node children. */    
    protected TreeNode[] children;
    /** The node identity. */    
    protected int id;
    
    /** Returns a new instance of a <CODE>NodeIterator</CODE>.
     * @param node the node whose children have to be iterated over
     * @param nodeClass the class of which a child must be instance in order to
     * be iterated
     */    
    protected final NodeIterator getNodeIteratorInstance(
        TreeNode node, Class nodeClass)
    {
        return new NodeIterator(node, nodeClass);
    }
    
    /** Returns the first child found being instance of the given class.
     * If the class is <CODE>null</CODE> the first child is returned.
     * @param c the class to be matched
     */    
    protected javax.swing.tree.TreeNode getNode(Class c)
    {
        int size = this.getChildCount();
        for (int i = 0; i < size; ++i)
        {
            javax.swing.tree.TreeNode node = this.getChildAt(i);
            if (c == null || node.getClass() == c) return node;
        }
        return null;
    }
    
    /** This class provides a way to iterate over the children of a given node,
     * possibly only the instances of a specified class.
     */    
    protected class NodeIterator
        extends CloneableIterator implements Enumeration
    {
        private TreeNode node;
        private Class nodeClass;
        private int index;
        private int size;        
        
        private NodeIterator(TreeNode node, Class nodeClass)
        {
            this.node = node;
            this.nodeClass = nodeClass;
            this.size = node.getChildCount();
            this.index = 0;
        }

        public void remove() { throw new UnsupportedOperationException(); }
        
        public Object next()
        {
            while (true)
            {
                if (index >= node.getChildCount()) 
                    throw new NoSuchElementException();

                TreeNode node = this.node.getChildAt(index++);
                if (nodeClass == null || node.getClass() == nodeClass)
                    return node;
            }
        }
        
        public boolean hasNext()
        {
            while (index < size)
            {
                TreeNode node = this.node.getChildAt(index);
                if (nodeClass == null || node.getClass() == nodeClass)
                    return true;
                ++index;
            }
            return false;
        }
        
        public int getIndex() { return this.index - 1; }
        public String toString()
        {
            return String.valueOf(this.index) + " (" + node + ")";
        }        
        public boolean equals(Object obj)
        {
            if (obj == null || !(obj instanceof NodeIterator)) return false;
            
            NodeIterator ni = (NodeIterator)obj;
            
            return index == ni.index && node.equals(ni.node);
        }
        
        // Enumeration implementation
        public final boolean hasMoreElements() { return this.hasNext(); }
        public final Object nextElement() { return this.next(); }        
    }
    
    /** Returns an iterator over all the descendant nodes belonging to the given 
     * class.
     * @param nodeClass the class to be matched
     * @return the first child found being instance of the given class
     */    
    protected Iterator descendantIterator(Class nodeClass)
    {
        Collection nodes = new LinkedList();
        this.findNodes(this, nodeClass, nodes);
        return nodes.iterator();
    }
    
    /* findNodes collects all the descendant nodes ofthe given class */
    private void findNodes(
        DefaultTreeNode root, Class nodeClass, Collection nodes)
    {
        Iterator it = root.childrenIterator();
        while (it.hasNext())
        {
            DefaultTreeNode child = (DefaultTreeNode)it.next();
            if (child.getClass() == nodeClass) nodes.add(child);
            this.findNodes(child, nodeClass, nodes);
        }
    }
    
    /** Returns a shallow copy of the current node. */
    public DefaultTreeNode cloneNode()
    {
        DefaultTreeNode cloned = null;
        try
        {
            cloned = (DefaultTreeNode)super.clone();
            cloned.children = new TreeNode[this.getChildCount()];
        }
        catch (CloneNotSupportedException cnse) {}
        return cloned;
    }
    
    /** Returns a copy of the subtree having the current node as root. */
    public Object clone()
    {
        DefaultTreeNode cloned = this.cloneNode();
        
        int childNum = this.getChildCount();
        for (int i = 0; i < childNum; ++i)
        {
            DefaultTreeNode child = (DefaultTreeNode)this.getChildAt(i);
            child = (DefaultTreeNode)child.clone();
            cloned.addChild(child, i);
//            child.parent = cloned;
        }
        
        return cloned;
    }

    /** Returns a string representation of the current node. */    
    public String toString() { return this.getClass().getName(); }
    
    /** Returns a string representation of the current node.
     * @param prefix a prefix to which the string representation will be 
     * appended
     */    
    public String toString(String prefix) { return prefix + toString(); }

    /** Returns a string representation of the subtree having the current node
     * as root.
     */    
    public final String getTreeImage() { return this.stringDump(""); }
    public String stringDump(String prefix)
    {
        StringBuffer buf = new StringBuffer(toString(prefix));
        buf.append("\n");
        if (children != null)
        {
            for (int i = 0; i < children.length; ++i)
            {
                DefaultTreeNode n = (DefaultTreeNode)children[i];
                if (n != null) buf.append(n.stringDump(prefix + " "));
            }
        }
        return buf.toString();
    }
    
    /** Returns an iterator over the children of the current node. */    
    public final Iterator childrenIterator()
    {
        return new NodeIterator(this, null);
    }
    
    /** Tells the node to prepare itself to receive the given number of 
     * children.
     * @param cap the number of chidren expected
     */    
    public void ensureCapacity(int cap)
    {
        if (children == null) 
            children = new TreeNode[cap];
        else if (cap >= children.length)
        {
            TreeNode c[] = new TreeNode[cap];
            System.arraycopy(children, 0, c, 0, children.length);
            children = c;
        }
    }
    
    /** Adds a new child to the current node.
     * @param n the child node
     * @param i the index of the child node
     */    
    public void addChild(Node n, int i)
    {
        if (n == null) return;
        this.ensureCapacity(i + 1);
        children[i] = n;
        n.setParent(this);
    }
    
    /** Returns the children of the receiver as an <CODE>Enumeration</CODE>. */    
    public final Enumeration children() { return (Enumeration)childrenIterator(); }
    /** Returns <CODE>true</CODE> if the receiver allows children. */    
    public final boolean getAllowsChildren() { return true; }
    /** Sets the given node as parent of the current one.
     * @param n the parent node
     */    
    public final void setParent(TreeNode n) { parent = n; }
    /** Returns the parent node of the receiver. */    
    public final TreeNode getParent() { return parent; }
    /** Returns <CODE>true</CODE> if the receiver is a leaf. */    
    public final boolean isLeaf() { return this.getChildCount() == 0; }
    /** Returns the child node at index <CODE>childIndex</CODE>. */    
    public final TreeNode getChildAt(int childIndex)
    {
        return children[childIndex];
    }
    /** Returns the number of children nodes the receiver contains. */    
    public final int getChildCount()
    {
        return (children == null) ? 0 : children.length;
    }

    /** eturns the index of node in the receiver's children.
     * @return Returns the index of <CODE>node</CODE> in the receivers
     * children. If the receiver does not contain <CODE>node</CODE>,
     * -1 will be returned.
     */    
    public int getIndex(javax.swing.tree.TreeNode node)
    {
        NodeIterator it = (NodeIterator)this.childrenIterator();
        while (it.hasNext())
            if (it.next().equals(node)) return it.getIndex() - 1;
        return -1;
    }
}
