/*
 * XPEProtocol.java
 *
 * Created on 5 gennaio 2005, 15.32
 */

package it.unive.dsi.pear.plugins.output.xpe;

import it.unive.dsi.pear.core.parser.Instruction;
import it.unive.dsi.pear.core.parser.KeyData;
import it.unive.dsi.pear.core.parser.Protocol;
import it.unive.dsi.pear.core.parser.ProtocolNode;
import it.unive.dsi.pear.core.parser.SequentialProcess;
import it.unive.dsi.pear.core.parser.SequentialProcessCall;

import java.util.Collection;
import java.util.Iterator;

import org.w3c.dom.Element;

/**
 *
 * @author Francesco Placella
 */
public class XPEProtocol extends XPEAbstractNode
{
    // protocol encoding
    static final String ELEMENT = "protocol-encoding";
    static final String EL_XMLNS = "xmlns";
    static final int EL_INDENT = 0;
    
    // associated rules
    static final String ASSOCIATED_RULES = "associated-rules";
    static final int AR_INDENT = EL_INDENT + 1;

    // rule system
    static final String RULE_SYSTEM = "rule-system";
    static final String RS_NAME = "name";
    static final int RS_INDENT = AR_INDENT + 1;
    
    // description
    static final String DESCRIPTION = "description";
    static final int DS_INDENT = EL_INDENT + 1;
    
    // protocol
    static final String PROTOCOL = "protocol";
    static final String PR_NAME = "name";
    static final int PR_INDENT = EL_INDENT + 1;
    static final int SP_INDENT = PR_INDENT + 1;
    
    // instantiation
    static final String INSTANTIATION = "instantiation";
    static final int IN_INDENT = PR_INDENT + 1;

    // key binding
    static final String KEY_BINDING = "key-binding";
    static final String KB_KEY = "key";
    static final int KB_INDENT = IN_INDENT + 1;
    static final int KT_INDENT = KB_INDENT + 1;

    // sequential process calls
    static final String SC_CALLS = "sequential-process-calls";
    static final int SC_INDENT = IN_INDENT + 1;

    // call
    static final String CALL = "call";
    static final String CL_PRINCIPAL = "principal";
    static final String CL_SEQUENTIAL_PROCESS = "sequential-process";
    static final String CL_REPLICATED = "replicated";
    static final int CL_INDENT = SC_INDENT + 1;

    
// ----------------------------------------------------------------------------
    
    private Protocol protocol;
    
    /** Creates a new instance of XPEProtocolEncoding */
    public XPEProtocol(Protocol p) { protocol = p; }
    
    public final String getElementName() { return ELEMENT; }
    
    protected void buildElement(Element e)
    {
        e.setAttribute(EL_XMLNS, XPEAbstractNode.NS_URI);

        this.appendAssociatedRules(e);
        this.appendDescription(e);
        this.appendProtocol(e);
        this.indent(e, EL_INDENT);
    }
    
// ----------------------------------------------------------------------------
    
    private void appendAssociatedRules(Element e)
    {
        Collection c = protocol.getAssociatedRules();
        
        if (c != null)
        {
            Element ar = this.createElement(ASSOCIATED_RULES, e, AR_INDENT);

            Iterator it = c.iterator();
            while (it.hasNext())
            {
                String rules = (String)it.next();
                Element rs = this.createElement(RULE_SYSTEM, ar, RS_INDENT);
                rs.setAttribute(RS_NAME, rules);
            }
            
            this.indent(ar, AR_INDENT);
        }
    }
    
    private void appendDescription(Element e)
    {
        String desc = protocol.getDescription();
        if (desc == null || desc.length() == 0) return;
        Element d = this.createElement(DESCRIPTION, e, DS_INDENT);
        this.appendText(d, desc);
    }
    
    private void appendProtocol(Element e)
    {
        Element pr = this.createElement(PROTOCOL, e, PR_INDENT);
        pr.setAttribute(PR_NAME, protocol.getProtocolName());
        
        Iterator it = protocol.seqProcIterator();
        while (it.hasNext())
        {
            SequentialProcess sp = (SequentialProcess)it.next();
            this.appendXPE(new XPESequentialProcess(sp), pr, SP_INDENT);
        }
        
        this.appendInstantiation(pr);
        this.indent(pr, PR_INDENT);
    }
    
    private void appendInstantiation(Element e)
    {
        Element in = this.createElement(INSTANTIATION, e, IN_INDENT);
        
        Iterator it = protocol.keyBindingIterator();
        while (it.hasNext())
        {
            ProtocolNode pn = (ProtocolNode)it.next();
            KeyData key = (KeyData)pn.getInstruction().getData()[0];
            Element kb = this.createElement(KEY_BINDING, in, KB_INDENT);
            kb.setAttribute(KB_KEY, key.getName());
            this.appendKeyType(kb, key, KT_INDENT);
            this.indent(kb, KB_INDENT);
        }
        
        this.appendCalls(in);
        
        this.indent(in, IN_INDENT);
    }
    
    private void appendCalls(Element e)
    {
        Element sp = this.createElement(SC_CALLS, e, SC_INDENT);
        
        Iterator it = protocol.callIterator();
        while (it.hasNext())
        {
            SequentialProcessCall spc = (SequentialProcessCall)it.next();
            Instruction instr = spc.getInstruction();
            Element cl = this.createElement(CALL, sp, CL_INDENT);
            cl.setAttribute(CL_PRINCIPAL, spc.getAssignedPrincipal().getInfo());
            cl.setAttribute(CL_SEQUENTIAL_PROCESS, instr.getName());
            cl.setAttribute(CL_REPLICATED, String.valueOf(spc.isReplicated()));
            this.appendData(cl, instr.getData(), CL_INDENT + 1);
            this.indent(cl, CL_INDENT);
        }
        
        this.indent(sp, SC_INDENT);
    }
}
