/*
 * JarClassLoader.java
 *
 * Created on 24 settembre 2004, 14.19
 */

package it.unive.dsi.pear.util;

import java.util.zip.*;
import java.util.*;
import java.io.*;
import java.net.URL;

/** This class implements a class loader that searches for the classes to be
 * loaded inside the jar archives found in the given paths. The class exploits
 * the delegation mechanism: the request is forwarded to the default class 
 * loader, if it couldn't find the class this will be searched by the jar class 
 * loader.
 * @author Francesco Placella
 */
public class JarClassLoader extends ClassLoader
{
    private static final int UPDATE_TIME = 10000;
    private static final String CLASS_EXT = ".class";
    private static final String URL_PREFIX = "jar:file:/";
    private static final FileFilter JAR_FILTER = new ExtensionBasedFileFilter("jar");
    
    private final Map foundClasses = new HashMap();
    private final Map foundResources = new HashMap();
    private final Map foundJar = new HashMap();
    private final List jarpath;
    private long updateId = 0;
    private long lastUpdate = 0;
    
    /** Creates new instances of JarClassLoader.
     * @param jarpath an array of string representing a path to be searched for
     * jar archives
     */    
    public JarClassLoader(String[] jarpath)
    {
        super(JarClassLoader.class.getClassLoader());
        this.jarpath = Arrays.asList(jarpath);

        try { this.updateFoundJar(this.jarpath); }
        catch (Exception ex) { ex.printStackTrace(); }
    }
    
    /** Test driver.
     * @param args not used
     */    
    public static final void main(String[] args)
    {
        try 
        {
            String jarpath = "d:\\documents and settings\\checco\\Documenti\\" +
                             "UniVe\\Tesi\\pear\\release\\plugins";
            ClassLoader cl = new JarClassLoader(new String[]{ jarpath });
            Class c = cl.loadClass("it.unive.dsi.pear.plugins.fresh");
            System.out.println(c);
        }
        catch (Throwable t) { t.printStackTrace(); }        
    }
    
    /** Finds the specified class. Overrides <CODE>ClassLoader.findClass</CODE>.
     * @param name the name of the class to be found
     * @throws ClassNotFoundException if the class could not be found
     * @return the resulting <CODE>Class</CODE> object
     */    
    protected Class findClass(String name)
//    public Class loadClass(String name)
        throws ClassNotFoundException
    {
/*
        try { this.updateFoundJar(jarpath); }
        catch (Exception ex) { ex.printStackTrace(); }
 */

        byte[] classData = (byte[])foundClasses.get(name);
        if (classData == null)
            throw new ClassNotFoundException(name);
//            return this.getParent().loadClass(name);
        
        return (Class)this.defineClass(name, classData, 0, classData.length);
    }
    
    /* updateFoundJar build and updates a list of all the classes and resources
     * contained by the jar archives found in the given paths.
     */
    private void updateFoundJar(List jarpath)
        throws IOException
    {
        long time = System.currentTimeMillis();
        if (time - lastUpdate < UPDATE_TIME) return;
        lastUpdate = time;
        
        Iterator it = jarpath.iterator();
        while (it.hasNext())
        {
            File path = new File((String)it.next());
            if (!path.isDirectory() || !path.exists()) continue;
            
            Iterator jt = new ArrayIterator(path.listFiles(JAR_FILTER));
            while (jt.hasNext())
            {
                File jarFile = (File)jt.next();
                if (jarFile.isDirectory()) continue;
                
                long lmod = jarFile.lastModified();
                Long pmod = (Long)foundJar.get(jarFile);
                
                if (pmod == null || pmod.longValue() < lmod)
                {
                    this.updateFoundElements(jarFile, updateId);
                    foundJar.put(jarFile, new Long(lmod));                
                }
            }            
        }
        
        ++updateId;
        lastUpdate = System.currentTimeMillis();
    }
    
    /* updateFoundElements builds or updates a list of the classes and 
     * resources found in the given archive.
     */    
    private void updateFoundElements(File file, long updateId)
        throws IOException
    {
        ZipFile jarFile = new ZipFile(file);
        try
        {
            List entries = Collections.list(jarFile.entries());
            Iterator it = entries.iterator();
            while (it.hasNext())
            {
                ZipEntry je = (ZipEntry)it.next();
                String name = je.getName();

                if (je.isDirectory()) continue;
                else if (!name.endsWith(CLASS_EXT))
                {
                    foundResources.put(name, file);
                    continue;            
                }

                InputStream jes = null;
                try
                {
                    jes = new BufferedInputStream(jarFile.getInputStream(je));
                    int size = (int)je.getSize();

                    if (size > 0)
                    {
                        int readBytes = 0;
                        byte[] data = new byte[size];
                        while (jes.available() > 0) jes.read(data);

                        name = name.replace('/', '.');
                        int nameLength = name.length() - CLASS_EXT.length();
                        foundClasses.put(name.substring(0, nameLength), data);
                    }
                    else System.err.println("can't load class data: " + je);
                }
                finally { jes.close(); }
            }
        }
        finally { jarFile.close(); }
    }

    /** Returns an input stream for reading the specified resource.
     * @param name the resource name
     * @return an input stream for reading the resource, or <CODE>null</CODE>
     * if the resource could not be found
     */    
    public InputStream getResourceAsStream(String name)
    {
        try
        {
            InputStream is = this.getParent().getResourceAsStream(name);
            
            if (is == null)
            {
                name = this.stripPrefix(name);
//                this.updateFoundJar(jarpath);
                File f = (File)foundResources.get(name);
                if (f != null) 
                {
                    ZipFile jf = new ZipFile(f);
                    is = jf.getInputStream(jf.getEntry(name));
                }
            }
            
            return is;
        }
        catch (Exception ex)
        {
            ex.printStackTrace();
            return null;
        }
    }
    
    public URL getResource(String name)
    {
        try
        {
            URL res = this.getParent().getResource(name);
            
            if (res == null)
            {
//                this.updateFoundJar(jarpath);
                File jar = (File)foundResources.get(this.stripPrefix(name));
                if (jar != null) 
                    res = new URL(URL_PREFIX + jar.getCanonicalPath() + "!" + this.addPrefix(name));
            }
            
//            System.err.println(foundResources);
            
            return res;
        }
        catch (Exception ex)
        {
            ex.printStackTrace();
            return null;
        }
    }
    
    public List getResourceList(String ext)
    {
        ArrayList res = new ArrayList(foundResources.size());
        Iterator it = foundResources.keySet().iterator();
        
        while (it.hasNext())
        {
            String name = (String)it.next();
            if (name.toLowerCase().endsWith(ext)) res.add(name);
        }
        
        res.trimToSize();
        return res;
    }
    
    private final String stripPrefix(String name)
    {
        if (name.startsWith("/")) return name.substring(1);
        return name;
    }
    
    private final String addPrefix(String name)
    {
        if (!name.startsWith("/")) return "/" + name;
        return name;
    }
}
